/*
    IONCurses.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cassert>

#include "IONCurses.h"
#include "util/calcTime.h"

using namespace std;

IONCurses::IONCurses (State* state__, WidiController* widiController__,
		      Option* option__):
  NCursesBase (),MainIO (state__, widiController__, option__),
  infoWindow (NULL),infoWindowLines (0),infoWindowCols (0),
  mainWindow (NULL),mainWindowLines (0),mainWindowCols (0),
  messageWindow (NULL),messageWindowLines (0),messageWindowCols (0)
{
  /*do nothing*/;
}

IONCurses::~IONCurses ()
{
  //do nothing
}


/////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////ɽ
/////////////////////////////////////////////////////////////////////
void IONCurses::initialize () throw (Error)
{
  screen=getScreen ();
  lockCurses ();
  (void) set_term (screen);
  infoWindowLines = 4;
  infoWindowCols = COLS;
  infoWindow = newwin (infoWindowLines,infoWindowCols,0,0);

  if (LINES-infoWindowLines-17>4)
    messageWindowLines=LINES-infoWindowLines-MIDI::MAX_PARTS-1;
  else
    messageWindowLines=4;
  messageWindowCols=COLS;
  messageWindow=newwin (messageWindowLines,messageWindowCols,
			infoWindowLines,0);

  mainWindowLines=LINES-infoWindowLines-messageWindowLines;
  mainWindowCols=COLS;
  mainWindow=newwin (mainWindowLines,mainWindowCols,
		     infoWindowLines+messageWindowLines,0);

  assert (infoWindow!=NULL);
  assert (mainWindow!=NULL);
  assert (messageWindow!=NULL);
  unlockCurses ();

  const char whatitis[]=
    "   "
    "instrum.     "
    "vel,On,prg,bkM,bkL,u,"
    "pBend,modul,"
    "  pan,chVol,  exp,"
    "dPdl,sst,sP,revS,"
    "chr,dly,"
    "vibR,vibD,vDly,"
    "COff,Rsn,"
    "atk,dcy,rls,ch.";
  mvwaddnstr (mainWindow,0,0,whatitis,mainWindowCols);
  for (int i = 0; i < MIDI::MAX_PARTS; i++)
    mvwprintw (mainWindow,i+1,0,"%02d",i+1);

    
  MainIO::initialize ();
}

void IONCurses::release ()
{
  //window˴
  delwin (messageWindow);
  delwin (mainWindow);
  delwin (infoWindow);
  //screen(terminal)˴
  lockCurses ();
  (void) set_term (screen);
  endwin ();
  delscreen (screen);
  unlockCurses ();
  MainIO::release ();
}

void IONCurses::update ()
{
  //windowι
  updateInfoWindow ();
  updateMainWindow ();
  //ºݤΥߥʥɽι
  lockCurses ();
  set_term (screen);
  wnoutrefresh (infoWindow);
  wnoutrefresh (mainWindow);
  doupdate ();
  unlockCurses ();
}

void IONCurses::updateInfoWindow ()
{
  mvwprintw (infoWindow,0,0,
	     "file=%s,title=%s",
	     cacheState.songStatus.getFileName ().c_str (),
	     cacheState.songStatus.getTitle ().c_str ());
  mvwprintw (infoWindow,1,0,
	     "TPQN=%d,"
	     "bpm=%d,"
	     "bar=%3d/%3d,"
	     "step=%4d/%4d,"
	     "%d/%d,"
	     "%d/%d,"
	     "key=%d,"
	     "cc=%d,bb=%d,sf=%d,mi=%d,",
	     cacheState.songStatus.getDivision (),
	     getBPM (cacheState.songStatus.getTempo ()),
	     cacheState.songStatus.measures,
	     cacheState.songStatus.totalMeasures,
	     cacheState.songStatus.ticksFromLastMeasure,
	     cacheState.songStatus.getTicksPerMeasure (),
	     
	     (cacheState.songStatus.ticksFromLastMeasure*
	      cacheState.songStatus.getNn ()-1)/
	     cacheState.songStatus.getTicksPerMeasure ()
	     +1,
	     cacheState.songStatus.getDd (),

	     cacheState.songStatus.getNn (),cacheState.songStatus.getDd (),
	     cacheState.songStatus.getKey (),
	     cacheState.songStatus.getCc (),cacheState.songStatus.getBb (),
	     cacheState.songStatus.getSf (),cacheState.songStatus.getMi ());

  const TimeInfo timeInfo = calcTime (cacheState.songStatus.logicalTicks);
  const TimeInfo totalTimeInfo=
    calcTime (cacheState.songStatus.totalLogicalTicks);
  mvwprintw (infoWindow,2,0,
	     "masterVolume=%d,"
	     "%3d'%02d'%02d/%3d'%02d'%02d,"
	     "%5d/%5d",
	     cacheState.sequencerStatus.masterVolume,
	     timeInfo.minute,timeInfo.second,timeInfo.milisecond/10,
	     totalTimeInfo.minute,totalTimeInfo.second,totalTimeInfo.milisecond/10,
	     cacheState.songStatus.events,cacheState.songStatus.totalEvents
	     );
  mvwprintw (infoWindow,3,0,
	     "instrum=%s",cacheState.sequencerStatus.displayedLetter);
}

void IONCurses::updateMainWindow ()
{
  for (int port = 0; port < MIDI::MAX_PORT; port++)
    {
      for (int channel = 0; channel < MIDI::MAX_CHANNEL; channel++)
	{
	  const int LEFT_PADDING=3;
	  const Part* const part=&cacheState.part[port][channel];//for simplicity
#define MAX_COLS 256
	  char buff[MAX_COLS];
#undef MAX_COLS
	  sprintf (buff,
		   "%-12s "                  //instrument
		   "%3d,%2d,%3d,%3d,%3d,%d,"//vel,nOn,prog,bMSB,bLSB
		   "%5d,%5d,"               //pBend,mod,
		   "%5d,%5d,%5d,"           //pan,chVol,exp,
		   "%4d,%3d,%2d,%4d,"       //dPdl,sost,sP,revSend
		   "%3d,%3d,"               //chorus,delay
		   "%4d,%4d,%4d,"           //vibRate,vibDepth,vibDelay,
		   "%4d,%3d,"               //CutOff,Reson
		   "%3d,%3d,%3d,"     //attak,decay,release
		   "%c%02d",//port
		   part->instrument,
		   ((part->notesOnNumber>0)?part->lastMaxVelocity:0),
		   part->notesOnNumber,
		   part->program,
		   part->bankMSB,
		   part->bankLSB,
		   part->useForRhythmPart,
		 	
		   part->pitchBend,
		   part->modulationWheel,
		 	
		   part->pan,
		   part->channelVolume,
		   part->expression,
		 	
		   part->damperPedal,
		   part->sostenute,
		   part->softPedal,
		   part->reverbSendLevel,
		 	
		   part->chorusSendLevel,
		   part->delaySendLevel,
		 	
		   part->vibrateRate,
		   part->vibrateDepth,
		   part->vibrateDelay,
		 	
		   part->tvfCutOff,
		   part->tvfResonance,
		 	
		   part->envelopeAttackTime,
		   part->envelopeDecayTime,
		   part->envelopeReleaseTime,
		   ((part->port==MIDI::PORT_A)?'A':'B'),part->RxChannel
		   );
	  mvwaddnstr (mainWindow,
		      port*MIDI::MAX_CHANNEL+channel+1, LEFT_PADDING,
		      buff, mainWindowCols-LEFT_PADDING);
	}
    }
}

void IONCurses::coutMessage (const char* message)
{
  //windowι
  scrollok (messageWindow,TRUE);
  scroll (messageWindow);
  scrollok (messageWindow,FALSE);
  mvwprintw (messageWindow,messageWindowLines-1,0,message);
  //ºݤΥߥʥɽι
  lockCurses ();
  set_term (screen);
  wrefresh (messageWindow);
  unlockCurses ();
}

void IONCurses::errorMessage (const char* errorMessage)
{normalMessage (errorMessage);};
void IONCurses::errorMessage (const char* errorMessage,int size)
{normalMessage (errorMessage,size);};


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////for dynamic loading
///////////////////////////////////////////////////////////////newIO<T>
///////////////////////////////////////////////////////////////newMainIO<T>
///////////////////////////////////////////////////////////////infoIO<T>
////////////////////////////////////////////////////////////////////////////

#include "IO/InfoIO.h"
static IO* newIONCurses (State* state, WidiController* widiController,
			 Option* option)
{
  return new IONCurses (state, widiController, option);
}

static MainIO* newMainIONCurses (State* state, WidiController* widiController,
				 Option* option)
{
  return new IONCurses (state, widiController, option);
}

InfoIO infoIONCurses=
{
  newIONCurses,
  newMainIONCurses,
  false
};
