#include <ctime>

#include "SherryWrd.h"
#include "util/KanjiConvert.h"
#include "util/schedYield.h"
#include "IO/ioutil/setSuffix.h"
#include "IO/ioutil/getDirectoryName.h"
#ifndef HAVE_NANOSLEEP
#include "compat/nanosleep.h"
#endif //HAVE_NANOSLEEP

//#define SHERRY_WRD_DEBUG
#undef SHERRY_WRD_DEBUG

using namespace std;

const int SherryWrd::MAX_COLORS;
const int SherryWrd::MAX_VIRTUAL_SCREENS;
const int SherryWrd::MAX_VIRTUAL_SCREEN_WIDTH;
const int SherryWrd::MAX_VIRTUAL_SCREEN_HEIGHT;
const int SherryWrd::MAX_VIRTUAL_PALETTES;
const int SherryWrd::REAL_SCREEN_SIZE_X;
const int SherryWrd::REAL_SCREEN_SIZE_Y;
const int SherryWrd::CHAR_WIDTH;
const int SherryWrd::CHAR_HEIGHT;

SherryWrd::SherryWrd (State* state__, WidiController* widiController__,
		      Option* option__):
  IO (state__, widiController__, option__),
  sherryWrdFile (static_cast<SherryWrdFile*> (NULL)),
  transferPalettesList (),
  transferScreensList ()
{
  for (int i = 0; i < MAX_VIRTUAL_PALETTES; i++)
    virtualPalette[i] = NULL;
  for (int i = 0; i < MAX_VIRTUAL_SCREENS; i++)
    virtualScreenInfo[i] = NULL;
}

SherryWrd::~SherryWrd ()
{
  
}

void SherryWrd::resetVariables ()
{
  for (int i = 0; i < MAX_VIRTUAL_PALETTES; i++)
    {
      delete virtualPalette[i];
      virtualPalette[i] = NULL;
    }
  realPalette.reset ();
  for (int i = 0; i < MAX_VIRTUAL_SCREENS; i++)
    {
      delete virtualScreenInfo[i];
      virtualScreenInfo[i] = NULL;
    }
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::resetVariables transferPalettesList size = "
       << transferPalettesList.size ()
       << " empty ="
       << transferPalettesList.empty ()
       << "transferScreensList size = "
       << transferScreensList.size ()
       << " empty = "
       << transferScreensList.empty ()
       << endl;
#endif
  
  if (!transferPalettesList.empty ())
    transferPalettesList.erase (transferPalettesList.begin (),
				transferPalettesList.end ());
  if (!transferScreensList.empty ())
    transferScreensList.erase (transferScreensList.begin (),
			       transferScreensList.end ());
}


void SherryWrd::newOpenSherryWrdFile (const string& songFileName)
{
  const size_t fileNameLength = songFileName.length ();
  if (sherryWrdFile != NULL && fileNameLength > 3 &&
      strncmp (songFileName.c_str (), sherryWrdFile->getFileName ().c_str (),
	       fileNameLength -3) == 0&&
      songFileName[fileNameLength-4] == '.')
    return;//$B4{$KFI$_9~$_:Q!#(B

  delete sherryWrdFile;
  string sherryWrdFileName = songFileName;
  setSuffix (sherryWrdFileName, "sry");
  //XXXfile$BL>$,BgJ8;z!">.J8;z$,0c$C$F$$$k>l9g$,$"$k!#(B
  //$B$3$l$KBP=h$9$Y$-!#(B
  try
    {
#ifdef DEBUG
      cerr << "SherryWrd::newOpenSherryWrdFile trying load "
	   << sherryWrdFileName << endl;
#endif
      sherryWrdFile = new SherryWrdFile (sherryWrdFileName);
    }
  catch (SherryWrdFile::SError&)
    {
#ifdef DEBUG
      cerr << "file " << sherryWrdFile << " is not found. trying SRY suffix."
	   << endl;
#endif
      setSuffix (sherryWrdFileName, "SRY");
      try
	{
	  sherryWrdFile = new SherryWrdFile (sherryWrdFileName);
	}
      catch (SherryWrdFile::SError&)
	{
	  sherryWrdFile = NULL;
#ifdef DEBUG
      cerr << "file " << sherryWrdFile << " is not found."
	   << " sherry wrd file is not found." << endl;
#endif
	}
    }
}


void SherryWrd::prePlay ()
{
  DEBUG_MESSAGE ("SherryWrd::prePlay");
  IO::prePlay ();
  //$B3F<oJQ?t$r=i4|@_Dj!#(B
  resetVariables ();

  getDirectoryName (songParentDirectory, cacheState.songStatus.getFileName ());
  //sherry wrd file$B$rC5$7$F%m!<%I$9$k!#(B
#ifdef DEBUG
  cerr << "SherryWrd::prePlay fileName="
       << cacheState.songStatus.getFileName () << endl;
#endif
  newOpenSherryWrdFile (cacheState.songStatus.getFileName ());

  if (sherryWrdFile != NULL)
    {//sherry wrd file$B$,8+$D$+$C$?$i1iAUA0$K<B9T$9$kItJ,$r<B9T$9$k!#(B
      
      //$BL?NaNs$N:G=i!#(B
      command = sherryWrdFile->getCommandsList ().begin ();
      
      do
	{
	  //start playing$BKx(Bcommand$B$r<B9T$9$k!#(B
	  //$B2>A[(B->$B<BE>Aw7OL?Na$,<B9T$5$l$?$+$I$&$+(B?
	  bool isRealPaletteChanged = false;//0x31$B$,<B9T$5$l$?$+(B?
	  bool isRealScreenChanged = false;//0x35 or 0x36$B$,<B9T$5$l$?$+(B?
	  try
	    {
	      executeCommand (command, 0UL,
			      isRealPaletteChanged, isRealScreenChanged);
	    }
	  catch (RunTimeError)
	    {
	      //XXX error$B=hM}(B
	      DEBUG_MESSAGE ("SherryWrd::prePlay RunTimeError!");
#if 0
	      delete sherryWrdFile;
	      sherryWrdFile = NULL;
	      resetVariables ();
	      return;
#endif
	    }
	  //updating
	  transferPaletteAndScreen (isRealPaletteChanged, isRealScreenChanged);
	  if (isRealPaletteChanged)
	    updatePalette ();
	  if (isRealScreenChanged)
	    updateScreen ();
	  
	}while ((*command)->getCommandType () != SherryCommand::START_PLAYING);

      command++;//skip SherryCommand::START_PLAYING command
    }

  DEBUG_MESSAGE ("SherryWrd::prePlay done");
}

void SherryWrd::postPlay ()
{
  //$B<B9T8e$N8e;OKv!#(B
  resetVariables ();
  
  IO::postPlay ();
}


void SherryWrd::playLoop ()
{
  if (sherryWrdFile == NULL)
    {//sherry wrd file$B$,L5$+$C$?!#(B
      //wrd file$B$,FI$_9~$^$l$F$$$J$$$3$H$N%A%'%C%/(B
      return;
    }

  for (;;)
    {
      //$B2>A[(B->$B<BE>Aw7OL?Na$,<B9T$5$l$?$+$I$&$+(B?
      bool isRealPaletteChanged = false;//0x31$B$,<B9T$5$l$?$+(B?
      bool isRealScreenChanged = false;//0x35 or 0x36$B$,<B9T$5$l$?$+(B?

      //$B<!2s$N(Bcommand$B<B9T$^$G$NBT$A;~4V$r7W;;(B
      state->lock ();
      state->getState (&cacheState);
      state->unlock ();
      //$BBT$A;~4VJ,$@$1(Bnanosleep
      //$BBT$A;~4V$r=P$9!#(B
      long timeOutUSec = getLeftUSec ((*command)->getTotalDeltaTime (),
				      sherryWrdFile->getTimeBase (),
				      cacheState.songStatus.ticks,
				      cacheState.songStatus.getTempo (),
				      cacheState.songStatus.getDivision ());
      //command$B$N<B9T(B
      try
	{
	  executeCommand (command, timeOutUSec,
			  isRealPaletteChanged, isRealScreenChanged);
	}
      catch (RunTimeError)
	{
	  //XXX error$B=hM}(B
	  DEBUG_MESSAGE ("SherryWrd::mailLoop RunTimeError!");
#if 0
	  return;
#endif
	}
      
      if (isRealPaletteChanged || isRealScreenChanged)
	{
	  //palette$B5Z$S2hA|$NE>Aw(B(realPalette$B$H(BrealScreen$B$N>pJs$r99?7!#(B)
	  transferPaletteAndScreen (isRealPaletteChanged, isRealScreenChanged);
	  
	  //$B<!2s$N(Bcommand$B<B9T$^$G$NBT$A;~4V$r7W;;(B
	  state->lock ();
	  state->getState (&cacheState);
	  state->unlock ();
	  //$BBT$A;~4V$r=P$9!#(B
	  timeOutUSec = getLeftUSec ((*command)->getTotalDeltaTime (),
				     sherryWrdFile->getTimeBase (),
				     cacheState.songStatus.ticks,
				     cacheState.songStatus.getTempo (),
				     cacheState.songStatus.getDivision ());

	  //$BI,MW$J$iBT$A;~4VJ,$@$1(Bnanosleep
	  if (timeOutUSec > 0)
	    {
	      timespec timeSpec;
	      timeSpec.tv_sec = timeOutUSec/1000000;
	      timeSpec.tv_nsec = (timeOutUSec%1000000)*1000;
	      nanosleep (&timeSpec, NULL);
	    }
	  
	  //$B<B:]$KI=<(2hLL$N99?7!#(B
	  if (isRealPaletteChanged)
	    updatePalette ();
	  if (isRealScreenChanged)
	    updateScreen ();
	}

      //pause,$B1iAU=*N;$N=hM}(B
      checkPause ();
      if (cacheState.songStatus.state==SongStatus::DONE ||
	  command == sherryWrdFile->getCommandsList ().end ())
	break;
      schedYield ();
    }
}

void SherryWrd::nonStopPlayLoop (bool waitKey)
{
  for (;;)
    {
      //$B2>A[(B->$B<BE>Aw7OL?Na$,<B9T$5$l$?$+$I$&$+(B?
      bool isRealPaletteChanged = false;//0x31$B$,<B9T$5$l$?$+(B?
      bool isRealScreenChanged = false;//0x35 or 0x36$B$,<B9T$5$l$?$+(B?

      if (waitKey)
	cin.get ();
      
      //command$B$N<B9T(B
      try
	{
	  executeCommand (command, abs (~0L),
			  isRealPaletteChanged, isRealScreenChanged);
	}
      catch (RunTimeError)
	{
	  //XXX error$B=hM}(B
	  DEBUG_MESSAGE ("SherryWrd::mailLoop RunTimeError!");
#if 0
	  return;
#endif
	}
      //$B<B:]$KI=<(2hLL$N99?7!#(B
      transferPaletteAndScreen (isRealPaletteChanged, isRealScreenChanged);
      if (isRealPaletteChanged)
	updatePalette ();
      if (isRealScreenChanged)
	updateScreen ();

      if (command == sherryWrdFile->getCommandsList ().end ())
	break;
    }  
}

void SherryWrd::transferPaletteAndScreen (bool isRealPaletteChanged,
					  bool isRealScreenChanged)
{
  //$B<B:]$KI=<(2hLL$N99?7!#(B
  if (isRealPaletteChanged)
    {
      assert (!transferPalettesList.empty ());
      for (list<uint16>::const_iterator i = transferPalettesList.begin ();
	   i != transferPalettesList.end ();
	   i++)
	{
	  //XXX parameter check
	  if (virtualPalette[*i] != NULL)
	    {
#ifdef SHERRY_WRD_DEBUG
	      cerr << "SherryWrd::updatePaletteAndScreen paletteNumber="
		   << *i << endl;
#endif
	      realTransferPalette (*i);
	    }
#ifdef SHERRY_WRD_DEBUG
	  else
	    {
	      cerr << "SherryWrd::playLoop transfer "
		   << static_cast<int> (*i)
		   << "-th palette but not allocated." << endl;
	    }
#endif
	}
      transferPalettesList.erase (transferPalettesList.begin (),
				  transferPalettesList.end ());
    }

  if (isRealScreenChanged)
    {
      assert (!transferScreensList.empty ());
      for (list<ScreenArea>::const_iterator i = transferScreensList.begin ();
	   i != transferScreensList.end ();
	   i++)
	{
	  //XXX parameter check
	  if (virtualScreenInfo[(*i).screen] != NULL)
	    {
#ifdef SHERRY_WRD_DEBUG
	      cerr << "SherryWrd::updatePaletteAndScreen screen="
		   << (*i).screen << endl;
#endif
	      realTransferScreen ((*i).screen,
				  (*i).fromLeft, (*i).fromTop,
				  (*i).toLeft, (*i).toTop,
				  (*i).width, (*i).height);
	    }
#ifdef DEBUG
	  else
	    {
	      cerr << "SherryWrd::playLoop transfer "
		   << (*i).screen << "-th screen but not allocated."
		   << endl;
	    }
#endif
	}
      transferScreensList.erase (transferScreensList.begin (),
				 transferScreensList.end ());
    }
}

void SherryWrd::realTransferPalette (uint16 paletteNumber)
{
  realPalette.copy (virtualPalette[paletteNumber]);
}


void SherryWrd::executeCommand (SherryWrdFile::Command& command__,
				long leftUSec,
				bool& isRealPaletteChanged,
				bool& isRealScreenChanged)
{
  if (command__ == sherryWrdFile->getCommandsList ().end ())
    return;

  //XXX skip bit$B$N=hM}(B
  //SherryCommand::END_OF_DATA$B$H(BSherryCommand::START_PLAYING$B$O(B
  //$BFCJL07$$!#(B
  do {
    switch ((*command)->getCommandType ())
      {
      case SherryCommand::END_OF_DATA:
	execEndOfData ();
	return;
	break;
      case SherryCommand::START_PLAYING:
	execStartPlaying ();
	return;
	break;

      case SherryCommand::RESERVED_COMMAND://falling through
      case SherryCommand::RESERVED_COMMAND+1://falling through
      case SherryCommand::RESERVED_COMMAND+2://falling through
      case SherryCommand::RESERVED_COMMAND+3://falling through
      case SherryCommand::RESERVED_COMMAND+4://falling through
      case SherryCommand::RESERVED_COMMAND+5://falling through
      case SherryCommand::RESERVED_COMMAND+6://falling through
      case SherryCommand::RESERVED_COMMAND+7://falling through
      case SherryCommand::RESERVED_COMMAND+8://falling through
      case SherryCommand::RESERVED_COMMAND+9://falling through
      case SherryCommand::RESERVED_COMMAND+10://falling through
      case SherryCommand::RESERVED_COMMAND+11://falling through
      case SherryCommand::RESERVED_COMMAND+12://falling through
      case SherryCommand::RESERVED_COMMAND+13://falling through
      case SherryCommand::RESERVED_COMMAND+14://falling through
      case SherryCommand::RESERVED_COMMAND+15:
	execReservedCommand ();
	break;
			  
      case SherryCommand::NEW_PALETTE:
	execNewPalette ((*command__)->getParameter (0).getUint16Array (),
			(*command__)->getParameter (0).getArraySize ());
	break;
      case SherryCommand::DELETE_PALETTE:
	execDeletePalette ((*command__)->getParameter (0).getUint16Array (),
			   (*command__)->getParameter (0).getArraySize ());
	break;
      case SherryCommand::NEW_SCREEN:
	execNewScreen ((*command__)->getParameter (0).getScreenArray (),
		       (*command__)->getParameter (0).getArraySize ());
	break;
      case SherryCommand::DELETE_SCREEN:
	execDeleteScreen ((*command__)->getParameter (0).getUint16Array (),
			  (*command__)->getParameter (0).getArraySize ());
	break;
      case SherryCommand::NEW_SCREEN_AND_LOAD_PNG:
	execNewScreenAndLoadPng ((*command__)->getParameter (0).getUint16 (),
				 (*command__)->getParameter (1).getUint16 (),
				 (*command__)->getParameter (2).getCharArray (),
				 (*command__)->getParameter (2).getArraySize ());
	break;
			  
      case SherryCommand::TRANSFER_PALETTE:
	execTransferPalette ((*command__)->getParameter (0).getUint16 ());
	isRealPaletteChanged = true;
	break;
      case SherryCommand::TRANSFER_FULL_SCREEN:
	execTransferFullScreen ((*command__)->getParameter (0).getUint16 (),
				(*command__)->getParameter (1).getUint16 (),
				(*command__)->getParameter (2).getUint16 ());
	isRealScreenChanged = true;
	break;
      case SherryCommand::TRANSFER_PARTIAL_SCREEN:
	execTransferPartialScreen
	  ((*command__)->getParameter (0).getUint16 (),
	   (*command__)->getParameter (1).getUint16 (),
	   (*command__)->getParameter (2).getUint16 (),
	   (*command__)->getParameter (3).getUint16 (),
	   (*command__)->getParameter (4).getUint16 (),
	   (*command__)->getParameter (5).getUint16 (),
	   (*command__)->getParameter (6).getUint16 ());
	isRealScreenChanged = true;
	break;
			  
      case SherryCommand::SET_PALETTE:
	execSetPalette ((*command__)->getParameter (0).getUint16 (),
			(*command__)->getParameter (1).getPaletteArray (),
			(*command__)->getParameter (1).getArraySize ());
	break;
      case SherryCommand::MERGE_PALETTE:
	execMergePalette
	  ((*command__)->getParameter (0).getUint16 (),
	   (*command__)->getParameter (1).getUint16 (),
	   (*command__)->getParameter (2).getUint16 (),
	   (*command__)->getParameter (3).getUint8 (),
	   (*command__)->getParameter (4).getUint8 (),
	   (*command__)->getParameter (5).getUint8 (),
	   (*command__)->getParameter (6).getUint8 (),
	   (*command__)->getParameter (7).getUint8 (),
	   (*command__)->getParameter (8).getUint8 (),
	   (*command__)->getParameter (9).getUint8 (),
	   (*command__)->getParameter (10).getUint8 ());
	break;
      case SherryCommand::COPY_PALETTE:
	execCopyPalette ((*command__)->getParameter (0).getUint16 (),
			 (*command__)->getParameter (1).getUint16 ());
	break;
      case SherryCommand::MASK_COPY_PALETTE:
	assert ((*command__)->getParameter (2).getArraySize ()== 32);
	execMaskCopyPalette ((*command__)->getParameter (0).getUint16 (),
			     (*command__)->getParameter (1).getUint16 (),
			     (*command__)->getParameter (2).getUint8Array ());
	break;
      case SherryCommand::PARTIAL_COPY_PALETTE:
	execPartialCopyPalette ((*command__)->getParameter (0).getUint16 (),
				(*command__)->getParameter (1).getUint16 (),
				(*command__)->getParameter (2).getPaletteNumberArray (),
				(*command__)->getParameter (2).getArraySize ());
	break;

      case SherryCommand::DRAW_TEXT:
	execDrawText ((*command__)->getParameter (0).getUint16 (),
		      (*command__)->getParameter (1).getUint8 (),
		      (*command__)->getParameter (2).getUint8 (),
		      (*command__)->getParameter (3).getUint8 (),
		      (*command__)->getParameter (4).getUint8 (),
		      (*command__)->getParameter (5).getUint16 (),
		      (*command__)->getParameter (6).getUint16 (),
		      (*command__)->getParameter (7).getCharArray (),
		      (*command__)->getParameter (7).getArraySize ());
	break;
      case SherryCommand::DRAW_BOX:
	execDrawBox ((*command__)->getParameter (0).getUint16 (),
		     (*command__)->getParameter (1).getUint8 (),
		     (*command__)->getParameter (2).getUint16 (),
		     (*command__)->getParameter (3).getUint16 (),
		     (*command__)->getParameter (4).getUint16 (),
		     (*command__)->getParameter (5).getUint16 (),
		     (*command__)->getParameter (6).getUint8 ());
	break;
      case SherryCommand::DRAW_VIRTIAL_LINE:
	execDrawVirtialLine ((*command__)->getParameter (0).getUint16 (),
			     (*command__)->getParameter (1).getUint8 (),
			     (*command__)->getParameter (2).getUint16 (),
			     (*command__)->getParameter (3).getUint16 (),
			     (*command__)->getParameter (4).getUint16 (),
			     (*command__)->getParameter (5).getUint8 ());
	break;
      case SherryCommand::DRAW_HORIZONTAL_LINE:
	execDrawHorizontalLine ((*command__)->getParameter (0).getUint16 (),
				(*command__)->getParameter (1).getUint8 (),
				(*command__)->getParameter (2).getUint16 (),
				(*command__)->getParameter (3).getUint16 (),
				(*command__)->getParameter (4).getUint16 (),
				(*command__)->getParameter (5).getUint8 ());
	break;
      case SherryCommand::DRAW_LINE:
	execDrawLine ((*command__)->getParameter (0).getUint16 (),
		      (*command__)->getParameter (1).getUint8 (),
		      (*command__)->getParameter (2).getUint16 (),
		      (*command__)->getParameter (3).getUint16 (),
		      (*command__)->getParameter (4).getUint16 (),
		      (*command__)->getParameter (5).getUint16 (),
		      (*command__)->getParameter (6).getUint8 ());
	break;
			  

      case SherryCommand::COPY_AREA:
	execCopyArea ((*command__)->getParameter (0).getUint16 (),
		      (*command__)->getParameter (1).getUint16 (),
		      (*command__)->getParameter (2).getUint8 (),
		      (*command__)->getParameter (3).getUint8 (),
		      (*command__)->getParameter (4).getUint16 (),
		      (*command__)->getParameter (5).getUint16 (),
		      (*command__)->getParameter (6).getUint16 (),
		      (*command__)->getParameter (7).getUint16 (),
		      (*command__)->getParameter (8).getUint16 (),
		      (*command__)->getParameter (9).getUint16 ());
	break;
      case SherryCommand::MASK_COPY:
	execMaskCopy ((*command__)->getParameter (0).getUint16 (),
		      (*command__)->getParameter (1).getUint16 (),
		      (*command__)->getParameter (2).getUint8 (),
		      (*command__)->getParameter (3).getUint8 (),
		      (*command__)->getParameter (4).getUint16 (),
		      (*command__)->getParameter (5).getUint16 (),
		      (*command__)->getParameter (6).getUint16 (),
		      (*command__)->getParameter (7).getUint16 (),
		      (*command__)->getParameter (8).getUint16 (),
		      (*command__)->getParameter (9).getUint16 (),
		      (*command__)->getParameter (10).getUint8 (),
		      (*command__)->getParameter (11).getUint8 (),
		      (*command__)->getParameter (12).getUint8Array (),
		      (*command__)->getParameter (12).getArraySize ());
	break;
      case SherryCommand::PLANE_COPY:
	execPlaneCopy ((*command__)->getParameter (0).getUint16 (),
		       (*command__)->getParameter (1).getUint16 (),
		       (*command__)->getParameter (2).getUint8 (),
		       (*command__)->getParameter (3).getUint8 (),
		       (*command__)->getParameter (4).getUint8 (),
		       (*command__)->getParameter (5).getUint8 (),
		       (*command__)->getParameter (6).getUint8 (),
		       (*command__)->getParameter (7).getUint16 (),
		       (*command__)->getParameter (8).getUint16 (),
		       (*command__)->getParameter (9).getUint16 (),
		       (*command__)->getParameter (10).getUint16 (),
		       (*command__)->getParameter (11).getUint16 (),
		       (*command__)->getParameter (12).getUint16 ());
	break;
			  
      case SherryCommand::SOURCE_FILE_POSITION:
	execSourceFilePosition ((*command__)->getParameter (0).getUint16 (),
				(*command__)->getParameter (1).getUint16 ());
	break;
      case SherryCommand::DEBUGGER_PRIVATE:
	execDebuggerPrivate ((*command__)->getParameter (0).getCharArray (),
			     (*command__)->getParameter (0).getArraySize ());
	break;

      default:
	DEBUG_MESSAGE ("ERROR::SherryWrd::execCommand: default");
	break;
      }

    ++command__;
  } while (command__ != sherryWrdFile->getCommandsList ().end () &&
	   (*command__)->getDeltaTime () == 0);
}


void SherryWrd::execEndOfData ()
{
  //do nothing
}

void SherryWrd::execStartPlaying ()
{
  //do nothing
}

			  
void SherryWrd::execReservedCommand ()
{
  DEBUG_MESSAGE ("SherryWrd::execReservedCommand called.Is there error?");
}

			  
void SherryWrd::execNewPalette (const uint16* palettes, int arraySize)
{
  assert (palettes != NULL);
  if (arraySize <= 0 || arraySize > 65536)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execNewPalette palettes=";
#endif
  for (int i = 0; i < arraySize; i++)
    {
      if (palettes[i] >= MAX_VIRTUAL_PALETTES)
	throw OutOfRange ();
      if (virtualPalette[palettes[i]] != NULL)
	throw AlreadyAllocated ();

#ifdef SHERRY_WRD_DEBUG
      cerr << palettes[i] << " ";
#endif
      virtualPalette[palettes[i]] = new SherryPalette;
      //XXX bad_alloc$B$r(Bcatch$B$9$k$3$H(B
      if (virtualPalette[palettes[i]] == NULL)
	throw CanNotAllocate ();
    }

#ifdef SHERRY_WRD_DEBUG
  cerr << endl;
#endif
}

void SherryWrd::execDeletePalette (const uint16* palettes, int arraySize)
{
  assert (palettes != NULL);
  if (arraySize <= 0 || arraySize > 65536)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDeletePalette aplettes=";
#endif
  for (int i = 0; i < arraySize; i++)
    {
      if (palettes[i] >= MAX_VIRTUAL_PALETTES)
	throw OutOfRange ();
      if (virtualPalette[palettes[i]] == NULL)
	throw NotAllocated ();

      //???needs this???
      for (list<uint16>::const_iterator j = transferPalettesList.begin ();
	   j != transferPalettesList.end ();
	   j++)
	{
	  //XXX parameter check
	  if (*j == palettes[i])
	    {
	      realTransferPalette (*j);
#ifdef SHERRY_WRD_DEBUG
	      cerr << "SherryWrd::execDeletePalette " << *j
		   << "-th palette deleted. but transfered." << endl;
#endif
	    }
	}
#ifdef SHERRY_WRD_DEBUG
      cerr << palettes[i] << " ";
#endif
      delete virtualPalette[palettes[i]];
      virtualPalette[palettes[i]] = NULL;
    }

#ifdef SHERRY_WRD_DEBUG
  cerr << endl;
#endif
  
}

void SherryWrd::execNewScreen (const SherryParameter::ScreenParameter* screens,
			       int arraySize)
{
  assert (screens != NULL);
  if (arraySize <= 0 || arraySize > 65536)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execNewScreen screens:";
#endif
  for (int i = 0; i < arraySize; i++)
    {

      //parameter check
      if (screens[i].screenNumber >= MAX_VIRTUAL_SCREENS)
	throw OutOfRange ();
      if (screens[i].sizeX > MAX_VIRTUAL_SCREEN_WIDTH)
	throw OutOfRange ();
      if (screens[i].sizeY > MAX_VIRTUAL_SCREEN_HEIGHT)
	throw OutOfRange ();
      if (virtualScreenInfo[screens[i].screenNumber] != NULL)
	throw AlreadyAllocated ();

#ifdef SHERRY_WRD_DEBUG
      cerr << " screenNumber=" << screens[i].screenNumber
	   << " wdith=" << screens[i].sizeX
	   << " height=" << screens[i].sizeY
	   << " transparancyColor="
	   << static_cast<int> (screens[i].transparentColor);
#endif
      virtualScreenInfo[screens[i].screenNumber] =
	new SherryParameter::ScreenParameter;
      // XXX bad_alloc$B$r(Bcatch$B$9$k$3$H(B
      if (virtualScreenInfo[screens[i].screenNumber] == NULL)
	throw CanNotAllocate ();
      virtualScreenInfo[screens[i].screenNumber]->screenNumber = screens[i].screenNumber;
      virtualScreenInfo[screens[i].screenNumber]->sizeX = screens[i].sizeX;
      virtualScreenInfo[screens[i].screenNumber]->sizeY = screens[i].sizeY;
      virtualScreenInfo[screens[i].screenNumber]->transparentColor = screens[i].transparentColor;
    }
  
#ifdef SHERRY_WRD_DEBUG
  cerr << endl;
#endif
  doNewScreen (screens, arraySize);
}

void SherryWrd::execDeleteScreen (const uint16* screens, int arraySize)
{
  assert (screens != NULL);
  if (arraySize <= 0 || arraySize > 65536)
    throw OutOfRange ();

  for (int i = 0; i < arraySize; i++)
    {
      if (virtualScreenInfo[i] == NULL)
	throw NotAllocated ();

      //??? needs this???
      for (list<ScreenArea>::const_iterator j = transferScreensList.begin ();
	   j != transferScreensList.end ();
	   j++)
	{
	  //XXX parameter check
	  //??? $B$3$l$G$ONI$/$J$$!#E>Aw=g=x$K0MB8$9$k$+$i!#(B???
	  //$B$I$&$$$&=hM}$r$9$k$N$,@5$7$$(B???
	  if ((*j).screen == screens[i])
	    {
	      realTransferScreen ((*j).screen,
				  (*j).fromLeft, (*j).fromTop,
				  (*j).toLeft, (*j).toTop,
				  (*j).width, (*j).height);
#ifdef SHERRY_WRD_DEBUG
	      cerr << "SherryWrd::execDeleteScreen " << (*j).screen
		   << "-th screene deleted. but transfered." << endl;
#endif
	    }
	}
    }
  
  doDeleteScreen (screens, arraySize);
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDeleteScreen screens";
#endif
  for (int i = 0; i < arraySize; i++)
    {
#ifdef SHERRY_WRD_DEBUG
      cerr << screens[i] << " ";
#endif
      delete virtualScreenInfo[screens[i]];
      virtualScreenInfo[screens[i]] = NULL;
    }
#ifdef SHERRY_WRD_DEBUG
  cerr << endl;
#endif
}


void SherryWrd::execNewScreenAndLoadPng (uint16 screen, uint16 palette,
					 const char* pngFileName,
					 int arraySize)
{
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execNewScreenAndLoadPng screen=" << screen
       << " palette=" << palette << " filename=";
  cerr.write (pngFileName, arraySize) << endl;
#endif
  
  assert (pngFileName != NULL);
  assert (arraySize > 0);
  if (screen >= MAX_VIRTUAL_SCREENS || palette >= MAX_VIRTUAL_PALETTES)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] != NULL)
    {
#ifdef SHERRY_WRD_DEBUG
      cerr << "SherryWrd::execNewScreenAndLoadPng screen=" << screen
	   << " is already allocated" << endl;
#endif
      throw AlreadyAllocated ();
    }


  /////////////////////////////////////////////png file$B$N(Bload
  string fullPngFileName =
    songParentDirectory + string (pngFileName, arraySize);
  PNG* png = NULL;
  //XXX$B8+$D$+$i$J$+$C$?>l9g3HD%;R!J(B.png)$B$,BgJ8;z!">.J8;z$HJQ$($F(B
  //$BC5$7D>$9!#!JCf$K$O%U%!%$%kL>$GBgJ8;z!">.J8;z$,0c$C$F$$$k>l9g$b9M$($k$Y$-(B?$B!K(B
  try
    {
      png = new PNG(fullPngFileName);
    }
  catch (PNG::PNGError&)
    {
#ifdef DEBUG
      cerr << "SherryWrd::execNewScreenAdnLoadPng can not find png file "
	   << fullPngFileName << ". trying .png suffix" << endl;
#endif
      setSuffix (fullPngFileName, "png");
      try
	{
	  png = new PNG(fullPngFileName);
	}
      catch (PNG::PNGError&)
	{
#ifdef DEBUG
	  cerr << "SherryWrd::execNewScreenAdnLoadPng can not find png file "
	       << pngFileName << ". trying .PNG suffix" << endl;
#endif
	  setSuffix (fullPngFileName, "PNG");
	  try
	    {
	      png = new PNG(fullPngFileName);
	    }
	  catch (PNG::PNGError&)
	    {
	      cerr << "SherryWrd::execNewScreenAndLoadPng "
		   << "can not find png file." << endl;
	      throw CanNotLoadPNG ();
	    }
	}
    }

  //depth channel$BEy$b%A%'%C%/$9$Y$-(B?
  if (png->getColorType () != 3)
    {
      DEBUG_MESSAGE ("SherryWrd::execNewScreenAndLoadPng png is not supported format");
      delete png;
      throw UnsupportedPNGFormat ();
    }

  
  /////////////////////////////////////png$B$N(Bdata$B$r(Bvirtual screen$B$K:n@.$9$k!#(B
  virtualScreenInfo[screen] = new SherryParameter::ScreenParameter;
  //XXX catch bac_alloc$B$9$Y$-!#(B
  if (virtualScreenInfo[screen] == NULL)
    {
      DEBUG_MESSAGE ("SherryWrd::execNewScreenAndLoadPng can not allocate virtual screen");
      delete png;
      throw CanNotAllocate ();
    }
  virtualScreenInfo[screen]->screenNumber = screen;
  virtualScreenInfo[screen]->sizeX = png->getWidth ();
  assert (virtualScreenInfo[screen]->sizeX > 0);
  assert (virtualScreenInfo[screen]->sizeX <= MAX_VIRTUAL_SCREEN_WIDTH);
  virtualScreenInfo[screen]->sizeY = png->getHeight ();
  assert (virtualScreenInfo[screen]->sizeY > 0);
  assert (virtualScreenInfo[screen]->sizeY <= MAX_VIRTUAL_SCREEN_HEIGHT);
  if (png->getValid (PNG_INFO_tRNS))
    virtualScreenInfo[screen]->transparentColor = png->getTrans ()[0];
  else
    virtualScreenInfo[screen]->transparentColor = 0;


  ///////////////////////////////////////////////////////palette$B$NFI$_9~$_!#(B
  if (png->getValid (PNG_INFO_PLTE))
    {
      const int numPalettes = png->getNumPalette ();
      if (numPalettes > 256)
	{
#ifdef DEBUG
	  cerr << "SherryWrd::execNewScreenAndLoadPng " << fullPngFileName
	       << " has " << numPalettes << " colors" << endl;
#endif
	  delete png;
	  throw UnsupportedPNGFormat ();
	}


      //??? palette$B$O3NJ]$9$Y$-(B?$B$=$l$H$b3NJ]$5$l$F$$$k$Y$-(B?
      if (virtualPalette[palette] == NULL)
	{
	  virtualPalette[palette] = new SherryPalette;
	  //XXX catch bac_alloc$B$9$Y$-!#(B
	  if (virtualPalette[palette] == NULL)
	    {
	      DEBUG_MESSAGE ("SherryWrd::execNewScreenAndLoadPng can not allocate virtual palette");
	      delete png;
	      throw CanNotAllocate ();
	    }
	}
      else
	{//virtualPalette[palette] != NULL
#ifdef DEBUG
	  cerr << "SherryWrd::execNewScreenAndLoadPng palette=" << palette
	       << " is already allocated" << endl;
#endif
	  //throw AlreadyAllocated ();//???
	}
      
      for (int i = 0; i < numPalettes; i++)
	{
	  virtualPalette[palette]->set (i,
					png->getPalette ()[i].red,
					png->getPalette ()[i].green,
					png->getPalette ()[i].blue);
#ifdef SHERRY_WRD_DEBUG
	  cerr << "SherryWrd::execNewScreenAndLoadPng "
	       << " png.r=" << static_cast<int> (png->getPalette ()[i].red)
	       << " png.g=" << static_cast<int> (png->getPalette ()[i].green)
	       << " png.b=" << static_cast<int> (png->getPalette ()[i].blue)
	       << ":"
	       << " r=" << static_cast<int> (virtualPalette[palette]->getEntry (i).getR ())
	       << " g=" << static_cast<int> (virtualPalette[palette]->getEntry (i).getG ())
	       << " b=" << static_cast<int> (virtualPalette[palette]->getEntry (i).getB ())
	       << endl;
#endif
	}
    }
  else
    {
#ifdef DEBUG
      cerr << "SherryWrd::execNewScreenAndLoadPng " << fullPngFileName
	   << " has no palette" << endl;
#endif
      delete png;
      throw UnsupportedPNGFormat ();
    }
  
  doNewScreenAndLoadPng (screen, palette, png);
  
  delete png;
  return;
}

			  
void SherryWrd::execTransferPalette (uint16 palette)
{
  //parameter check???
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execTransferPalette palette=" << palette << endl;
#endif
  transferPalettesList.push_back (palette);
}

void SherryWrd::execTransferFullScreen (uint16 screen, uint16 left, uint16 top)
{
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execTransferFullScreen"
       << " screen=" << screen
       << " left=" << left
       << " top=" << top
       << endl;
#endif
  //parameter check???
  transferScreensList.push_back (ScreenArea (screen, left, top,
					     0, 0, 640, 480));
}

void SherryWrd::execTransferPartialScreen (uint16 screen,
					   uint16 fromLeft, uint16 fromTop,
					   uint16 toLeft, uint16 toTop,
					   uint16 width, uint16 height)
{
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execTransferPartialScreen"
       << " screen=" << screen
       << " fromLeft=" << fromLeft
       << " fromTop=" << fromTop
       << " toLeft=" << toLeft
       << " toTop=" << toTop
       << " width=" << width
       << " height=" << height
       << endl;
#endif
  //parameter check???
  transferScreensList.push_back (ScreenArea (screen,
					     fromLeft, fromTop,
					     toLeft, toTop, width, height));
}

			  
void SherryWrd::execSetPalette (uint16 palette,
				const SherryParameter::PaletteParameter* rgbs,
				int arraySize)
{
  if (virtualPalette[palette] == NULL)
    throw NotAllocated ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execSetPalette palette=" << palette << " ";
#endif
  for (int i = 0; i < arraySize; i++)
    {
#ifdef SHERRY_WRD_DEBUG
      cerr << static_cast<int> (rgbs[i].paletteNumber) << "-th "
	   << "r=" << static_cast<int> (rgbs[i].red) << " "
	   << "g=" << static_cast<int> (rgbs[i].green) << " "
	   << "b=" << static_cast<int> (rgbs[i].blue) << " ";
#endif
	virtualPalette[palette]->set (rgbs[i].paletteNumber,
				      rgbs[i].red, rgbs[i].green, rgbs[i].blue);
    }
#ifdef SHERRY_WRD_DEBUG
  cerr << endl;
#endif
}

void SherryWrd::execMergePalette (uint16 fromPalette1, uint16 fromPalette2,
				  uint16 toPalette,
				  uint8 palette1In, uint8 palette1Bit,
				  uint8 palette1Out,
				  uint8 palette2In, uint8 palette2Bit,
				  uint8 palette2Out,
				  uint8 transparency1, uint8 transparency2)
{
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execMergePalette"
       << " formPalette1=" << fromPalette1
       << " fromPalette2=" << fromPalette2
       << " toPalette=" << toPalette
       << " palette1In=" << static_cast<uint> (palette1In)
       << " palette1Bit=" << static_cast<uint> (palette1Bit)
       << " palette1Out=" << static_cast<uint> (palette1Out)
       << " palette2In=" << static_cast<uint> (palette2In)
       << " palette2Bit=" << static_cast<uint> (palette2Bit)
       << " palette2Out=" << static_cast<uint> (palette2Out)
       << " transparency1=" << static_cast<uint> (transparency1)
       << " transparency2=" << static_cast<uint> (transparency2)
       << endl;
#endif
  if (fromPalette1 >= MAX_VIRTUAL_PALETTES ||
      fromPalette2 >= MAX_VIRTUAL_PALETTES ||
      toPalette >= MAX_VIRTUAL_PALETTES)
    throw OutOfRange ();
  if (virtualPalette[fromPalette1] == NULL ||
      virtualPalette[fromPalette2] == NULL ||
      virtualPalette[toPalette] == NULL)
    throw NotAllocated ();

#if 0
  const uint32 pal1Mask = (0x11111111 >> (8-palette1Bit)) << palette1Out;
  const uint32 pal2Mask = (0x11111111 >> (8-palette2Bit)) << palette2Out;
#else
  const uint32 pal1Mask = (0xff >> (8-palette1Bit)) << palette1Out;
  const uint32 pal2Mask = (0xff >> (8-palette2Bit)) << palette2Out;
#endif
  const uint32 palMask = pal1Mask | pal2Mask;

  for (uint i = 0; i < MAX_COLORS; i++)
    {
      if ( (i & palMask) == i)
	{
	  const int palette1Entry = i & (pal1Mask >> palette1In);
	  const int palette2Entry = i & (pal1Mask >> palette2In);
	  const int red =
	    (virtualPalette[fromPalette1]->getEntry (palette1Entry).getR ()
	     *static_cast<int> (transparency1)
	     +virtualPalette[fromPalette2]->getEntry (palette2Entry).getR ()
	     *static_cast<int> (transparency2))/100;
	  const int green =
	    (virtualPalette[fromPalette1]->getEntry (palette1Entry).getG ()
	     *static_cast<int> (transparency1)
	     +virtualPalette[fromPalette2]->getEntry (palette2Entry).getG ()
	     *static_cast<int> (transparency2))/100;
	  const int blue =
	    (virtualPalette[fromPalette1]->getEntry (palette1Entry).getB ()
	     *static_cast<int> (transparency1)
	     +virtualPalette[fromPalette2]->getEntry (palette2Entry).getB ()
	     *static_cast<int> (transparency2))/100;
	  virtualPalette[toPalette]->set (i,
					  (red   > 0xff)? 0xff: red,
					  (green > 0xff)? 0xff: green,
					  (blue  > 0xff)? 0xff: blue);
	}
    }
}

void SherryWrd::execCopyPalette (uint16 fromPalette, uint16 toPalette)
{
  if (fromPalette >= MAX_VIRTUAL_PALETTES || toPalette >= MAX_VIRTUAL_PALETTES)
    throw OutOfRange ();
  if (virtualPalette[fromPalette] == NULL || virtualPalette[toPalette] == NULL)
    throw NotAllocated ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execCopyPalette"
       << " fromPalette=" << fromPalette
       << " toPlatte=" << toPalette
       << endl;
#endif
  virtualPalette[toPalette]->copy (virtualPalette[fromPalette]);
}

void SherryWrd::execMaskCopyPalette (uint16 fromPalette, uint16 toPalette,
				     const uint8* paletteCopyMask)
{
  assert (paletteCopyMask != NULL);
  if (fromPalette >= MAX_VIRTUAL_PALETTES || toPalette >= MAX_VIRTUAL_PALETTES)
    throw OutOfRange ();
  if (virtualPalette[fromPalette] == NULL || virtualPalette[toPalette] == NULL)
    throw NotAllocated ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execMaskCopyPalette"
       << " fromPalette=" << fromPalette
       << " toPalette=" << toPalette
       << endl;
#endif
  
  for (int i = 0; i < 32; i++)
    {
      for (int j = 0; j < 8; j++)
	{
	  if (paletteCopyMask[i] & (1 << j))
	    {
	      virtualPalette[toPalette]->copyEntry (i*8+j,
						    virtualPalette[fromPalette]->getEntry (i*8+j));
	    }
	}
    }
}

void SherryWrd::execPartialCopyPalette (uint16 fromPalette, uint16 toPalette,
					const SherryParameter::PaletteNumberParameter* paletteNumbers, int arraySize)
{
  assert (paletteNumbers != NULL);
  assert (arraySize > 0);
  if (fromPalette >= MAX_VIRTUAL_PALETTES || toPalette >= MAX_VIRTUAL_PALETTES)
    throw OutOfRange ();
  if (virtualPalette[fromPalette] == NULL || virtualPalette[toPalette] == NULL)
    throw NotAllocated ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execPartialCopyPalette"
       << " fromPalette=" << fromPalette
       << " toPalette=" << toPalette
       << endl;
#endif
  for (int i = 0; i < arraySize; i++)
    {
      virtualPalette[toPalette]->copyEntry (paletteNumbers[i].toPalette,
					    virtualPalette[fromPalette]->getEntry (paletteNumbers[i].fromPalette));
    }
}

			  
void SherryWrd::execDrawText (uint16 screen,
			      uint8 planeMask, uint8 mode,
			      uint8 foreground, uint8 background,
			      uint16 left, uint16 top,
			      const char* sjisStr, int arraySize)
{
  assert (mode == 0 || mode == 1 || mode == 2 || mode == 3);
  assert (sjisStr != NULL);
  //assert (arraySize > 0);
  if (screen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] == NULL)
    throw NotAllocated ();
  if (left > virtualScreenInfo[screen]->sizeX ||
      top > virtualScreenInfo[screen]->sizeY)
    throw OutOfRange ();

  if (arraySize == 0)
    {
#ifdef SHERRY_WRD_DEBUG
      cerr << "Warning:SherryWrd::execDrawText arraySize==0" << endl;
#endif
      return;
    }
  
  char* eucStr = new char[arraySize*2];
  const int eucLength = convertSjis2Euc (sjisStr, arraySize, eucStr);
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDrawText"
       << " screen=" << screen
       << " planeMask=" << static_cast<int> (planeMask)
       << " mode=" << static_cast<int> (mode)
       << " foreground=" << static_cast<int> (foreground)
       << " background=" << static_cast<int> (background)
       << " left=" << left
       << " top=" << top
       << " str=\"" ;
  cerr.write (eucStr, eucLength) << "\""<< endl;
#endif

  doDrawText (screen,
	      planeMask, mode,
	      foreground, background,
	      left, top,
	      eucStr, eucLength);
  delete[] eucStr;
}

void SherryWrd::execDrawBox (uint16 screen, uint8 planeMask,
			     uint16 x1, uint16 y1,
			     uint16 x2, uint16 y2,
			     uint8 foreground)
{
  if (screen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[screen]->sizeX ||
      y1 > virtualScreenInfo[screen]->sizeY ||
      x2 > virtualScreenInfo[screen]->sizeX ||
      y2 > virtualScreenInfo[screen]->sizeY)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDrawBox"
       << " screen=" << screen
       << " planeMask=" << static_cast<int> (planeMask)
       << " x1=" << x1
       << " y1=" << y1
       << " x2=" << x2
       << " y2=" << y2
       << " foreground=" << static_cast<int> (foreground)
       << endl;
#endif
  doDrawBox (screen, planeMask, x1, y1, x2, y2, foreground);
}

void SherryWrd::execDrawVirtialLine (uint16 screen, uint8 planeMask,
				     uint16 x1, uint16 y1, uint16 y2,
				     uint8 foreground)
{
  if (screen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[screen]->sizeX ||
      y1 > virtualScreenInfo[screen]->sizeY ||
      y2 > virtualScreenInfo[screen]->sizeY)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDrawVirtialLine"
       << " screen=" << screen
       << " planeMask=" << static_cast<int> (planeMask)
       << " x1=" << x1
       << " y1=" << y1
       << " y2=" << y2
       << " foreground=" << static_cast<int> (foreground)
       << endl;
#endif
  
  doDrawLine (screen, planeMask,
	      x1, y1,
	      x1, y2,
	      foreground);
}

void SherryWrd::execDrawHorizontalLine (uint16 screen, uint8 planeMask,
					uint16 x1, uint16 x2, uint16 y1,
					uint8 foreground)
{
  if (screen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[screen]->sizeX ||
      x2 > virtualScreenInfo[screen]->sizeX ||
      y1 > virtualScreenInfo[screen]->sizeY)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDrawHorizontalLine"
       << " screen=" << screen
       << " planeMask=" << static_cast<int> (planeMask)
       << " x1=" << x1
       << " x2=" << x2
       << " y1=" << y1
       << " foreground=" << static_cast<int> (foreground)
       << endl;
#endif
  
  doDrawLine (screen, planeMask,
	      x1, y1,
	      x2, y1,
	      foreground);
}

void SherryWrd::execDrawLine (uint16 screen, uint8 planeMask,
			      uint16 x1, uint16 y1,
			      uint16 x2, uint16 y2,
			      uint8 foreground)
{
  if (screen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[screen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[screen]->sizeX ||
      y1 > virtualScreenInfo[screen]->sizeY ||
      x2 > virtualScreenInfo[screen]->sizeX ||
      y2 > virtualScreenInfo[screen]->sizeY)
    throw OutOfRange ();
  
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execDrawLine"
       << " screen=" << screen
       << " planeMask=" << static_cast<int> (planeMask)
       << " x1=" << x1
       << " x2=" << x2
       << " y1=" << y1
       << " y2=" << y2
       << " foreground=" << static_cast<int> (foreground)
       << endl;
#endif
  doDrawLine (screen, planeMask,
	      x1, y1,
	      x2, y2,
	      foreground);
}

			  
void SherryWrd::execCopyArea (uint16 fromScreen, uint16 toScreen,
			      uint8 planeMask, uint8 transparencyFlag,
			      uint16 x1, uint16 y1,
			      uint16 x2, uint16 y2,
			      uint16 toX, uint16 toY)
{
  if (fromScreen >= MAX_VIRTUAL_SCREENS || toScreen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[fromScreen] == NULL ||
      virtualScreenInfo[toScreen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[fromScreen]->sizeX ||
      y1 > virtualScreenInfo[fromScreen]->sizeY ||
      x2 > virtualScreenInfo[fromScreen]->sizeX ||
      y2 > virtualScreenInfo[fromScreen]->sizeY ||
      toX > virtualScreenInfo[toScreen]->sizeX ||
      toY > virtualScreenInfo[toScreen]->sizeY ||
      toX+abs (x1-x2) > virtualScreenInfo[toScreen]->sizeX ||
      toY+abs (y1-y2) > virtualScreenInfo[toScreen]->sizeY)
    throw OutOfRange ();

#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execCopyArea"
       << " fromScreen=" << fromScreen
       << " toScreen=" << static_cast<int> (toScreen)
       << " planeMask=" << static_cast<int> (planeMask)
       << " transparencyFlag=" << static_cast<int> (transparencyFlag)
       << " x1=" << x1
       << " y1=" << y1
       << " x2=" << x2
       << " y2=" << y2
       << " toX=" << toX
       << " toY=" << toY
       << endl;
#endif
  doCopyArea (fromScreen, toScreen,
	      planeMask, transparencyFlag,
	      x1, y1,
	      x2, y2,
	      toX, toY);
}

void SherryWrd::execMaskCopy (uint16 fromScreen, uint16 toScreen,
			      uint8 planeMask, uint8 transparencyFlag,
			      uint16 x1, uint16 y1,
			      uint16 x2, uint16 y2,
			      uint16 toX, uint16 toY,
			      uint8 maskX, uint8 maskY,
			      const uint8* maskData, int arraySize)
{
  assert (maskData != NULL);
  assert (arraySize > 0);
  //assert (arraySize == ...);//XXX
  if (fromScreen >= MAX_VIRTUAL_SCREENS || toScreen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[fromScreen] == NULL ||
      virtualScreenInfo[toScreen] == NULL)
    throw NotAllocated ();
  if (x1 > virtualScreenInfo[fromScreen]->sizeX ||
      y1 > virtualScreenInfo[fromScreen]->sizeY ||
      x2 > virtualScreenInfo[fromScreen]->sizeX ||
      y2 > virtualScreenInfo[fromScreen]->sizeY ||
      toX > virtualScreenInfo[toScreen]->sizeX ||
      toY > virtualScreenInfo[toScreen]->sizeY ||
      toX+abs (x1-x2) > virtualScreenInfo[toScreen]->sizeX ||
      toY+abs (y1-y2) > virtualScreenInfo[toScreen]->sizeY)
    throw OutOfRange ();


#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execMaskCopy"
       << " fromScreen=" << fromScreen
       << " toScreen=" << toScreen
       << " planeMask=" << static_cast<int> (planeMask)
       << " transparencyFlag=" << static_cast<int> (transparencyFlag)
       << " x1=" << x1
       << " y1=" << y1
       << " x2=" << x2
       << " y2=" << y2
       << " toX=" << toX
       << " toY=" << toY
       << " maskX" << static_cast<int> (maskX)
       << " maskY" << static_cast<int> (maskY)
       << endl;
#endif
  doMaskCopy (fromScreen, toScreen,
	      planeMask, transparencyFlag,
	      x1, y1,
	      x2, y2,
	      toX, toY,
	      maskX, maskY,
	      maskData, arraySize);
}

void SherryWrd::execPlaneCopy (uint16 fromScreen, uint16 toScreen,
			       uint8 planeMask, uint8 transparencyFlag,
			       uint8 fromPlane, uint8 toPlane,
			       uint8 planeNumber,
			       uint16 x1, uint16 y1,
			       uint16 x2, uint16 y2,
			       uint16 toX, uint16 toY)
{
  if (fromScreen >= MAX_VIRTUAL_SCREENS || toScreen >= MAX_VIRTUAL_SCREENS)
    throw OutOfRange ();
  if (virtualScreenInfo[fromScreen] == NULL ||
      virtualScreenInfo[toScreen] == NULL)
    throw NotAllocated ();
  if (fromPlane >= MAX_PLANES ||
      toPlane >= MAX_PLANES ||
      fromPlane + planeNumber >= MAX_PLANES ||
      toPlane + planeNumber >= MAX_PLANES ||
      x1 > virtualScreenInfo[fromScreen]->sizeX ||
      y1 > virtualScreenInfo[fromScreen]->sizeY ||
      x2 > virtualScreenInfo[fromScreen]->sizeX ||
      y2 > virtualScreenInfo[fromScreen]->sizeY ||
      toX > virtualScreenInfo[toScreen]->sizeX ||
      toY > virtualScreenInfo[toScreen]->sizeY ||
      toX+abs (x1-x2) > virtualScreenInfo[toScreen]->sizeX ||
      toY+abs (y1-y2) > virtualScreenInfo[toScreen]->sizeY)
    throw OutOfRange ();
  
#ifdef SHERRY_WRD_DEBUG
  cerr << "SherryWrd::execPlaneCopy"
       << " fromScreen=" << fromScreen
       << " toScreen=" << toScreen
       << " planeMask=" << static_cast<int> (planeMask)
       << " transparencyFlag=" << static_cast<int> (transparencyFlag)
       << " fromPlane=" << static_cast<int> (fromPlane)
       << " toPlane=" << static_cast<int> (toPlane)
       << " planeNumber=" << static_cast<int> (planeNumber)
       << " x1=" << x1
       << " y1=" << y1
       << " x2=" << x2
       << " y2=" << y2
       << " toX=" << toX
       << " toY=" << toY
       << endl;
#endif
  doPlaneCopy (fromScreen, toScreen,
	       planeMask, transparencyFlag,
	       fromPlane, toPlane,
	       planeNumber,
	       x1, y1,
	       x2, y2,
	       toX, toY);
}

			  
void SherryWrd::execSourceFilePosition (uint16 sourceFileNumber,
					uint16 sourceFileLineNumber)
{
  //do nothing
}

void SherryWrd::execDebuggerPrivate (const char* privateData, int arraySize)
{
  //do nothing
}
