#include "SherryWrdFile.h"
#include "SherryInputStream.h"

const int16 SherryWrdFile::TIME_BASE_IS_NOT_SET;

SherryWrdFile::SherryWrdFile (const string& filename__):
  filename (filename__), version (-1),timeBase (TIME_BASE_IS_NOT_SET)
{
  try
    {
      load (filename__);
    }
  catch (Error&)
    {
#ifdef DEBUG
      cerr << "SherryWrdFile::SherryWrdFile can not find the file="
	   << filename__ << endl;
#endif
      throw FileIsNotFound ();
    }
}

SherryWrdFile::~SherryWrdFile ()
{
  for (CompilerPrivateInfo::iterator i = compilerPrivateInfo.begin ();
       i != compilerPrivateInfo.end (); i++)
    delete *i;

  for (list<const SherryCommand*>::iterator i = commands.begin ();
       i != commands.end (); i++)
    delete *i;
}


void SherryWrdFile::load (const string& filename__)
{
  SherryInputStream ifs (filename__);
  if (!ifs)
    throw FileIsNotFound ();
  
  const int fileSize = ifs.getFileSize ();
  if (fileSize < 16+4+6)//file header + header block + data block
    throw FileIsTooShort();

  //read file header
  char magic[12];
  ifs.read (magic, 12);
  if (memcmp (magic, SherryMagic, 12) != 0)
    throw FileIsNotSherryWrd ();
  version = ifs.get32L ();
#ifdef DEBUG
  cerr << "SherryWrdFile::load version=" << version << endl;
#endif
  
  //read header block
  while (1)
    {
      const ulong length = ifs.getVariableLength ();
      if (length == 0)
	continue;

      uint8 command = ifs.get8 ();
      switch (command)
	{
	case 0x00:
	  assert (length == 1);
	  goto END_OF_HEADER_BLOCK;
	  break;
	case 0x10://falling through
	case 0x11://falling through
	case 0x12://falling through
	case 0x13://falling through
	case 0x14://falling through
	case 0x15://falling through
	case 0x16://falling through
	case 0x17://falling through
	case 0x18://falling through
	case 0x19:
	  cerr << "SherryWrdFile::load reserved packet data found.ignoring" << endl;
	  ifs.seekg (length-1);
	  break;
	case 0x20:
	  //XXX SECL bug???
	  assert (length == 3);
	  timeBase = ifs.get16L ();
	  break;
	case 0x21:
	  assert (length == 3);
	  timeBase = ifs.get16L ();
	  break;
	case 0x71:
	  {
	    //assert (compilerName.size () == 0);//???
	    char* tmp = new char[length];
	    ifs.read (tmp, length-1);
	    tmp[length-1] = '\0';
	    compilerName.push_back (tmp);
#ifdef DEBUG
	    cerr << "SherryWrdFile::load compiler=";
	    cerr.write (tmp, length-1) <<endl;
#endif
	    delete[] tmp;
	    break;
	  }
	case 0x72:
	  {
	    const uint16 fileNumber = ifs.get16L ();
	    char* tmp = new char[length-2];
	    ifs.read (tmp, length-2-1);
	    tmp[length-2-1] = '\0';
	    sourceFiles.push_back (SourceFileItem (fileNumber, tmp));
#ifdef DEBUG
	    cerr << "SherryWrdFile::load srouceFileName=";
	    cerr.write (tmp, length-2-1) << endl;
#endif
	    delete[] tmp;
	    break;
	  }
	  break;
	case 0x73:
	  {
	    assert (compilerName.size () == 0);
	    char* tmp = new char[length-1];
	    ifs.read (tmp, length-1);
	    compilerPrivateInfo.push_back (new CompilerPrivateInfoItem (tmp, length-1));
	    delete[] tmp;
	    break;
	  }
	default:
#ifdef DEBUG
	  cerr << "SherryWrdFile::load read header block. found unknown packet data = " << hex << static_cast<int> (command) << dec << endl;
	  
#endif
	  ifs.seekg (length-1, ios::cur);
	  break;
	}
    }
 END_OF_HEADER_BLOCK:
  if (timeBase == TIME_BASE_IS_NOT_SET)
    throw TimeBaseIsNotSet ();

  //read data block
  readDataBlock (ifs);
}

void SherryWrdFile::readDataBlock (SherryInputStream& ifs)
{
  ulong totalDeltaTime = 0;

  while (1)
    {
      const ulong deltaTime = ifs.getVariableLength ();
      const ulong slength = ifs.getVariableLength ();
      totalDeltaTime +=  deltaTime;
      if (slength == 0)
	continue;

      const uint8 commandType = ifs.get8 ();
      ulong nowSLength = 1;//$B$^$:(B,command$B$N(B1byte$B$rFI$s$@!#(B
      if (slength == 1)
	{
#ifdef DEBUG
	  cerr << "SherryWrdFile::readDataBlock slength=1" << endl;
#endif
	}
      

#ifdef DEBUG
      //cerr << " command = " << hex << static_cast<int>(commandType & 0x7f) << dec << " : slength = " << slength << " : tellg = " << hex << ifs.tellg ()-1 << dec <<endl;
#endif
      //parameter$B$N8D?t$r?t$($k!#(B
      int nrParameter = 0;
      for (; SherryCommand::getParameterType (commandType, nrParameter) != SherryParameter::END_OF_PARAMETER; nrParameter++)
	{
	  if (SherryCommand::getParameterType (commandType, nrParameter) == SherryParameter::INVALID)
	    {
#ifdef DEBUG
	      //XXX error$B=hM}(B
	      DEBUG_MESSAGE ("SherryWrdFile::readDataBlock invalid command found");
#endif
	      ifs.seekg (slength-nowSLength);
	      continue;
	    }
	}
      if (commandType == SherryCommand::START_PLAYING)
	totalDeltaTime = 0;
      
      SherryCommand* const sherryCommand = new SherryCommand (deltaTime,
							      totalDeltaTime,
							      commandType,
							      nrParameter);

      //$B<B:]$K%Q%i%a!<%?!<$r3P$($F$*$/!#(B
      for (int i = 0; i < nrParameter; i++)
	{
	  switch (SherryCommand::getParameterType (commandType, i))
	    {
	    case SherryParameter::CHAR:
	      {
		nowSLength += 1;
		const char charParameter = ifs.get8 ();
		sherryCommand->appendParameter (charParameter);
		break;
	      }
	    case SherryParameter::UINT8:
	      {
		nowSLength += 1;
		const uint8 uint8Parameter = ifs.get8 ();
		sherryCommand->appendParameter (uint8Parameter);
		break;
	      }
	    case SherryParameter::UINT16:
	      {
		nowSLength += 2;
		const uint16 uint16Parameter = ifs.get16L ();
		sherryCommand->appendParameter (uint16Parameter);
		break;
	      }
	    case SherryParameter::UINT32:
	      {
		
		nowSLength += 4;
		const uint32 uint32Parameter = ifs.get32L ();
		sherryCommand->appendParameter (uint32Parameter);
		break;
	      }
	    
	    case SherryParameter::CHAR_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		const ulong sizeOfArray = slength - nowSLength;
		nowSLength += sizeOfArray;
		char* const data = new char[sizeOfArray];
		ifs.read (data, sizeOfArray);
		sherryCommand->appendParameter (data, sizeOfArray);
		delete[] data;
		break;
	      }
	    case SherryParameter::UINT8_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		const ulong sizeOfArray = slength - nowSLength;
		nowSLength += sizeOfArray;
		uint8* const data = new uint8[sizeOfArray];
		ifs.read (data, sizeOfArray);
		sherryCommand->appendParameter (data, sizeOfArray);
		delete[] data;
		break;
	      }
	    case SherryParameter::UINT16_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		assert ((slength-nowSLength)%2 == 0);
		const ulong sizeOfArray = (slength - nowSLength)/2;
		nowSLength += slength - nowSLength;
		uint16* const data = new uint16[sizeOfArray];
		for (uint i = 0; i < sizeOfArray; i++)
		  data[i] = ifs.get16L ();
		sherryCommand->appendParameter (data, sizeOfArray);
		delete[] data;
		break;
	      }


	    case SherryParameter::SCREEN_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		assert ((slength-nowSLength)%7 == 0);
		const ulong sizeOfArray = (slength - nowSLength)/7;
		nowSLength += slength - nowSLength;
		SherryParameter::ScreenParameter* screenParameter =
		  new SherryParameter::ScreenParameter[sizeOfArray];
		for (uint i = 0; i < sizeOfArray; i++)
		  {
		    screenParameter[i].screenNumber = ifs.get16L ();
		    screenParameter[i].sizeX = ifs.get16L ();
		    screenParameter[i].sizeY = ifs.get16L ();
		    screenParameter[i].transparentColor = ifs.get8 ();
		  }
		sherryCommand->appendParameter (screenParameter, sizeOfArray);
		delete[] screenParameter;
		break;
	      }
  
	    case SherryParameter::PALETTE_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		assert ((slength-nowSLength)%4 == 0);
		const ulong sizeOfArray = (slength - nowSLength)/4;
		nowSLength += slength - nowSLength;
		SherryParameter::PaletteParameter* paletteParameter =
		  new SherryParameter::PaletteParameter[sizeOfArray];
		for (uint i = 0; i < sizeOfArray; i++)
		  {
		    paletteParameter[i].paletteNumber = ifs.get8 ();
		    paletteParameter[i].red = ifs.get8 ();
		    paletteParameter[i].green = ifs.get8 ();
		    paletteParameter[i].blue = ifs.get8 ();
		  }
		sherryCommand->appendParameter (paletteParameter, sizeOfArray);
		delete[] paletteParameter;
		break;
	      }

	    case SherryParameter::PALETTE_COPY_MASK:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		const ulong sizeOfArray = slength - nowSLength;
		assert (sizeOfArray == 32);
		nowSLength += sizeOfArray;
		uint8 mask[32];
		ifs.read (mask, 32);
		sherryCommand->appendParameter (mask);
		break;
	      }

	    case SherryParameter::PALETTE_NUMBER_ARRAY:
	      {
		assert (SherryCommand::getParameterType (commandType, i+1) ==
			SherryParameter::END_OF_PARAMETER);
		assert ((slength - nowSLength)%2 == 0);
		const ulong sizeOfArray = (slength - nowSLength)/2;
		nowSLength += slength - nowSLength;
		SherryParameter::PaletteNumberParameter* paletteNumberParameter = new SherryParameter::PaletteNumberParameter[sizeOfArray];
		for (uint i = 0; i < sizeOfArray; i++)
		  {
		    paletteNumberParameter[i].fromPalette = ifs.get8 ();
		    paletteNumberParameter[i].toPalette = ifs.get8 ();
		  }
		sherryCommand->appendParameter (paletteNumberParameter,
						sizeOfArray);
		delete[] paletteNumberParameter;
		break;
	      }
    
	    case SherryParameter::END_OF_PARAMETER:
	      if (slength != nowSLength)
		{
		  cerr << "ERROR::SherryWrdFile::readDataBlock"
		       << " END_OF_PARAMETER slength != nowLength" << endl;
		}
	      goto END_OF_PARAMETER;
	      break;
	    case SherryParameter::INVALID:
	      cerr << "ERROR::SherryWrdFile::readDataBlock INVALID" << endl;
	      //XXX
	      break;
	    }
	}
    END_OF_PARAMETER:

      commands.push_back (sherryCommand);
      if (commandType == SherryCommand::END_OF_DATA)
	break;//end of commnads
    }
}
