/*
    WtkButton.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <cassert>

#include "WtkButton.h"

using namespace std;

const char* const WtkButton::defaultResource=
"*.Foreground: Black\n"
"*.Background: Gray80\n"
"*.Light: Gray90\n"
"*.Mid: Gray80\n"
"*.Dark: Gray70\n"
"*.TextColor: Black\n"
"*.ActiveFontSet: -misc-fixed-*--14-*\n"
"*.ActiveForeground: Red\n"
"*.ActiveBackground: Gray90\n"
"*.ActiveLight: Gray100\n"
"*.ActiveMid: Gray90\n"
"*.ActiveDark: Gray80\n"
"*.ActiveTextColor: Black\n"
"*.BitmapX: 0\n"
"*.BitmapY: 0\n";


WtkButton::WtkButton (Display* display,const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_,
		      int depth,Visual* visual,Colormap colormap_,
		      char* resourceFileName):
  WtkWidget (display,className_,name_,x_,y_,width_,height_,
	     depth,visual,colormap_,resourceFileName),
  buttonState(BUTTON_UP), bitmap (NULL),bitmapX (0),bitmapY (0)
{
  //do needed
  
  initialize ();
}

WtkButton::WtkButton (Display* display,const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_,
		      int depth,Visual* visual,Colormap colormap_,
		      XrmDatabase displayRdb_):
  WtkWidget (display,className_,name_,x_,y_,width_,height_,
	     depth,visual,colormap_,displayRdb_),
  buttonState (BUTTON_UP), bitmap (NULL),bitmapX (0),bitmapY (0)
{
  //do needed
  
  initialize ();
}

WtkButton::WtkButton (WtkWidget* parent,const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_):
  WtkWidget (parent,className_,name_,x_,y_,width_,height_),
  buttonState (BUTTON_UP), bitmap (NULL),bitmapX (0),bitmapY (0)
{
  //do needed
  
  initialize ();
}

WtkButton::~WtkButton ()
{
  delete bitmap;
}

void WtkButton::initialize ()
{
  window->selectInput (ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask|ExposureMask);
  XrmCombineDatabase (XrmGetStringDatabase (defaultResource),&privateRdb,False);
  getResource ();
  setResource ();
}

void WtkButton::getResource ()
{
  activeStyle.setColormap (window->getDisplay (),style.getColormap ());
  activeStyle.allocColor (window->getDisplay (),FOREGROUND,
		    getResource ("activeForeground","ActiveForeground"));
  activeStyle.allocColor (window->getDisplay (),BACKGROUND,
		    getResource ("activeBackground","ActiveBackground"));
  activeStyle.allocColor (window->getDisplay (),LIGHT,
		    getResource ("activeLight","ActiveLight"));
  activeStyle.allocColor (window->getDisplay (),MID,
		    getResource ("activeMid","ActiveMid"));
  activeStyle.allocColor (window->getDisplay (),DARK,
		    getResource ("activeDark","ActiveDark"));
  activeStyle.allocColor (window->getDisplay (),TEXTCOLOR,
		    getResource ("activeTextColor","ActiveTextColor"));
  activeStyle.createFontSet (window->getDisplay (),getResource ("activeFontSet","ActiveFontSet"));

  delete bitmap;
  char* filename=getResource ("bitmap","Bitmap");
  if (filename!=NULL)
    bitmap=new WdkPixmap (getWindow (),filename);
  bitmapX=getResourceValue ("bitmapX","BitmapX");
  bitmapY=getResourceValue ("bitmapY","BitmapY");
  gc->setFillStyle (FillStippled);
}

void WtkButton::setBitmap (char* data,uint bitmapWidth,uint bitmapHeight)
{
  bitmap=new WdkPixmap (getWindow (),data,bitmapWidth,bitmapHeight);
}

void WtkButton::setBitmap (WdkPixmap& bitmap_)
{
  assert (bitmap_.getDepth ()==1);
  delete bitmap;
  bitmap=&bitmap_;
}

void WtkButton::setBitmap (char* filename)
{
  delete bitmap;
  bitmap=new WdkPixmap (getWindow (),filename);
}

void WtkButton::exposed (XEvent* /*event*/)
{
  decorate ();
}


void WtkButton::buttonPressed (XEvent* event)
{
  lockEventState ();
  eventState.isButtonPressed[event->xbutton.button]=true;
  buttonState=BUTTON_DOWN;
  unlockEventState ();
  decorate ();
};

void WtkButton::buttonReleased (XEvent* event)
{
  lockEventState ();
  if (eventState.isButtonPressed[event->xbutton.button])
    {
      eventState.isButtonPressed[event->xbutton.button]=false;
      eventState.isClicked[event->xbutton.button]=true;
    }
  else
    {
      eventState.isButtonReleased[event->xbutton.button]=true;
    }
  if (eventState.crossing==EVENT_CROSSING_ENTER)
    {
      //cerr << "Button released." << endl;
      clicked (event);
    }
  buttonState=BUTTON_UP;
  unlockEventState ();
  decorate ();
};

void WtkButton::enterNotify (XEvent* /*event*/)
{
  lockEventState ();
  eventState.crossing=EVENT_CROSSING_ENTER;
  unlockEventState ();
  window->setBackground (activeStyle.bg);
  //window->clear ();
  decorate ();
};

void WtkButton::leaveNotify (XEvent* /*event*/)
{
  lockEventState ();
  if (eventState.crossing==EVENT_CROSSING_ENTER)
    eventState.crossing=EVENT_CROSSING_NONE;
  else
    eventState.crossing=EVENT_CROSSING_LEAVE;
  window->setBackground (style.bg);
  //window->clear ();
  unlockEventState ();
  decorate ();
};

  
void WtkButton::drawBitmap ()
{
  int offsetX=bitmapX+style.leftOffset;
  int offsetY=bitmapY+style.topOffset;
  lockData ();
  if (buttonState==BUTTON_DOWN)
    {
      offsetX+=style.verticalShadowWidth*2;
      offsetY+=style.horizontalShadowWidth*2;
    }
  unlockData ();
  gc->setTSOrigin (offsetX,offsetY);
  if (bitmap!=NULL)
    {
      gc->setFillStyle (FillStippled);
      gc->setStipple (bitmap);
      gc->setForeground (style.fg);
      window->fillRectangle (gc,offsetX,offsetY,
			     bitmap->getWidth (),bitmap->getHeight ());
      gc->setFillStyle (FillSolid);
    }
}



void WtkButton::decorate ()
{
  //draw bitmap
  window->clear ();
  drawBitmap ();
  
  //set color
  WtkStyle* nowStyle=&style;
  if (eventState.crossing==EVENT_CROSSING_ENTER)
    nowStyle=&activeStyle;
  ShadowType shadowType=SHADOW_UP;
  for (int i=0;i<5;i++)
    {
      if (eventState.isButtonPressed[i])
	{
	  shadowType=SHADOW_DOWN;
	  break;
	}
    }
  WtkWidget::decorate (window,width,height,*nowStyle,shadowType);
}
