/*
    WtkRange.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <cassert>

#include "WtkRange.h"

using namespace std;

const char* const WtkRange::defaultResource=
"*.Foreground: Black\n"
"*.Background: Gray80\n"
"*.Light: Gray90\n"
"*.Mid: Gray80\n"
"*.Dark: Gray70\n"
"*.TextColor: Black\n"
"*.ActiveFontSet: -misc-fixed-*--14-*\n"
"*.ActiveForeground: Red\n"
"*.ActiveBackground: Gray90\n"
"*.ActiveLight: Gray100\n"
"*.ActiveMid: Gray90\n"
"*.ActiveDark: Gray80\n"
"*.ActiveTextColor: Black\n"
"*.Direction: Horizontal\n"
"*.Limit: 100.0\n"
"*.Thickness: 15\n"
"*.MinSliderSize: 10";


WtkRange::WtkRange (Display* display,const char* className_,const char* name_,
		    int x_,int y_,uint width_,uint height_,
		    int depth,Visual* visual,Colormap colormap_,
		    char* resourceFileName):
  WtkWidget (display,className_,name_,x_,y_,width_,height_,
	     depth,visual,colormap_,resourceFileName),
  isForwardActive (false),isForwardUp (true),
  isBackActive (false),isBackUp (true),isSliderActive (false),
  sliderMotionOffset (0),isSliderDragged (false)
{
  initialize ();
}

WtkRange::WtkRange (Display* display,const char* className_,const char* name_,
		    int x_,int y_,uint width_,uint height_,
		    int depth,Visual* visual,Colormap colormap_,
		    XrmDatabase displayRdb_):
  WtkWidget (display,className_,name_,x_,y_,width_,height_,
	     depth,visual,colormap_,displayRdb_),
  isForwardActive (false),isForwardUp (true),
  isBackActive (false),isBackUp (true),isSliderActive (false),
  sliderMotionOffset (0),isSliderDragged (false)
{
  initialize ();
}

WtkRange::WtkRange (WtkWidget* parent,const char* className_,const char* name_,
		    int x_,int y_,uint width_,uint height_):
  WtkWidget (parent,className_,name_,x_,y_,width_,height_),
  isForwardActive (false),isForwardUp (true),
  isBackActive (false),isBackUp (true),isSliderActive (false),
  sliderMotionOffset (0),isSliderDragged (false)
{
  initialize ();
}

WtkRange::~WtkRange ()
{
  delete forward;
  delete back;
  delete slider;
}

void WtkRange::initialize ()
{
  XrmCombineDatabase (XrmGetStringDatabase (defaultResource),&privateRdb,False);
  getResource ();
  setResource ();
}

Direction WtkRange::getResourceDirection
(char* resourceName,char* resourceClass)
{
  const char* result=getResource (resourceName,resourceClass);
  if (strcmp (result,"DIRECTION_UP")==0)
    return DIRECTION_UP;
  else if (strcmp (result,"DIRECTION_DOWN")==0)
    return DIRECTION_DOWN;
  else if (strcmp (result,"DIRECTION_RIGHT")==0)
    return DIRECTION_RIGHT;
  else if (strcmp (result,"DIRECTION_LEFT")==0)
    return DIRECTION_LEFT;
  else if (strcmp (result,"VERTICAL")==0)
    return VERTICAL;
  else 
    return HORIZONTAL;
}


void WtkRange::getResource ()
{
  activeStyle.setColormap (window->getDisplay (),style.getColormap ());
  activeStyle.allocColor (window->getDisplay (),FOREGROUND,
			  getResource ("activeForeground","ActiveForeground"));
  activeStyle.allocColor (window->getDisplay (),BACKGROUND,
			  getResource ("activeBackground","ActiveBackground"));
  activeStyle.allocColor (window->getDisplay (),LIGHT,
			  getResource ("activeLight","ActiveLight"));
  activeStyle.allocColor (window->getDisplay (),MID,
			  getResource ("activeMid","ActiveMid"));
  activeStyle.allocColor (window->getDisplay (),DARK,
			  getResource ("activeDark","ActiveDark"));
  activeStyle.allocColor (window->getDisplay (),TEXTCOLOR,
			  getResource ("activeTextColor","ActiveTextColor"));
  activeStyle.createFontSet (window->getDisplay (),
			     getResource ("activeFontSet","ActiveFontSet"));
  
  const uint min=max(style.horizontalShadowWidth,style.verticalShadowWidth)*3;
  style.shadowType=SHADOW_DOWN;
  direction=getResourceDirection ("direction","Direction");
  thickness=getResourceValue ("thickness","Thickness");
  if (thickness<=min) thickness=min;
  minSliderSize=getResourceValue ("minSliderSize","MinSliderSize");
  if (minSliderSize<min)minSliderSize=min;
  limit=getResourceValue ("limit","Limit");
  if (limit<=0) limit=100;//100$B$KFC$K:,5r$OL5$$!#(B
  position=0;
  scope=limit*90/100;//90%$B$H$$$&;v$KFC$K:,5r$O$J$$(B
  if (scope==0)scope=3;//3$B$H$$$&?t;z$K:,5rL5$7!#(B
  unitStep=scope/3;
}

void WtkRange::setResource ()
{

  //button$B$NBg$-$5$r7hDj!#(B
  switch (direction)
    {
    case VERTICAL://falling through
    case DIRECTION_DOWN://falling through
    case DIRECTION_UP:
      //$B=D(B
      forwardWidth=width;
      forwardHeight=thickness;
      backWidth=width;
      backHeight=thickness;
      break;
    case HORIZONTAL://falling through
    case DIRECTION_RIGHT://falling through
    case DIRECTION_LEFT:
      //$B2#8~$-(B
      forwardWidth=thickness;
      forwardHeight=height;
      backWidth=thickness;
      backHeight=height;
      break;
    }

  //slider$B$N0LCV5Z$SBg$-$5$H(Bbutton$B$N0LCV$r7hDj!#(B
  switch (direction)
    {
    case VERTICAL://falling through
    case DIRECTION_DOWN:
      forward=new WdkWindow (window,0,height-thickness*2,width,thickness);
      back=new WdkWindow (window,0,height-thickness,width,thickness);
      sliderX=style.verticalShadowWidth*2;
      sliderY=style.horizontalShadowWidth*2;
      sliderWidth=width-style.verticalShadowWidth*4;
      sliderHeight=max (height-thickness*2,minSliderSize);
      break;
    case DIRECTION_UP:
      forward=new WdkWindow (window,0,0,width,thickness);
      back=new WdkWindow (window,0,thickness,width,thickness);
      sliderX=style.verticalShadowWidth*2;
      sliderY=thickness*2+style.horizontalShadowWidth*2;
      sliderWidth=width-style.verticalShadowWidth*4;
      sliderHeight=max (height-thickness*2,minSliderSize);
      break;
    case HORIZONTAL://falling through
    case DIRECTION_RIGHT:
      forward=new WdkWindow (window,width-thickness*2,0,thickness,height);
      back=new WdkWindow (window,width-thickness,0,thickness,height);
      sliderX=style.verticalShadowWidth*2;
      sliderY=style.horizontalShadowWidth*2;
      sliderWidth=max (width-thickness*2,minSliderSize);
      sliderHeight=height-style.horizontalShadowWidth*4;
      break;
    case DIRECTION_LEFT:
      forward=new WdkWindow (window,0,0,thickness,height);
      back=new WdkWindow (window,thickness,0,thickness,height);
      sliderX=thickness*2+style.verticalShadowWidth*2;
      sliderY=style.horizontalShadowWidth*2;
      sliderWidth=max (width-thickness*2,minSliderSize);
      sliderHeight=height-style.horizontalShadowWidth*4;
      break;
    }
  /*
  cerr << "sliderX=" << sliderX << ":sliderY=" << sliderY
       << ":sliderWidth=" << sliderWidth << ":sliderHeight=" << sliderHeight
       << ":thickness=" << thickness
       << ":forwardWidth=" << forwardWidth << ":forwardHeight=" <<forwardHeight
       << ":backWidth=" << backWidth << ":backHeight=" << backHeight
       << endl;
       */
  slider=new WdkWindow (window,sliderX,sliderY,sliderWidth,sliderHeight);

  window->setBackground (style.dark);
  slider->setBackground (style.bg);
  forward->setBackground (style.bg);
  back->setBackground (style.bg);
  window->selectInput (ButtonPressMask|ButtonReleaseMask|ButtonMotionMask|ExposureMask);
  slider->selectInput (EnterWindowMask|LeaveWindowMask|ExposureMask);
  forward->selectInput (EnterWindowMask|LeaveWindowMask|ButtonPressMask|ButtonReleaseMask|ExposureMask);
  back->selectInput (EnterWindowMask|LeaveWindowMask|ButtonPressMask|ButtonReleaseMask|ExposureMask);
  slider->map ();
  forward->map ();
  back->map ();
}

void WtkRange::unlockedSetLimit (double limit_,bool isNeedScrolled=true)
{
  if (limit_<=0) limit_=1;
  if (limit!=limit_)
    {
      limit=limit_;
      updateSlider ();
      if (isNeedScrolled)
	scrolled (limit,position,scope);
    }
};

void WtkRange::unlockedSetScope (double scope_,bool isNeedScrolled=true)
{
  double minScope=0;
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
    case DIRECTION_UP:
      minScope=limit*minSliderSize/(height-forwardHeight-backHeight-style.horizontalShadowWidth*2);
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
    case DIRECTION_LEFT:
      minScope=limit*minSliderSize/(width-forwardWidth-backWidth-style.verticalShadowWidth*2);
      break;
    }
  assert (minScope>0);
  if (scope_<minScope) scope_=minScope;
  if (scope_>limit) scope_=limit;
  
  if (scope!=scope_)
    {
      scope=scope_;
      unitStep=scope/3;
      if (unitStep==0) unitStep=scope;
      updateSlider ();
      if (isNeedScrolled)
	scrolled (limit,position,scope);
    }
}

void WtkRange::unlockedSetPosition (double position_,bool isNeedScrolled=true)
{
  if (position_>limit-scope) position_=limit-scope;
  if (position_<0) position_=0;
  if (position!=position_)
    {
      position=position_;
      updateSlider ();
      if (isNeedScrolled)
	scrolled (limit,position,scope);
    }
}

void WtkRange::registerEventHandler (WtkEventHandler* eventHandler)
{
  eventHandler->addWidget (this);
  eventHandler->addWidget (this,forward);
  eventHandler->addWidget (this,back);
  eventHandler->addWidget (this,slider);
}


void WtkRange::exposed (XEvent* event)
{
  assert (event!=NULL);
  if (event->xany.window==window->getWindow ())
    widgetExposed (event);
  else if(event->xany.window==forward->getWindow ())
    forwardExposed (event);
  else if(event->xany.window==back->getWindow ())
    backExposed (event);
  else if(event->xany.window==slider->getWindow ())
    sliderExposed (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::exposed: incorrect XEvent." << endl;
    }
}

void WtkRange::buttonPressed (XEvent* event)
{
  assert (event!=NULL);
  if(event->xany.window==window->getWindow ())
    widgetButtonPressed (event);
  else if(event->xany.window==forward->getWindow ())
    forwardButtonPressed (event);
  else if(event->xany.window==back->getWindow ())
    backButtonPressed (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::buttonPressed: incorrect XEvent." << endl;
    }
}

void WtkRange::buttonReleased (XEvent* event)
{
  assert (event!=NULL);
  if(event->xany.window==window->getWindow ())
    widgetButtonReleased (event);
  else if(event->xany.window==forward->getWindow ())
    forwardButtonReleased (event);
  else if(event->xany.window==back->getWindow ())
    backButtonReleased (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::buttonReleased: incorrect XEvent." << endl;
    }
}

void WtkRange::pointerMotion (XEvent* event)
{
  assert (event!=NULL);
  if(event->xany.window==window->getWindow ())
    widgetPointerMotion (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::pointerMotion: incorrect XEvent." << endl;
    }
}

void WtkRange::enterNotify (XEvent* event)
{
  assert (event!=NULL);
  if(event->xany.window==forward->getWindow ())
    forwardEnterNotify (event);
  else if(event->xany.window==back->getWindow ())
    backEnterNotify (event);
  else if(event->xany.window==slider->getWindow ())
    sliderEnterNotify (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::EnterNotify: incorrect XEvent." << endl;
    }
}

void WtkRange::leaveNotify (XEvent* event)
{
  assert (event!=NULL);
  if(event->xany.window==forward->getWindow ())
    forwardLeaveNotify (event);
  else if(event->xany.window==back->getWindow ())
    backLeaveNotify (event);
  else if(event->xany.window==slider->getWindow ())
    sliderLeaveNotify (event);
  else
    {//should not reached
      cerr << "Error:WtkRange::LeaveNotify: incorrect XEvent." << endl;
    }
}


void WtkRange::widgetExposed (XEvent* /*event*/)
{
  decorate (window,width,height,style,SHADOW_DOWN);
}

void WtkRange::widgetButtonPressed (XEvent* event)
{
  double step=0;
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
    case DIRECTION_UP:
      if (event->xbutton.y<sliderY)
	step=-unitStep*2;//$B:8$K0\F0(B
      else if (event->xbutton.y<sliderY+(int) sliderHeight)
	{//slider$B$r(Bdrag
	  sliderMotionOffset=event->xbutton.y-sliderY;
	  isSliderDragged=true;
	}
      else
	step=unitStep*2;//$B1&$K0\F0(B
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
    case DIRECTION_LEFT:
      if (event->xbutton.x<sliderX)
	step=-unitStep*2;//$B>e$K0\F0(B
      else if (event->xbutton.x<sliderX+static_cast<int> (sliderWidth))
	{//slider$B$r(Bdrag
	  sliderMotionOffset=event->xbutton.x-sliderX;
	  isSliderDragged=true;
	}
      else 
	step=unitStep*2;//$B2<$K0\F0(B
      break;
    }
  unlockedSetPosition (position+step,true);
}

void WtkRange::widgetButtonReleased (XEvent* /*event*/)
{
  isSliderDragged=false;
}

double WtkRange::calcPosition (XEvent* event)
{
  double nowPosition=0;
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
      nowPosition=(event->xmotion.y-sliderMotionOffset)*limit/
	(height-forwardHeight-backHeight-style.horizontalShadowWidth*2);
      break;
    case DIRECTION_UP:
      nowPosition=(event->xmotion.y-sliderMotionOffset
		   -static_cast<double>(forwardHeight+backHeight))
	*limit
	/(height-forwardHeight-backHeight-style.horizontalShadowWidth*2);
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
      nowPosition=(event->xmotion.x-sliderMotionOffset)*limit/
	(width-forwardWidth-backWidth-style.verticalShadowWidth*2);
      break;
    case DIRECTION_LEFT:
      nowPosition=(event->xmotion.x-sliderMotionOffset
		   -static_cast<double>(forwardWidth+backWidth))
	*limit
	/(width-forwardWidth-backWidth-style.verticalShadowWidth*2);
      break;
    }
  return nowPosition;
}


void WtkRange::widgetPointerMotion (XEvent* event)
{
  if (isSliderDragged)
    unlockedSetPosition (calcPosition (event),true);
}


void WtkRange::updateSlider ()
{
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
      sliderY=static_cast<int> ((height-forwardHeight-backHeight-style.horizontalShadowWidth*2)*position/limit)
	+style.horizontalShadowWidth*2;
      sliderHeight=static_cast<uint> ((height-forwardHeight-backHeight-style.horizontalShadowWidth*2)*scope/limit);
      if (sliderHeight<minSliderSize) sliderHeight=minSliderSize;
      break;
    case DIRECTION_UP:
      sliderY=static_cast<int> ((height-forwardHeight-backHeight-style.horizontalShadowWidth*2)*position/limit)
	+forwardHeight+backHeight;
      sliderHeight=static_cast<uint> ((height-forwardHeight-backHeight-style.horizontalShadowWidth*2)*scope/limit);
      if (sliderHeight<minSliderSize) sliderHeight=minSliderSize;
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
      sliderX=static_cast<int> ((width-forwardWidth-backWidth-style.verticalShadowWidth*2)*position/limit)
	+style.verticalShadowWidth*2;
      sliderWidth=static_cast<uint> ((width-forwardWidth-backWidth-style.verticalShadowWidth*2)*scope/limit);
      if (sliderWidth<minSliderSize) sliderWidth=minSliderSize;
      break;
    case DIRECTION_LEFT:
      sliderX=static_cast<int> ((width-forwardWidth-backWidth-style.verticalShadowWidth*2)*position/limit)
	+forwardWidth+backWidth;
      sliderWidth=static_cast<uint> ((width-forwardWidth-backWidth-style.verticalShadowWidth*2)*scope/limit);
      if (sliderWidth<minSliderSize) sliderWidth=minSliderSize;
      break;
    }
  
  slider->moveResize (sliderX,sliderY,sliderWidth,sliderHeight);
  //if (isSliderDragged) isSliderActive=true;//???
  decorate (slider,sliderWidth,sliderHeight,isSliderActive,true);
  /*
  cerr << "WtkRange::updateSlider:limit=" << limit << ":position=" << position
       << ":scope=" << scope << endl;
  cerr << "sliderX=" << sliderX << ":sliderY=" << sliderY
       << ":sliderWidth=" << sliderWidth << ":sliderHeight=" << sliderHeight
       << endl;
       */
}


void WtkRange::sliderExposed (XEvent* /*event*/)
{
  decorate (slider,sliderWidth,sliderHeight,isSliderActive,true);
}

void WtkRange::sliderEnterNotify (XEvent* /*event*/)
{
  isSliderActive=true;
  decorate (slider,sliderWidth,sliderHeight,isSliderActive,true);
}

void WtkRange::sliderLeaveNotify (XEvent* /*event*/)
{
  isSliderActive=false;
  decorate (slider,sliderWidth,sliderHeight,isSliderActive,true);
}

void WtkRange::forwardDecorate ()
{
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
    case DIRECTION_UP:
      decorateTriangle
	(forward,forwardWidth,forwardHeight,isForwardActive,isForwardUp,
	 SHIFT_UP);
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
    case DIRECTION_LEFT:
      decorateTriangle
	(forward,forwardWidth,forwardHeight,isForwardActive,isForwardUp,
	 SHIFT_LEFT);
      break;
    }
}


void WtkRange::forwardExposed (XEvent* /*event*/)
{
  forwardDecorate ();
}


void WtkRange::forwardButtonPressed (XEvent* /*event*/)
{
  isForwardUp=false;
  unlockedSetPosition (position-unitStep,true);
  forwardDecorate ();
}

void WtkRange::forwardButtonReleased (XEvent* /*event*/)
{
  isForwardUp=true;
  forwardDecorate ();
}

void WtkRange::forwardEnterNotify (XEvent* /*event*/)
{
  isForwardActive=true;
  forwardDecorate ();
}

void WtkRange::forwardLeaveNotify (XEvent* /*event*/)
{
  isForwardActive=false;
  forwardDecorate ();
}


void WtkRange::backDecorate ()
{
  switch (direction)
    {
    case VERTICAL:
    case DIRECTION_DOWN:
    case DIRECTION_UP:      
      decorateTriangle (back,backWidth,backHeight,isBackActive,isBackUp,
			SHIFT_DOWN);
      break;
    case HORIZONTAL:
    case DIRECTION_RIGHT:
    case DIRECTION_LEFT:
      decorateTriangle (back,backWidth,backHeight,isBackActive,isBackUp,
			SHIFT_RIGHT);
      break;
    }
}


void WtkRange::backExposed (XEvent* /*event*/)
{
  backDecorate ();
}

void WtkRange::backButtonPressed (XEvent* /*event*/)
{
  isBackUp=false;
  unlockedSetPosition (position+unitStep,true);
  backDecorate ();
}

void WtkRange::backButtonReleased (XEvent* /*event*/)
{
  isBackUp=true;
  backDecorate ();
}

void WtkRange::backEnterNotify (XEvent* /*event*/)
{
  isBackActive=true;
  backDecorate ();
}

void WtkRange::backLeaveNotify (XEvent* /*event*/)
{
  isBackActive=false;
  backDecorate ();
}

void WtkRange::decorate (WdkWindow* wdkWindow,
			 uint windowWidth,uint windowHeight,
			 bool isActive,bool isUp)
{
  if (isActive)
    wdkWindow->setBackground (activeStyle.bg);
  else
    wdkWindow->setBackground (style.bg);
  wdkWindow->clear ();

  decorate (wdkWindow,windowWidth,windowHeight,
	    (isActive)? activeStyle:style,(isUp)?SHADOW_UP:SHADOW_DOWN);
};


void WtkRange::decorateTriangle (WdkWindow* wdkWindow,
				 uint windowWidth,uint windowHeight,
				 bool isActive,bool isUp,SHIFT shift)
{
  if (isActive)
    wdkWindow->setBackground (activeStyle.bg);
  else
    wdkWindow->setBackground (style.bg);
  wdkWindow->clear ();
#if 0
  XPoint points1[3];
  XPoint points2[3];
  ulong lineWidth=1;
  switch (shift)
    {
    case SHIFT_UP:
      lineWidth=style.horizontalShadowWidth;
      points1[0].x=style.verticalShadowWidth*2;
      points1[0].y=windowHeight-style.horizontalShadowWidth*3;
      points1[1].x=windowWidth/2;
      points1[1].y=style.horizontalShadowWidth*2;
      points1[2].x=windowWidth-style.verticalShadowWidth*2;
      points1[2].y=windowHeight-style.horizontalShadowWidth*3;

      points2[0].x=style.verticalShadowWidth*2;
      points2[0].y=windowHeight-style.horizontalShadowWidth*2;
      points2[1].x=windowWidth/2;
      points2[1].y=style.horizontalShadowWidth*3;
      points2[2].x=windowWidth-style.verticalShadowWidth*2;
      points2[2].y=windowHeight-style.horizontalShadowWidth*2;
      break;
    case SHIFT_DOWN:
      lineWidth=style.horizontalShadowWidth;
      points1[0].x=style.verticalShadowWidth*2;
      points1[0].y=style.horizontalShadowWidth*2;
      points1[1].x=windowWidth/2;
      points1[1].y=windowHeight-style.horizontalShadowWidth*3;
      points1[2].x=windowWidth-style.verticalShadowWidth*2;
      points1[2].y=style.horizontalShadowWidth*2;

      points2[0].x=style.verticalShadowWidth*2;
      points2[0].y=style.horizontalShadowWidth*3;
      points2[1].x=windowWidth/2;
      points2[1].y=windowHeight-style.horizontalShadowWidth*2;
      points2[2].x=windowWidth-style.verticalShadowWidth*2;
      points2[2].y=style.horizontalShadowWidth*3;
      break;
    case SHIFT_LEFT:
      lineWidth=style.verticalShadowWidth;
      points1[0].x=windowWidth-style.verticalShadowWidth*3;
      points1[0].y=style.horizontalShadowWidth*2;
      points1[1].x=style.verticalShadowWidth*2;
      points1[1].y=windowHeight/2;
      points1[2].x=windowWidth-style.verticalShadowWidth*3;
      points1[2].y=windowHeight-style.horizontalShadowWidth*2;

      points2[0].x=windowWidth-style.verticalShadowWidth*2;
      points2[0].y=style.horizontalShadowWidth*2;
      points2[1].x=style.verticalShadowWidth*3;
      points2[1].y=windowHeight/2;
      points2[2].x=windowWidth-style.verticalShadowWidth*2;
      points2[2].y=windowHeight-style.horizontalShadowWidth*2;
      break;
    case SHIFT_RIGHT:
      lineWidth=style.verticalShadowWidth;
      points1[0].x=style.verticalShadowWidth*2;
      points1[0].y=style.horizontalShadowWidth*2;
      points1[1].x=windowWidth-style.verticalShadowWidth*3;
      points1[1].y=windowHeight/2;
      points1[2].x=style.verticalShadowWidth*2;
      points1[2].y=windowHeight-style.horizontalShadowWidth*2;

      points2[0].x=style.verticalShadowWidth*3;
      points2[0].y=style.horizontalShadowWidth*2;
      points2[1].x=windowWidth-style.verticalShadowWidth*2;
      points2[1].y=windowHeight/2;
      points2[2].x=style.verticalShadowWidth*3;
      points2[2].y=windowHeight-style.horizontalShadowWidth*2;
      break;
    }

  lockData ();
  gc->setLineAttributes (lineWidth,LineSolid,CapButt,JoinMiter);
  gc->setForeground (style.black);
  //gc->setForeground (style.dark);
  wdkWindow->drawLines (gc,points1,3,CoordModeOrigin);
  //gc->setForeground (style.dark);
  gc->setForeground (style.light);
  wdkWindow->drawLines (gc,points2,3,CoordModeOrigin);
  unlockData ();
#else
  XPoint points[4];
  switch (shift)
    {
    case SHIFT_UP:
      points[0].x=windowWidth/2;
      points[0].y=style.horizontalShadowWidth*3;
      points[1].x=windowWidth-style.verticalShadowWidth*3;
      points[1].y=windowHeight-style.horizontalShadowWidth*3;
      points[2].x=windowWidth/2;
      points[2].y=windowHeight/2;
      points[3].x=style.verticalShadowWidth*3;
      points[3].y=windowHeight-style.horizontalShadowWidth*3;
      break;
    case SHIFT_DOWN:
      points[0].x=windowWidth/2;
      points[0].y=windowHeight-style.horizontalShadowWidth*3;
      points[1].x=windowWidth-style.verticalShadowWidth*3;
      points[1].y=style.horizontalShadowWidth*3;
      points[2].x=windowWidth/2;
      points[2].y=windowHeight/2;
      points[3].x=style.verticalShadowWidth*3;
      points[3].y=style.horizontalShadowWidth*3;
      break;
    case SHIFT_RIGHT:
      points[0].x=windowWidth-style.verticalShadowWidth*3;
      points[0].y=windowHeight/2;
      points[1].x=style.verticalShadowWidth*3;
      points[1].y=style.horizontalShadowWidth*3;
      points[2].x=windowWidth/2;
      points[2].y=windowHeight/2;
      points[3].x=style.verticalShadowWidth*3;
      points[3].y=windowHeight-style.horizontalShadowWidth*3;
      break;
    case SHIFT_LEFT:
      points[0].x=style.verticalShadowWidth*3;
      points[0].y=windowHeight/2;
      points[1].x=windowWidth-style.verticalShadowWidth*3;
      points[1].y=style.horizontalShadowWidth*3;
      points[2].x=windowWidth/2;
      points[2].y=windowHeight/2;
      points[3].x=windowWidth-style.verticalShadowWidth*3;
      points[3].y=windowHeight-style.horizontalShadowWidth*3;
      break;
    }

  lockData ();
  gc->setForeground (style.black);
  //gc->setForeground (style.light);
  //gc->setForeground (style.middle);
  //gc->setForeground (style.dark);
  gc->setFillRule (EvenOddRule);
  wdkWindow->fillPolygon (gc,points,4,Nonconvex,CoordModeOrigin);
  unlockData ();
#endif
  decorate (wdkWindow,windowWidth,windowHeight,
	    (isActive)? activeStyle:style,(isUp)?SHADOW_UP:SHADOW_DOWN);
}
