/*
    WtkWidget.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <iostream>

#include "WtkWidget.h"

using namespace std;

const char* const WtkWidget::defaultResource=
"*Foreground: black\n"
"*Background: gray80\n"
"*Light: gray90\n"
"*Mid: gray80\n"
"*Dark: gray70\n"
"*TextColor: black\n"
"*VerticalShadowWidth: 1\n"
"*HorizontalShadowWidth: 1\n"
"*ShadowType: UP"
"*X: 0\n"
"*Y: 0\n"
"*Width: 640\n"
"*Height: 400\n"
"*FontSet: -misc-fixed-*--14-*\n";

void WtkWidget::initialize ()
{
  getResource ();
  window->setColormap (style.getColormap ());
  window->selectInput (ExposureMask);
  gc=new WdkGC (window);
  setResource ();
}

WtkWidget::WtkWidget (Display* display,
		      const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_,
		      int depth,Visual* visual,Colormap colormap_,
		      char* resourceFileName):
  parent (NULL),window (NULL),gc (NULL),
  x (x_),y (y_),width (width_),height (height_),
  dataLock (),
  className (NULL),name (NULL),
  eventStateLock ()
{
  style.setColormap (display,colormap_);
  char* const rdbString=XResourceManagerString (display);
  if (rdbString==NULL)
    displayRdb=XrmGetStringDatabase ("");
  else
    displayRdb=XrmGetStringDatabase (rdbString);
  privateRdb=XrmGetStringDatabase (defaultResource);
  if (resourceFileName!=NULL)
    XrmCombineFileDatabase (resourceFileName,&privateRdb,True);
  setResourceName (NULL,className_,NULL,name_);
  getWindowSize ();
  window=new WdkWindow (display,NULL,x,y,width,height,0,depth,visual);
  initialize ();
}

WtkWidget::WtkWidget (Display* display,
		      const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_,
		      int depth,Visual* visual,Colormap colormap_,
		      XrmDatabase displayRdb_):
  parent (NULL),window (NULL),gc (NULL),
  x (x_),y (y_),width (width_),height (height_),
  dataLock (),
  displayRdb (displayRdb_),className (NULL),name (NULL),
  eventStateLock ()
{
  privateRdb=XrmGetStringDatabase (defaultResource);
  style.setColormap (display,colormap_);
  setResourceName (NULL,className_,NULL,name_);
  getWindowSize ();
  window=new WdkWindow (display,NULL,x,y,width,height,0,depth,visual);
  initialize ();
}

WtkWidget::WtkWidget (WtkWidget* parent_,
		      const char* className_,const char* name_,
		      int x_,int y_,uint width_,uint height_):
  parent (parent_),window (NULL),gc (NULL),
  x (x_),y (y_),width (width_),height (height_),
  dataLock (),
  displayRdb (parent->displayRdb),privateRdb (parent->privateRdb),
  className (NULL),name (NULL),
  eventStateLock ()
{
  XrmCombineDatabase(XrmGetStringDatabase (defaultResource),&privateRdb,False);
  parent->addSubWindow (this);
  style.setColormap (parent->window->getDisplay (),parent->style.getColormap ());
  setResourceName (parent->className,className_,parent->name,name_);
  getWindowSize ();
  window=new WdkWindow (parent->getWindow (),x,y,width,height);
  initialize ();
}


WtkWidget::~WtkWidget ()
{
  delete gc;
  delete window;
  delete[] className;
  delete[] name;

  //XrmDestroyDatabase (displayRdb);
  if (parent==NULL)
    XrmDestroyDatabase (privateRdb);
}

char* WtkWidget::getResourceName (const char* parent,const char* me)
{
  assert (me!=NULL);
  char* dest;
  if (parent==NULL)
    {
      dest=new char[strlen (me)+1];
      strcpy (dest,me);
    }
  else
    {
      dest=new char[strlen (parent)+strlen (me)+2];
      sprintf (dest,"%s.%s",parent,me);
    }
  return dest;
}

void WtkWidget::setResourceName (const char* parentClassName_,
				  const char* ClassName_,
				  const char* parentName_,const char* name_)
{
  className=getResourceName (parentClassName_,ClassName_);
  name=getResourceName (parentName_,name_);
}

char* WtkWidget::getResource (char* resourceName,char* resourceClass)
{
  char* fullResource=getResourceName (name,resourceName);
  char* fullClass=getResourceName (className,resourceClass);
  
  char* typeReturn;
  XrmValue valueReturn;
  Bool result=XrmGetResource (displayRdb,fullResource,fullClass,
			      &typeReturn,&valueReturn);
  if (!result)//user did not specified the resouce.so use default resource.
    result=XrmGetResource (privateRdb,fullResource,fullClass,
			   &typeReturn,&valueReturn);
  if (result)
    {//user specifiyed the resource
      /*
	#ifdef DEBUG
	cerr << "name=" << name
	<< ":res=" << resourceName << ":class=" << resourceClass
	<< ":value=" << (char*) valueReturn.addr << endl;
	#endif
	*/
      delete[] fullResource;
      delete[] fullClass;
      return (char*) valueReturn.addr;
    }
  
  //apprication should prepare default resource not to reach here.
#ifdef DEBUG
  cerr << "not found resource.name=" << name
       << ":resource=" << resourceName
       << ":class="  << resourceClass << endl;
#endif
  return NULL;
}

int WtkWidget::getResourceValue (char* resourceName,char* resourceClass)
{
  const char* value=getResource (resourceName,resourceClass);
  return (value==NULL)?0:atoi (value);//XXX
}

void WtkWidget::getResource ()
{
  //XrmPutFileDatabase (displayRdb,"testtesttest.ad");
  style.allocColor (window->getDisplay (),FOREGROUND,
		    getResource ("foreground","Foreground"));
  style.allocColor (window->getDisplay (),BACKGROUND,
		    getResource ("background","Background"));
  style.allocColor (window->getDisplay (),LIGHT,
		    getResource ("light","Light"));
  style.allocColor (window->getDisplay (),MID,
		    getResource ("mid","Mid"));
  style.allocColor (window->getDisplay (),DARK,
		    getResource ("dark","Dark"));
  style.allocColor (window->getDisplay (),TEXTCOLOR,
		    getResource ("textColor","TextColor"));
#if 0
  //$B$3$l$@$H(BLC_CTYPE=ja_JP.eucJP$B$r;XDj$7$?$H$-@5>o$K5/F0$7$J$$!#(B
  style.createFontSet (window->getDisplay (),getResource ("fontSet","FontSet"));
#else
  //XXX
  //should use hash!!!!!!!!!!!!!!!!!!!!!!!!!!
  if (parent==NULL)
    style.createFontSet (window->getDisplay (),getResource ("fontSet","FontSet"));
  else
    style.fontSet=parent->style.fontSet;
#endif
  

  style.verticalShadowWidth=getResourceValue ("verticalShadowWidth",
					      "VerticalShadowWidth");
  style.horizontalShadowWidth=getResourceValue ("horizontalShadowWidth",
						"HorizontalShadowWidth");
  const char* shadowType=getResource ("shadowType","ShadowType");
  if (strcmp (shadowType,"FLAT")==0)
    style.shadowType=SHADOW_FLAT;
  else if (strcmp (shadowType,"DOWN")==0)
    style.shadowType=SHADOW_DOWN;
  else
    style.shadowType=SHADOW_UP;
  if (style.shadowType==SHADOW_FLAT)
    {//$BOH$J$7$J$N$G(B,$BOH$NI}$O(B0
      style.leftOffset=0;
      style.topOffset=0;
      style.rightOffset=0;
      style.bottomOffset=0;
    }
  else
    {//$BOH$rIA$/$N$G$=$NJ,!#$3$l$O(Bdecorate ()$B$K0MB8!#(B
      style.leftOffset=style.verticalShadowWidth+1;
      style.topOffset=style.horizontalShadowWidth+1;
      style.rightOffset=3+style.verticalShadowWidth;
      style.bottomOffset=2+style.horizontalShadowWidth;
    }
}

void WtkWidget::getWindowSize ()
{
  if (x<0)
    x=getResourceValue ("x","X");
  if (y<0)
    y=getResourceValue ("y","Y");
  if (width==0)
    width=getResourceValue ("width","Width");
  if (height==0)
    height=getResourceValue ("height","Height");
}


void WtkWidget::setResource ()
{
  gc->setForeground (style.fg);
  gc->setBackground (style.bg);
  window->setBackground (style.bg);
}


void WtkWidget::decorate (WdkWindow* wdkWindow,
			  uint windowWidth,uint windowHeight,
			  WtkStyle& nowStyle,ShadowType shadowType)
{
  WdkColor* leftTopColor=NULL;
  WdkColor* topColor=NULL;
  WdkColor* rightBottomColor=NULL;
  WdkColor* bottomColor=NULL;
  //set color
  switch (shadowType)
    {
    case SHADOW_UP:
      leftTopColor=&nowStyle.light;
      topColor=&nowStyle.mid;
      rightBottomColor=&nowStyle.dark;
      bottomColor=&nowStyle.black;
      break;
    case SHADOW_FLAT:
      return;//no decoration
      break;
    case SHADOW_DOWN:
      leftTopColor=&nowStyle.dark;
      topColor=&nowStyle.black;
      rightBottomColor=&nowStyle.light;
      bottomColor=&nowStyle.mid;
      break;
    }
  decorate (wdkWindow,windowWidth,windowHeight,
	    nowStyle,leftTopColor,topColor,rightBottomColor,bottomColor);
}



void WtkWidget::decorate (WdkWindow* wdkWindow,
			  uint windowWidth,uint windowHeight,
			  WtkStyle& nowStyle,
			  WdkColor* leftTopColor,WdkColor* topColor,
			  WdkColor* rightBottomColor,WdkColor* bottomColor)
{
  assert (leftTopColor!=NULL);
  assert (topColor!=NULL);
  assert (rightBottomColor!=NULL);
  assert (bottomColor!=NULL);
  //draw shadow
  /*
    #ifdef DEBUG
    cerr << "style=" << style.shadowType << endl;
    cerr << "verticalShadowWidth=" << style.verticalShadowWidth
    << ":horizontalSadowWidth=" << style.horizontalShadowWidth << endl;
    cerr << "leftTop"
    << ":red=" << leftTopColor->color.red
    << ":green=" << leftTopColor->color.green
    << ":red=" << leftTopColor->color.red << endl;
    cerr << "rightBottom"
    << ":red=" << rightBottomColor->color.red
    << ":green=" << rightBottomColor->color.green
    << ":red=" << rightBottomColor->color.red << endl;
    cerr << "windowWidth=" << windowWidth << ":windowHeight=" << windowHeight << endl;
    #endif
    */

  lockData ();
  //right bottom
  gc->setForeground (rightBottomColor);
  gc->setLineAttributes (nowStyle.horizontalShadowWidth,
			 LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       1,            windowHeight-2-nowStyle.horizontalShadowWidth/2,
		       windowWidth-1,windowHeight-2-nowStyle.horizontalShadowWidth/2);
  gc->setLineAttributes (nowStyle.verticalShadowWidth,
			 LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       windowWidth-2-nowStyle.verticalShadowWidth/2,1,
		       windowWidth-2-nowStyle.verticalShadowWidth/2,windowHeight-1);

  gc->setForeground (bottomColor);
  gc->setLineAttributes (1,LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       0,          windowHeight-1,
		       windowWidth,windowHeight-1);
  gc->setLineAttributes (1,LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       windowWidth-1,0,
		       windowWidth-1,windowHeight-1);


  //left top
  gc->setForeground (leftTopColor);
  gc->setLineAttributes (nowStyle.horizontalShadowWidth,
			 LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       0,            nowStyle.horizontalShadowWidth/2,
		       windowWidth-1,nowStyle.horizontalShadowWidth/2);
  gc->setLineAttributes (nowStyle.verticalShadowWidth,
			 LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       nowStyle.verticalShadowWidth/2,0,
		       nowStyle.verticalShadowWidth/2,windowHeight-1);

  gc->setForeground (topColor);
  gc->setLineAttributes (1,LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       nowStyle.verticalShadowWidth,nowStyle.horizontalShadowWidth,
		       windowWidth-2,               nowStyle.horizontalShadowWidth);
  gc->setLineAttributes (1,LineSolid,CapButt,JoinMiter);
  wdkWindow->drawLine (*gc,
		       nowStyle.verticalShadowWidth,nowStyle.horizontalShadowWidth,
		       nowStyle.verticalShadowWidth,windowHeight-2);
  unlockData ();
}


void WtkWidget::updateEventState (EventState& srcEventState)
{
  lockEventState ();
  eventState.isExposed=srcEventState.isExposed;
  
  if (eventState.crossing==EVENT_CROSSING_ENTER &&
      srcEventState.crossing==EVENT_CROSSING_LEAVE)
    eventState.crossing=EVENT_CROSSING_NONE;
  else if (srcEventState.crossing!=EVENT_CROSSING_NONE)
    eventState.crossing=srcEventState.crossing;
    
  for (int i=0;i<5;i++)
    {
      if (eventState.isButtonPressed[i] && srcEventState.isButtonReleased[i])
	{
	  eventState.isButtonPressed[i]=false;
	  eventState.isButtonReleased[i]=false;
	  eventState.isClicked[i]=true;
	}
      else
	{
	  eventState.isButtonPressed[i]=srcEventState.isButtonPressed[i];
	  eventState.isButtonReleased[i]=srcEventState.isButtonReleased[i];
	  if (srcEventState.isClicked[i])
	    eventState.isClicked[i]=true;
	}
    }
  unlockEventState ();
}


void WtkWidget::buttonPressed (XEvent* event)
{
  lockEventState ();
  eventState.isButtonPressed[event->xbutton.button]=true;
  unlockEventState ();
};

void WtkWidget::buttonReleased (XEvent* event)
{
  lockEventState ();
  if (eventState.isButtonPressed[event->xbutton.button])
    {
      eventState.isButtonPressed[event->xbutton.button]=false;
      eventState.isClicked[event->xbutton.button]=true;
    }
  else
    {
      eventState.isButtonReleased[event->xbutton.button]=true;
    }
  unlockEventState ();
};

void WtkWidget::enterNotify (XEvent* /*event*/)
{
  lockEventState ();
  eventState.crossing=EVENT_CROSSING_ENTER;
  unlockEventState ();
};

void WtkWidget::leaveNotify (XEvent* /*event*/)
{
  lockEventState ();
  if (eventState.crossing==EVENT_CROSSING_ENTER)
    eventState.crossing=EVENT_CROSSING_NONE;
  else
    eventState.crossing=EVENT_CROSSING_LEAVE;
  unlockEventState ();
};


void WtkWidget::printResource () const
{
  cout << "classname=" << className << endl;
  cout << "res. name=" << name << endl;
  cout << name << ".x=" << x << endl;
  cout << name << ".y=" << y << endl;
  cout << name << ".width=" << width << endl;
  cout << name << ".height=" << height << endl;
}
