#include <algorithm>
#include <functional>
#include <vector>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "IOXSherry.h"
#include "SherryArea.h"
#include "IO/ioutil/countCharSize.h"
#include "util/PThreadAttribute.h"
#include "util/schedYield.h"
#include "PlugIn/X/xutil/getResource.h"
#include "PlugIn/X/xutil/createFontSet.h"
#include "PlugIn/X/xutil/findVisual.h"
#include "PlugIn/X/WidiToolkit/WdkXImage.h"

#define SHERRY_WRD_DEBUG
//#undef SHERRY_WRD_DEBUG

using namespace std;

const char* const IOXSherry::defaultResource;
const char* const IOXSherry::resourceFileName;
const char* const IOXSherry::className;
const char* const IOXSherry::resourceName;
const Atom IOXSherry::XSHERRY_EVENT_HANDLER_LOOP_END_ATOM;
const int IOXSherry::CLIENT_MESSAGE_LENGTH;
const char IOXSherry::XSHERRY_EVENT_HANDLER_LOOP_CANCEL_MESSAGE[CLIENT_MESSAGE_LENGTH];

IOXSherry::IOXSherry (State* state__,
		      WidiController* widiController__,
		      Option* option__):
  SherryWrd (state__, widiController__, option__),
  isPrivateColormapAllocated (false),
  isEventHandleThreadCreated (false)
{
  for (int i = 0; i < MAX_VIRTUAL_SCREENS; i++)
    virtualScreen[i] = NULL;
}

IOXSherry::~IOXSherry ()
{
  
}

void IOXSherry::initialize ()
{
  SherryWrd::initialize ();
  
  allocateXResource ();
}

void IOXSherry::release ()
{
  releaseXResource ();
  
  SherryWrd::release ();
}


void IOXSherry::allocateXResource ()
{
  //XXX
  //display name$B$r;XDj$9$kJ}K!$r9M$($k$Y$7!#(B
  //$B0?$O(Bresource$B$+$i<hF@(B?
  display = XOpenDisplay (NULL);
  screenNumber = DefaultScreen (display);
  
  ////////////////////////////////////////////////resource data base$B$N@_Dj(B
  //display resource
  const char* const str = XResourceManagerString (display);
  if (str == NULL)
    displayRdb = NULL;
  else
    displayRdb = XrmGetStringDatabase (str);
  //this app's defualt resource
  privateRdb = XrmGetFileDatabase (resourceFileName);
  XrmCombineDatabase (XrmGetStringDatabase (defaultResource), &privateRdb,
		      False);
  
  //////////////////////////////////////////////////////////fontset$B$N@_Dj(B
  const char* const fontName = getResource (privateRdb, displayRdb,
					    className, "Fontset",
					    resourceName, "fontset");
#ifdef DEBUG
  cerr << "IOXSherry::allocateXResource fontName=" << fontName << endl;
#endif
  fontSet = createFontSet (display, fontName);
  
  /////////////////////////////////////////////////////////////visual$B$N<hF@(B
  findVisual (display, screenNumber, &visual, &visualClass, &depth);
  
  //////////////////////////////////////////////////////////////window$B$N@8@.(B
  window = new WdkWindow (display, NULL, 0, 0,
			  REAL_SCREEN_SIZE_X, REAL_SCREEN_SIZE_Y,
			  0, depth, visual);
  XSizeHints* sizeHints = XAllocSizeHints ();
  if (sizeHints != NULL)
    {
      sizeHints->flags = PMaxSize | PMinSize | PResizeInc | PBaseSize;
      sizeHints->max_width = REAL_SCREEN_SIZE_X;
      sizeHints->max_height = REAL_SCREEN_SIZE_Y;
      sizeHints->min_width = REAL_SCREEN_SIZE_X;
      sizeHints->min_height = REAL_SCREEN_SIZE_Y;
      sizeHints->width_inc = 0;
      sizeHints->height_inc = 0;
      sizeHints->base_width = REAL_SCREEN_SIZE_X;
      sizeHints->base_height = REAL_SCREEN_SIZE_Y;
      window->setWMNormalHints (sizeHints);
      XFree (sizeHints);
    }
  else
    {
      cerr << "IOXSherry::allocateXResource:XAllocSizeHints can not allocate."
	   << endl;
    }
  realScreen = new WdkPixmap (window, REAL_SCREEN_SIZE_X, REAL_SCREEN_SIZE_Y);
  defaultGC = new WdkGC (window);
  gc = new WdkGC (window);
  
  //////////////////////////////////////////////////colormap,color cell$B$N<hF@(B
  colormap = DefaultColormap (display, screenNumber);
  if (!XAllocColorCells (display, colormap, False,
			 planePixel, MAX_PLANES,
			 &basePixel, 1))
    {//$B6&M-(Bcolor cell$B$N3NJ]$K<:GT!#(B
      //private color cell$B$r3NJ]$9$k!#(B
      cerr << "Can not allocate color."
	   << "trying to create private colormap" << endl;
      colormap = XCreateColormap (display, window->getWindow (),
				  visual, AllocNone);
      isPrivateColormapAllocated = true;
      if (!XAllocColorCells (display, colormap, False,
			     planePixel, MAX_PLANES,
			     &basePixel, 1))
	{
	  cerr << "Can not allocate color cells of private colormap." << endl;
	  //XXX$B$b$C$H$^$7$J%(%i!<=hM}$N;EJ}$,$"$kH&!#(B
	  exit (1);
	}
      //XInstallColormap (display, colormap);
      window->setColormap (colormap);
    }
  
  for (int i = 0; i < MAX_COLORS; i++)
    {
      palette2Pixel[i] = basePixel;
      for (int j = 0; j < MAX_PLANES; j++)
	{
	  if ((i & (1 << j)) != 0) palette2Pixel[i] |= planePixel[j];
	}
    }

  ///////////////////////////////////////////////////window$B$N(Bmap
  window->setBackground (basePixel);
  window->selectInput (ExposureMask);
  window->map ();

  /////////////////////////////////////////////////map$B$5$l$?$3$H$N3NG'(B
  XFlush ();
  XEvent event;
  XWindowEvent (display, window->getWindow (), ExposureMask, &event);
  window->selectInput (0);
  
  createEventHandleThread ();
}
    
void IOXSherry::releaseXResource ()
{
  cancelEventHandleThread ();
  
  //XUninstallColormap (display, colormap);
  delete gc;
  delete defaultGC;
  delete realScreen;
  delete window;
  unsigned long planeMask = 0;
  for (int i = 0; i < MAX_PLANES; i++)
    planeMask |= planePixel[i];
  XFreeColors (display, colormap, &basePixel, 1, planeMask);//???
  if (isPrivateColormapUsed ())
    XFreeColormap (display, colormap);
  XFreeFontSet (display, fontSet);
  XrmDestroyDatabase (privateRdb);
  XrmDestroyDatabase (displayRdb);
  XCloseDisplay (display);
}



//////////////////////////////////////////////////////////////////
void IOXSherry::prePlay ()
{
  window->setBackground (basePixel);
  window->clear ();
  defaultGC->setForeground (basePixel);
  defaultGC->setBackground (basePixel);
  gc->setForeground (basePixel);
  gc->setBackground (basePixel);
  gc->resetClipMask ();
  //gc->setPlaneMask (~0UL);
  gc->setPlaneMask (0xffffffff);
  realScreen->fillRectangle (defaultGC,
			     0, 0, REAL_SCREEN_SIZE_X, REAL_SCREEN_SIZE_Y);
  for (int i = 0; i < MAX_VIRTUAL_SCREENS; i++)
    delete virtualScreen[i];
  XColor xcolors[MAX_COLORS];
  for (int i = 0; i < MAX_COLORS; i++)
    {
      xcolors[i].pixel = palette2Pixel[i];
      xcolors[i].red   = 0;
      xcolors[i].green = 0;
      xcolors[i].blue  = 0;
      xcolors[i].flags = DoRed | DoGreen | DoBlue;
    }
  XStoreColors (display, colormap, xcolors, MAX_COLORS);
  
  SherryWrd::prePlay ();
  if (sherryWrdFile == NULL)//sherry wrd file is not found
    return;

  string title = "Widi:XSherry ";
  title += sherryWrdFile->getFileName ();
  window->storeName (title.c_str ());
  window->setIconName (title.c_str ());
  XFlush ();
}

void IOXSherry::postPlay ()
{


  
  SherryWrd::postPlay ();
}



///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////Sherry$B$N:F@8;~$N=hM}(B
///////////////////////////////////////////////////////////////////////
void IOXSherry::realTransferScreen (uint16 screen,
				    uint16 fromLeft, uint16 fromTop,
				    uint16 toLeft, uint16 toTop,
				    uint16 width, uint16 height)
{
  assert (virtualScreen[screen] != NULL);
  virtualScreen[screen]->copyArea (realScreen, defaultGC,
				   fromLeft, fromTop, width, height,
				   toLeft, toTop);
}

void IOXSherry::updatePalette ()
{
  XColor xcolors[MAX_COLORS];
  for (int i = 0; i < MAX_COLORS; i++)
    {
      xcolors[i].pixel = palette2Pixel[i];
      //NOTE 255*257 = 65535
#if 1
      xcolors[i].red   = realPalette.getEntry (i).getR ()*257UL;
      xcolors[i].green = realPalette.getEntry (i).getG ()*257UL;
      xcolors[i].blue  = realPalette.getEntry (i).getB ()*257UL;
#else
      xcolors[i].red   = realPalette.getEntry (i).getR () << 8;
      xcolors[i].green = realPalette.getEntry (i).getG () << 8;
      xcolors[i].blue  = realPalette.getEntry (i).getB () << 8;
#endif
      xcolors[i].flags = DoRed | DoGreen | DoBlue;
      
#ifdef SHERRY_WRD_DEBUG
      cerr << "IOXSherry::updatePalette"
	   << i << "-th color:"
	   << "pixel=" << xcolors[i].pixel
	   << " r=" << xcolors[i].red
	   << " g=" << xcolors[i].green
	   << " b=" << xcolors[i].blue
	   << endl;
#endif
    }
  XStoreColors (display, colormap, xcolors, MAX_COLORS);
  XFlush ();
}

void IOXSherry::updateScreen ()
{
  //$B$3$l$G(Bexpose event$B$N=hM}$OI,MWL5$7!#(B
  window->setBackgroundPixmap (realScreen);
  window->clear ();
  XFlush ();
}

void IOXSherry::doNewScreen (const SherryParameter::ScreenParameter* screens,
			     int arraySize)
{
  assert (arraySize > 0);
  for (int i = 0; i < arraySize; i++)
    {
      const uint16 screenNumber = screens[i].screenNumber;
      assert (virtualScreen[screenNumber] == NULL);
      virtualScreen[screenNumber] =
	new WdkPixmap (window, screens[i].sizeX, screens[i].sizeY);
      gc->setForeground (palette2Pixel[screens[i].transparentColor]);
      //gc->setPlaneMask (~0UL);
      gc->setPlaneMask (0xffffffff);
      virtualScreen[screenNumber]->fillRectangle (gc, 0, 0,
						  screens[i].sizeX,
						  screens[i].sizeY);
    }
}

void IOXSherry::doDeleteScreen (const uint16* screens, int arraySize)
{
  assert (arraySize > 0);
  for (int i = 0; i < arraySize; i++)
    {
      assert (virtualScreen[screens[i]] != NULL);
      delete virtualScreen[screens[i]];
      virtualScreen[screens[i]] = NULL;
    }
}

void IOXSherry::doNewScreenAndLoadPng (uint16 screen, uint16 palett,
				       const PNG* png)
{
  assert (png != NULL);
  assert (virtualScreen[screen] == NULL);

  const unsigned int width = png->getWidth ();
  const unsigned int height = png->getHeight ();
  WdkXImage ximage (display, visual, depth, ZPixmap, 0, width ,height);
  for (uint y = 0; y < height; y++)
    {
      for (uint x = 0; x < width; x++)
	{
	  ximage.putPixel (x, y, palette2Pixel[png->getPixel (x, y)]);
	}
    }
  virtualScreen[screen] = new WdkPixmap (window, width, height);
  ximage.putImage (virtualScreen[screen], defaultGC,
		   0, 0, 0, 0, width, height);
}

void IOXSherry::doDrawText (uint16 screen,
			    uint8 planeMask, uint8 mode,
			    uint8 foreground, uint8 background,
			    uint16 left, uint16 top,
			    const char* eucStr, int eucLength)
{
  assert (mode == 0 || mode == 1 || mode == 2 || mode ==3);
  setGC (gc, planeMask, foreground);
  gc->setBackground (palette2Pixel[background]);

  XRectangle ink;
  XRectangle logical;
#if 1
  XmbTextExtents (fontSet, eucStr, eucLength, &ink, &logical);
#else
  char* tmpEucStr = new char[eucLength];
  memcpy (tmpEucStr, eucStr, eucLength);
  XmbTextExtents (fontSet, tmpEucStr, eucLength, &ink, &logical);
  delete[] tmpEucStr;
#endif

  switch (mode)
    {
    case 0:
      //do nothing
      break;
    case 1:
      virtualScreen[screen]->drawString (fontSet, gc,
					 left-logical.x, top-logical.y,
					 eucStr, eucLength);
      break;
    case 2:
      {
	const int charSize = countCharSize (eucStr, eucLength);
	const uint clipWidth = charSize * CHAR_WIDTH;
	const uint clipHeight = CHAR_HEIGHT;
#ifdef SHERRY_WRD_DEBUG
	cerr << "IOXSherry::doDrawText"
	     << " charSize=" << charSize
	     << " clipWidth=" << clipWidth
	     << " clipHeight=" << clipHeight
	     << endl;
	
#endif
	WdkPixmap clipMask (window, clipWidth, clipHeight, 1);
	WdkGC clipGC (&clipMask);
	clipGC.setForeground (0UL);
	clipGC.setBackground (1UL);
	clipMask.drawImageString (fontSet, clipGC,
				  -logical.x, -logical.y, eucStr, eucLength);
	gc->setClipMask (clipMask);
	gc->setClipOrigin (left, top);
	gc->setForeground (palette2Pixel[background]);
	virtualScreen[screen]->fillRectangle (gc,
					      left, top,
					      clipWidth, clipHeight);
	gc->resetClipMask ();
	break;
      }
    case 3:
      virtualScreen[screen]->drawImageString (fontSet, gc,
					      left-logical.x, top-logical.y,
					      eucStr, eucLength);
      break;
    default:
#ifdef DEBUG
      cerr << "IOXSherry::doDrawText mode is out of range."
	   << " mode = " << mode << endl;
#endif
      break;
    }
}

void IOXSherry::doDrawBox (uint16 screen, uint8 planeMask,
			   uint16 x1, uint16 y1,
			   uint16 x2, uint16 y2,
			   uint8 foreground)
{
  setGC (gc, planeMask, foreground);
  
  const SherryArea sherryArea (x1, y1, x2, y2);
  virtualScreen[screen]->fillRectangle (gc,
					sherryArea.getLeft (),
					sherryArea.getTop (),
					sherryArea.getWidth (),
					sherryArea.getHeight ());
}

void IOXSherry::doDrawLine (uint16 screen, uint8 planeMask,
			    uint16 x1, uint16 y1,
			    uint16 x2, uint16 y2,
			    uint8 foreground)
{
  setGC (gc, planeMask, foreground);
  virtualScreen[screen]->drawLine (gc, x1, y1, x2, y2);
}

void IOXSherry::doCopyArea (uint16 fromScreen, uint16 toScreen,
			    uint8 planeMask, uint8 transparencyFlag,
			    uint16 x1, uint16 y1,
			    uint16 x2, uint16 y2,
			    uint16 toX, uint16 toY)
{
  setPlaneMask (gc, planeMask);
  SherryArea sherryArea (x1, y1, x2, y2);
#ifdef SHERRY_WRD_DEBUG
      cerr << "IOXSherry::doCopyArea"
	   << " fromScreen=" << fromScreen
	   << " toScreen=" << toScreen
	   << " planeMask=" << static_cast<int> (planeMask)
	   << " transparencyFlag=" << static_cast<int> (transparencyFlag)
	   << " x1=" << x1
	   << " y1=" << y1
	   << " x2=" << x2
	   << " y2=" << y2
	   << " toX=" << toX
	   << " toY=" << toY
	   << endl;
#endif
  if (transparencyFlag)
    {//$BF)2a=hM}$,I,MW(B
      //clipmask$B$HF)2a=hM}MQ$N%^%9%/$r:n@.$9$k!#(B
      WdkXImage fromImage (virtualScreen[fromScreen],
			   sherryArea.getLeft (), sherryArea.getTop (),
			   sherryArea.getWidth (), sherryArea.getHeight (),
			   0xffffffff, ZPixmap);
      WdkXImage ximage = WdkXImage (display, visual, 1, XYBitmap, 0,
				    sherryArea.getWidth (),
				    sherryArea.getHeight ());
      for (uint y = 0; y < sherryArea.getHeight (); y++)
	{
	  for (uint x = 0; x < sherryArea.getWidth (); x++)
	    {
	      if (fromImage.getPixel (x, y) != palette2Pixel[virtualScreenInfo[fromScreen]->transparentColor])
		{
		  ximage.putPixel (x, y, 1UL);
		}
	    }
	}
      
      //mask$B$r(Bbitmap$B$K%;%C%H$7$F(BGC$B$K@_Dj$9$k!#(B
      WdkPixmap clipMask (window,
			  sherryArea.getWidth (), sherryArea.getHeight (), 1);
      WdkGC clipGC (&clipMask);
      clipGC.setForeground (1UL);
      clipGC.setBackground (0UL);
      ximage.putImage (clipMask, clipGC, 0, 0, 0, 0,
		       sherryArea.getWidth (), sherryArea.getHeight ());
      gc->setClipMask (clipMask);
      gc->setClipOrigin (sherryArea.getLeft (), sherryArea.getTop ());

      //$B<B:]$K%3%T!<(B
      virtualScreen[fromScreen]->copyArea (virtualScreen[toScreen], gc,
					   sherryArea.getLeft (),
					   sherryArea.getTop (),
					   sherryArea.getWidth (),
					   sherryArea.getHeight (),
					   toX, toY);
      //$B8e;OKv(B
      gc->resetClipMask ();
    }
  else
    {//$BF)2a=hM}$OITMW(B
      virtualScreen[fromScreen]->copyArea (virtualScreen[toScreen], gc,
					   sherryArea.getLeft (),
					   sherryArea.getTop (),
					   sherryArea.getWidth (),
					   sherryArea.getHeight (),
					   toX, toY);
    }
}

void IOXSherry::doMaskCopy (uint16 fromScreen, uint16 toScreen,
			    uint8 planeMask, uint8 transparencyFlag,
			    uint16 x1, uint16 y1,
			    uint16 x2, uint16 y2,
			    uint16 toX, uint16 toY,
			    uint8 maskX, uint8 maskY,
			    const uint8* maskData, int arraySize)
{
  setPlaneMask (gc, planeMask);
  SherryArea sherryArea (x1, y1, x2, y2);
  //maskData$B$,(B1line$BEv$j2?%P%$%HM-$k$+5a$a$k!#(B
  //(width$B$r(B8$B$NG\?t$K@Z$j>e$2(B)/8 $B$,5a$a$kJ*!#(B
  const int bytesPerLine = sherryArea.getWidth ()/8
    + ((sherryArea.getWidth ()%8 == 0)? 0:1);
  
  
  //clipmask$B$HF)2a=hM}MQ$N%^%9%/$r:n@.$9$k!#(B
  WdkXImage ximage = WdkXImage (display, visual, 1, XYBitmap, 0,
				sherryArea.getWidth (),
				sherryArea.getHeight ());
  if (transparencyFlag)
    {//$BF)2a=hM}$,I,MW(B
      WdkXImage fromImage (virtualScreen[fromScreen],
			   sherryArea.getLeft (), sherryArea.getTop (),
			   sherryArea.getWidth (), sherryArea.getHeight (),
			   ~0UL, ZPixmap);
      for (uint y = 0; y < sherryArea.getHeight (); y++)
	{
	  for (uint x = 0; x < sherryArea.getWidth (); x++)
	    {
	      if (maskData[x/8+y*bytesPerLine] & (0x70 >> (x%8)) ||
		  fromImage.getPixel (x, y) != palette2Pixel[virtualScreenInfo[fromScreen]->transparentColor])
		{
		  ximage.putPixel (x, y, 1UL);
		}
	    }
	}
    }
  else
    {//$BF)2a=hM}$OITMW(B
      for (uint y = 0; y < sherryArea.getHeight (); y++)
	{
	  for (uint x = 0; x < sherryArea.getWidth (); x++)
	    {
	      if (maskData[x/8+y*bytesPerLine] & (0x70 >> (x%8)))
		{
		  ximage.putPixel (x, y, 1UL);
		}
	    }
	}
    }

  //mask$B$r(Bbitmap$B$K%;%C%H$7$F(BGC$B$K@_Dj$9$k!#(B
  WdkPixmap clipMask (window,
		      sherryArea.getWidth (), sherryArea.getHeight (), 1);
  ximage.putImage (clipMask, defaultGC, 0, 0, 0, 0,
		   sherryArea.getWidth (), sherryArea.getHeight ());
  gc->setClipMask (clipMask);
  gc->setClipOrigin (sherryArea.getLeft (), sherryArea.getTop ());

  //$B<B:]$K%3%T!<(B
  virtualScreen[fromScreen]->copyArea (virtualScreen[toScreen], gc,
				       sherryArea.getLeft (),
				       sherryArea.getTop (),
				       sherryArea.getWidth (),
				       sherryArea.getHeight (),
				       toX, toY);

  //$B8e;OKv(B
  gc->resetClipMask ();
}

void IOXSherry::doPlaneCopy (uint16 fromScreen, uint16 toScreen,
			     uint8 planeMask, uint8 transparencyFlag,
			     uint8 fromPlane, uint8 toPlane,
			     uint8 planeNumber,
			     uint16 x1, uint16 y1,
			     uint16 x2, uint16 y2,
			     uint16 toX, uint16 toY)
{
  SherryArea sherryArea (x1, y1, x2, y2);

  //copy$B85$N%$%a!<%8$r<hF@!#(B
  WdkXImage fromImage (virtualScreen[fromScreen],
		       sherryArea.getLeft (), sherryArea.getTop (),
		       sherryArea.getWidth (), sherryArea.getHeight (),
		       ~0UL, ZPixmap);
  
  //$BIA2h@h$N%$%a!<%8$r:n@.$9$k!#(B
  WdkXImage toImage (display, visual, depth, ZPixmap, 0,
		     sherryArea.getWidth (), sherryArea.getHeight ());
  for (uint y = 0; y < sherryArea.getWidth (); y++)
    {
      for (uint x = 0; x < sherryArea.getHeight (); x++)
	{
	  const unsigned long fromPixel = fromImage.getPixel (x, y);
	  unsigned long toPixel = 0UL;
	  for (int i = 0; i < planeNumber; i++)
	    {
	      if (fromPixel & planePixel[i+fromPlane])
		toPixel |= planePixel[i+toPlane];
	    }
	  toImage.putPixel (x, y, toPixel);
	}
    }
  
  //plane mask$B$N@_Dj(B
  uint8 copyPlaneMask = 0;
  for (int i = 0; i < planeNumber; i++)
    copyPlaneMask |= (1 << i+toPlane);
  setPlaneMask (gc, (planeMask & copyPlaneMask));

  if (transparencyFlag)
    {//$BF)2a=hM}(B
      //clip mask$B$r:n@.(B
      unsigned long clipMaskPixel = 0UL;
      for (int i = 0; i < planeNumber; i++)
	clipMaskPixel |= planePixel[i+fromPlane];
      WdkXImage clipXImage (display, visual, 1, XYBitmap, 0,
			    sherryArea.getWidth (), sherryArea.getHeight ());
      for (uint y = 0; y < sherryArea.getHeight (); y++)
	{
	  for (uint x = 0; x < sherryArea.getWidth (); x++)
	    {
	      if ((clipMaskPixel & fromImage.getPixel (x, y)) != 0)
		clipXImage.putPixel (x, y, 1UL);
	    }
	}
      WdkPixmap clipMask (window,
			  sherryArea.getWidth (), sherryArea.getHeight (), 1);
      clipXImage.putImage (clipMask, defaultGC, 0, 0, 0, 0,
			   sherryArea.getWidth (), sherryArea.getHeight ());
      
      //clilp mask$B$r(BdrawGC$B$K@_Dj!#(B
      gc->setClipMask (clipMask);
      gc->setClipOrigin (toX, toY);
    }

  toImage.putImage (virtualScreen[toScreen], gc,
		    0, 0, toX, toY, sherryArea.getWidth (),
		    sherryArea.getHeight ());

  //$B8e;OKv(B
  if (transparencyFlag)
    gc->resetClipMask ();
}

////////////////////////////////////////////////////////////event handling

//$BK\Ev$O(BNoExpose event$B$OM^@)$9$l$P(Bevent handle$B$9$kI,MW$OL5$$H&!#(B
void IOXSherry::handleEvent ()
{
  for (;;)
    {
      schedYield ();
      XEvent event;
      XNextEvent (display, &event);
      switch (event.type)
	{
	case Expose:
	  //do nothing
	  break;
	case ClientMessage:
	  DEBUG_MESSAGE ("IOXSherry::handleEvent");
	  if (event.xclient.message_type ==
	      XSHERRY_EVENT_HANDLER_LOOP_END_ATOM &&
              event.xclient.format == 8 &&
              strncmp (event.xclient.data.b,
		       XSHERRY_EVENT_HANDLER_LOOP_CANCEL_MESSAGE,
		       CLIENT_MESSAGE_LENGTH) == 0)
	    //pthread_exit (NULL);
	    return;
          else
            cerr << "IOXSherry::handleEvent: unknown"
		 <<" XClientMessageEvent recieved." << endl;
	  break;
	case NoExpose:
	  //NoExpose,GraphicExpose event$B$O<u$1<h$i$J$$$h$&$K$9$Y$-(B?
	  break;
	default:
	  cerr << "IOXSherry::handleEvent: unknown event recieved."
	       << " Is there error?" << endl;
	  cerr << "event.type=" << event.type << endl;
	  break;
	}
    }
}

void* IOXSherry::startEventHandleThread (void* me)
{
  static_cast<IOXSherry*>(me)->handleEvent ();
  return NULL;
}

void IOXSherry::createEventHandleThread ()
{
  XLockDisplay (display);
  assert (isEventHandleThreadCreated == false);
  PThreadAttribute attribute;
  pthread_create (&eventHandleThread, &attribute.getPThreadAttribute (),
		  &IOXSherry::startEventHandleThread,
		  static_cast<void*> (this));
  isEventHandleThreadCreated = true;
  XUnlockDisplay (display);
}


void IOXSherry::cancelEventHandleThread ()
{
  XLockDisplay (display);
  assert (isEventHandleThreadCreated == true);
  XUnlockDisplay (display);
  
  XEvent clientMessage;
  clientMessage.xany.type = ClientMessage;
  clientMessage.xany.display = display;
  clientMessage.xany.window = window->getWindow ();
  clientMessage.xclient.message_type = XSHERRY_EVENT_HANDLER_LOOP_END_ATOM;
  clientMessage.xclient.format = 8;
  memset (clientMessage.xclient.data.b, 0, CLIENT_MESSAGE_LENGTH);
  strncpy (clientMessage.xclient.data.b,
	   XSHERRY_EVENT_HANDLER_LOOP_CANCEL_MESSAGE,
	   CLIENT_MESSAGE_LENGTH);
  DEBUG_MESSAGE ("IOXSherry::cancelEventLoopThread:now sending Client"
		 "Message event.");
  XSendEvent (display, clientMessage.xany.window, True, ExposureMask,
	      &clientMessage);
  DEBUG_MESSAGE ("IOXSherry::cancelEventLoopThread now flushing");
  XFlush ();
  DEBUG_MESSAGE ("IOXSherry::cancelEventLoopThread now joining");
  if (pthread_join (eventHandleThread, NULL) != 0)
    cerr << "IOXSherry::cancelEventHandler:"
	 << " pthread_join returned non-zero status." << endl;
  DEBUG_MESSAGE ("IOXSherry::cancelEventLoopThread:joined");
  
  XLockDisplay (display);
  isEventHandleThreadCreated = false;
  XUnlockDisplay (display);
}



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////for dynamic loading
///////////////////////////////////////////////////////////////newIO<T>
///////////////////////////////////////////////////////////////newMainIO<T>
///////////////////////////////////////////////////////////////infoIO<T>
////////////////////////////////////////////////////////////////////////////

#include "IO/InfoIO.h"
static IO* newIOXSherry (State* state, WidiController* widiController,
                         Option* option)
{
  return new IOXSherry (state, widiController, option);
}

InfoIO infoIOXSherry=
{
  newIOXSherry,
  NULL,
  false
};
