/*
    IOXWrd.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cassert>
#include <vector>
#include <functional>
#include <algorithm>

#include "util/KanjiConvert.h"
#include "util/getABSTimeSpec.h"
#include "IOXWrd.h"
#include "WidiToolkit/WdkWindow.h"
#include "WidiToolkit/WdkXImage.h"
#include "PlugIn/Wrd/magfile.h"
#include "PlugIn/X/xutil/createFontSet.h"
#include "PlugIn/X/xutil/findVisual.h"

#ifndef HAVE_NANOSLEEP
# include "compat/nanosleep.h"
#endif // HAVE_NANOSLEEP

using namespace std;

const char* IOXWrd::defaultResource =
"Widi*XWrd*Fontset: "
"-adobe-courier-medium-r-normal--16-110-100-0-m-*-iso8859-*,"
"-jis-fixed-medium-r-normal--16-110-100-100-*-*-jisx0208.1983-*,"
"-misc-fixed-medium-r-normal--16-110-100-*-*-*-jisx0201.1976-*\n"
"Widi*XWrd*TextColor0: Black\n"
"Widi*XWrd*TextColor1: Red\n"
"Widi*XWrd*TextColor2: Green\n"
"Widi*XWrd*TextColor3: Yellow\n"
"Widi*XWrd*TextColor4: Blue\n"
"Widi*XWrd*TextColor5: Magenta\n"
"Widi*XWrd*TextColor6: Cyan\n"
"Widi*XWrd*TextColor7: White\n";
const char IOXWrd::resourceFileName[] = WIDI_DATA_DIR"/WidiXWrd.ad";
const char* IOXWrd::className = "Widi.XWrd";
const char* IOXWrd::resourceName = "widi.xwrd";
const int IOXWrd::color2PixelIndex[47+1] =
{
  7,7,//0,1
  0,0,
  7,7,0,7,0,//4,5,6,7,8
  0,0,0,0,0,0,0,
  0,1,4,6,2,3,5,7,//16,17,18,19,20,21,22,23
  0,0,0,0,0,0,
  0,1,2,3,4,5,6,7,//30,31,32,33,34,35,36,37
  0,0,
  0,1,2,3,4,5,6,7,//40,41,42,43,44,45,46.47
};
const int IOXWrd::MAX_PLANE_NUM;

IOXWrd::IOXWrd (State* state__, WidiController* widiController__, 
		Option* option__):
  XBase (), Wrd (state__, widiController__, option__),
  window (NULL)
{
  //do nothing
}

IOXWrd::~IOXWrd ()
{
  //do nothing
}

void IOXWrd::concatinate (char* dest, const char* first, const char* second)
{
  assert (dest != NULL);
  assert (first != NULL);
  assert (second != NULL);
  
  strcpy (dest, first);
  strcat (dest, ".");
  strcat (dest, second);
}

char* IOXWrd::getResource (XrmDatabase privateRdb, XrmDatabase displayRdb,
			   const char* className_, const char* classResource_,
			   const char* name_, const char* resourceName_)
{
  assert (className_ != NULL);
  assert (classResource_ != NULL);
  assert (name_ != NULL);
  assert (resourceName_ != NULL);
  
  char* const fullClassName = new char[strlen (className_)+1+strlen (classResource_)+1];
  concatinate (fullClassName, className_, classResource_);
  char* const fullName = new char[strlen (name_)+1+strlen (resourceName_)+1];
  concatinate (fullName, name_, resourceName_);

  char* retType;
  XrmValue retValue;
  int result = (displayRdb !=NULL &&
		XrmGetResource (displayRdb, fullClassName, fullName,
				&retType, &retValue));
  if (!result)
    result = XrmGetResource (privateRdb, fullClassName, fullName,
			     &retType, &retValue);

  delete[] fullClassName;
  delete[] fullName;
  
  if (!result)
    return NULL;
  return retValue.addr;
}


void IOXWrd::setupResource (Display* display_, const char* resourceFileName_,
			    const char* defaultResource_,
			    const char* className_, const char* resourceName_,
			    char* colorName_[MAX_TEXT_COLORS],
			    XFontSet* fontSet_)
{
  assert (resourceFileName_ != NULL);
  assert (defaultResource_ != NULL);
  assert (className_ != NULL);
  assert (resourceName_ != NULL);
  assert (fontSet_ != NULL);
  
  //display resource
  const char* str = XResourceManagerString (display_);
  if (str == NULL)
    displayRdb = NULL;
  else
    displayRdb = XrmGetStringDatabase (str);
  
  //this app's defualt resource
  privateRdb = XrmGetFileDatabase (resourceFileName_);
  XrmCombineDatabase (XrmGetStringDatabase (defaultResource_), &privateRdb,
		      False);
  

  //get color name
  char baseClassName[] = "TextColor0";
  const int classNameLength = strlen (baseClassName);
  char baseName[] = "textColor0";
  const int nameLength = strlen (baseName);
  for (int i = 0; i < MAX_TEXT_COLORS; i++)
    {
      baseClassName[classNameLength-1] = '0'+i;
      baseName[nameLength-1] = '0'+i;
      colorName_[i] = getResource (privateRdb, displayRdb,
				   className_, baseClassName,
				   resourceName_, baseName);
    }
  
  //get font struct
  const char* fontName = getResource (privateRdb, displayRdb,
				      className_, "Fontset",
				      resourceName_, "fontset");
  *fontSet_ = createFontSet (display, fontName);
}



void IOXWrd::allocateColors (Display* display_, Colormap colormap_, 
			     WdkColor tcolor_[MAX_TEXT_COLORS], 
			     char* colorName_[MAX_TEXT_COLORS],
			     ulong* basePixel_,
			     ulong planePixel_[MAX_PLANE_NUM],
			     ulong graphicPixel_[MAX_GRAPHIC_COLORS])
{
  assert (colorName_ != NULL);
  
  //textΥ顼ݡ
  for (int i = 0; i < MAX_TEXT_COLORS; i++)
    {
      try
	{
	  tcolor_[i].allocNamed2 (display_, colormap_, colorName_[i]);
	}
      catch (WdkColor::CanNotAllocate)
	{//γݤ˼
	  for (int j = i-1; j >= 0; j--)
	    tcolor_[j].free (display, colormap);
	  throw CanNotAllocateColors ();
	}
    }
  
  //color cellγݡ
  if (!XAllocColorCells (display_, colormap_, False,
			 planePixel_, MAX_PLANE_NUM,
			 basePixel_, 1))
    {
      for (int j = 0; j < MAX_TEXT_COLORS; j++)
	tcolor_[j].free (display, colormap);
      throw CanNotAllocateColors ();
    }

  for (int i = 0; i < MAX_GRAPHIC_COLORS; i++)
    {
      graphicPixel_[i] = *basePixel_;
      for (int j = 0; j < MAX_GRAPHIC_DEPTH; j++)
	{
	  if ((i & (1 << j)) != 0) graphicPixel_[i] |= planePixel_[j];
	}
    }
}

void IOXWrd::handleXEvent ()
{
  XEvent xevent;
  //using XCheckWindowEvent cause strange results why???
  //if (XCheckWindowEvent (display, window->getWindow (),
  //ExposureMask, &xevent) == True)
  if (XCheckTypedWindowEvent (display, window->getWindow (),
			      Expose, &xevent) != 0)
    {
      int left = xevent.xexpose.x;
      int top = xevent.xexpose.y;
      int right = left + xevent.xexpose.width-1;
      int bottom = top + xevent.xexpose.height-1;
      assert (xevent.type == Expose);
      //while (XCheckWindowEvent (display, window->getWindow (),
      //ExposureMask, &xevent) == True)
      while (XCheckTypedWindowEvent (display, window->getWindow (),
				     Expose, &xevent) != 0)
	{
	  assert (xevent.type == Expose);
	  left = min (xevent.xexpose.x, left);
	  top = min (xevent.xexpose.y, top);
	  right = max (xevent.xexpose.x+xevent.xexpose.width-1, right);
	  bottom = max (xevent.xexpose.y+xevent.xexpose.height-1, bottom);
	}

      redrawText (left, top, right-left+1, bottom-top+1, textScreen);
      //redrawText (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, textScreen);
    }
}


void IOXWrd::checkPause ()
{
  if (cacheState.songStatus.state==SongStatus::PAUSING)
    {//pauseʤΤǥݡޤԤġ
      DEBUG_MESSAGE ("IO::checkPause pausing");
      state->lock ();
      while (state->songStatus.state==SongStatus::PAUSING)
	{
	  state->unlock ();
	  handleXEvent ();
	  timespec timeSpec;
	  timeSpec.tv_sec = 0;
	  timeSpec.tv_nsec = 100*1000*1000;//0.1 sec
	  nanosleep (&timeSpec, NULL);

	  //0.5xevent롣
	  getABSTimeSpec (&timeSpec, 0, 500*100*100);
	  state->lock ();
	  state->timedWaitChange (&timeSpec);
	}
      state->unlock ();
      DEBUG_MESSAGE ("IO::checkPause all done");
    }

  handleXEvent ();
}

/////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////ɽ
/////////////////////////////////////////////////////////////////////
void IOXWrd::initialize ()
{
  Wrd::initialize ();

  //connect display
  display = XOpenDisplay (NULL);//XXX
  if (display == NULL)
    {//XXX
      cerr << "IOXWrd::initialize can not connect to the display." << endl;
    }
  screenNumber = DefaultScreen (display);

  setupResource (display, resourceFileName, defaultResource,
		 className, resourceName, colorName, &fontSet);
  
  //visualγ
  findVisual (display, screenNumber, &visual, &visualClass, &depth);
  
  //window
  window = new WdkWindow (display, NULL,
			  0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, 0, depth, visual);
    XSizeHints* sizeHints = XAllocSizeHints ();
  if (sizeHints != NULL)
    {
      sizeHints->flags = PMaxSize | PMinSize | PResizeInc | PBaseSize;
      sizeHints->max_width = GRAPHIC_SIZE_X;
      sizeHints->max_height = GRAPHIC_SIZE_Y;
      sizeHints->min_width = GRAPHIC_SIZE_X;
      sizeHints->min_height = GRAPHIC_SIZE_Y;
      sizeHints->width_inc = 0;
      sizeHints->height_inc = 0;
      sizeHints->base_width = GRAPHIC_SIZE_X;
      sizeHints->base_height = GRAPHIC_SIZE_Y;
      window->setWMNormalHints (sizeHints);
      XFree (sizeHints);
    }
  else
    {
      cerr << "IOXSherry::allocateXResource:XAllocSizeHints can not allocate."
	   << endl;
    }
  //pixmap
  pixmap[0] = new WdkPixmap (window, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  pixmap[1] = new WdkPixmap (window, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  
  //colormap, colorcellγ
  colormap = DefaultColormap (display, screenNumber);
  try
    {
      allocateColors (display, colormap, tcolor, colorName, &basePixel, planePixel, graphicPixel);
    }
  catch (CanNotAllocateColors)
    {//default colorǿγݤ˼ԡ
      cerr << "Can not allocate color.trying to create private colormap" << endl;
      colormap=XCreateColormap (display, window->getWindow (), visual, AllocNone);
      try
	{
	  allocateColors (display, colormap, tcolor, colorName, &basePixel, planePixel, graphicPixel);
	}
      catch (CanNotAllocateColors)
	{
	  //XXXҤ
	  throw IError ("IOXWrd::initliaze Can not allocate colors");
	}
    }
  window->setColormap (colormap);

  gc = new WdkGC (window);
  copyGc = new WdkGC (window);
  gc->setForeground (tcolor[0]);
  gc->setBackground (tcolor[0]);
  window->setBackground (tcolor[0]);
  window->clear ();
  pixmap[0]->fillRectangle (gc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  pixmap[1]->fillRectangle (gc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  
  //windowmap
  window->selectInput (ExposureMask);//XXX
  window->map ();

  //map줿Τǧ
  XFlush ();
  XEvent xevent;
  XWindowEvent (display, window->getWindow (), ExposureMask, &xevent);
}

void IOXWrd::release ()
{
  if (colormap!=DefaultColormap (display, screenNumber))
    {
      for (int i=0; i < MAX_TEXT_COLORS; i++)
	tcolor[i].free (display, colormap);
    }
  XFreeColors (display, colormap, graphicPixel, MAX_GRAPHIC_COLORS, 0);
  XFreeColormap (display,colormap);
  //window->unmap ();
  delete window;

  delete gc;
  delete copyGc;
  delete[] pixmap;
  delete[] virtualScreen;
  virtualScreen = NULL;
  nOfVirtualScreen = 0;
  XFreeFontSet (display,fontSet);

  XrmDestroyDatabase (displayRdb);
  XrmDestroyDatabase (privateRdb);
  XCloseDisplay (display);
  
  Wrd::release ();
}


void IOXWrd::prePlay ()
{
  gc->setPlaneMask (~0UL);
  gc->setForeground (tcolor[0]);
  gc->setBackground (tcolor[0]);
  window->clear ();
  pixmap[0]->fillRectangle (gc,0,0,GRAPHIC_SIZE_X,GRAPHIC_SIZE_Y);
  pixmap[1]->fillRectangle (gc,0,0,GRAPHIC_SIZE_X,GRAPHIC_SIZE_Y);

  delete[] virtualScreen;
  virtualScreen = NULL;
  nOfVirtualScreen = 0;


  Wrd::prePlay ();
  if (wrdFile == NULL) return;
  
  //window pixmapν
  const char* prefix = "Widi:XWrd:";
  char* name = new char[strlen (prefix)+strlen (wrdFile->getFileName ())+1];
  strcpy (name, prefix);
  strcat (name, wrdFile->getFileName ());
  window->storeName (name);
  window->setIconName (name);
  
  XFlush ();
}

void IOXWrd::postPlay ()
{
  //XXX

  
  Wrd::postPlay ();
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////̿μ¹
void IOXWrd::execNextLine ()
{
  //do nothing
}

void IOXWrd::execATMARK ()
{
  //do nothing
}

void IOXWrd::execCOLOR (uint32 color_)
{
  //do nothing
}

void IOXWrd::execEND ()
{
  //do nothing
}

void IOXWrd::execEXEC (const char* filename)
{
  //do nothing
}

void IOXWrd::doFading (int fadeCount_, uint32 fromPalette_, uint32 toPalette_)
{
  realUpdatePal ();
  XFlush ();
}


void IOXWrd::execFADE (uint32 p1, uint32 p2, int speed)
{
  assert (speed >= 0);
  
  realUpdatePal ();
  XFlush ();
}

void IOXWrd::execGCIRCLE (uint32 x, uint32 y, uint r, uint32 p1, uint32 sw, uint32 p2)
{
  assert (0 <= x);
  assert (x < static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= y);
  assert (y < static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (0 < r);
  assert (0 <= p1);
  assert (p1 < static_cast<uint32> (MAX_GRAPHIC_COLORS));
  assert (sw == 0 || sw == 1 || sw == 2);
  assert (0 <= p2);
  assert (p2 < static_cast<uint32> (MAX_GRAPHIC_COLORS));

  const int left = x - r;
  const int top = y - r;
  const uint width = 2*r;
  const uint height = 2*r;
  
  if (sw == 2)
    {
      gc->setForeground (graphicPixel[p2]);
      pixmap[activePage]->fillArc (gc, left, top, width, height, 0, 64*360);
    }

  gc->setForeground (graphicPixel[p1]);
  pixmap[activePage]->drawArc (gc, left, top, width, height, 0, 64*360);

  checkCopyWindow (left, top, width, height);
}

void IOXWrd::execGCLS (uint32 sw)
{
  assert (0 <= sw);
  assert (sw <= 15);

  WdkGC drawGc (window);
  drawGc.copy (gc, GCPlaneMask);
  drawGc.setForeground (0UL);
  if (option->xwrdDoesPlaneMask)
    {
      ulong mask = basePixel;
      if (!(sw == 0 || sw == 15))
	{
	  mask = ~0UL;
	}
      else
	{
	  if (sw & 1) mask |= planePixel[0];
	  if (sw & 2) mask |= planePixel[1];
	  if (sw & 4) mask |= planePixel[2];
	  if (sw & 8) mask |= planePixel[3];
	}
      drawGc.setPlaneMask (mask);
    }
  pixmap[activePage]->fillRectangle (drawGc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  drawGc.setPlaneMask (~0UL);

  checkCopyWindow (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
}

void IOXWrd::execGINIT ()
{
  //init grrphic
  gc->setForeground (tcolor[0]);
  gc->setPlaneMask (~0UL);
  pixmap[0]->fillRectangle (gc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  pixmap[1]->fillRectangle (gc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
  window->clear ();

  redrawText (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, textScreen);
  
  XFlush ();
}

void IOXWrd::execGLINE
(uint32 x1, uint32 y1, uint32 x2, uint32 y2, uint32 p1, uint32 sw, uint32 p2)
{
  //XXX
  //(p1 == 0 || p1 ==1)λ p2line style
  //ɤˤʤäƤΤʤΤǼ괺̵뤹롣
  assert (0 <= x1);
  assert (x1 <= static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= y1);
  assert (y1 <= static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (0 <= x2);
  assert (x2 <= static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= y2);
  assert (y2 <= static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (0 <= p1);
  assert (p1 < static_cast<uint32> (MAX_GRAPHIC_COLORS));
  assert (sw == 0 || sw == 1 || sw == 2);
  assert (0 <= p2);
  assert (p2 < static_cast<uint32> (MAX_GRAPHIC_COLORS));
  
  const int left = min (x1, x2);
  const int top = min (y1, y2);
  const int right = max (x1,x2);
  const int bottom = max (y1, y2);
  const int width = right-left+1;
  const int height = bottom-top+1;
  switch (sw)
    {
    case 0:
      gc->setForeground (graphicPixel[p1]);
      pixmap[activePage]->drawLine (gc, x1, y1, x2, y2);
      break;
    case 2:
      gc->setForeground (graphicPixel[p2]);
      pixmap[activePage]->fillRectangle (gc, left, top, width, height);
      //falling throgh
    case 1:
      gc->setForeground (graphicPixel[p1]);
      pixmap[activePage]->drawLine (gc, left,  top,    left,  bottom);
      pixmap[activePage]->drawLine (gc, left,  top,    right, top);
      pixmap[activePage]->drawLine (gc, right, top,    right, bottom);
      pixmap[activePage]->drawLine (gc, left,  bottom, right, bottom);
      break;
    default:
      cerr << "ERROR:IOXWrd::execGLINE sw is invalied!!!" << endl;
      break;
    }


  checkCopyWindow (left, top, width, height);
}

void IOXWrd::execGMODE (uint32 sw)
{//gcplane maskѹƤΤϤʳGC򥳥ԡƻȤȡ
  assert (0 <= sw);
  assert (sw < static_cast<uint32> (MAX_GRAPHIC_COLORS));

  if (option->xwrdDoesPlaneMask)
    {
#if 1
      ulong mask = basePixel;
      if (sw & 1) mask |= planePixel[0];
      if (sw & 2) mask |= planePixel[1];
      if (sw & 4) mask |= planePixel[2];
      if (sw & 8) mask |= planePixel[3];
      gc->setPlaneMask (mask);
#else
      ulong mask = ~0UL;
      if ((sw & 1) != 0) mask &= ~planePixel[0];
      if ((sw & 2) != 0) mask &= ~planePixel[1];
      if ((sw & 4) != 0) mask &= ~planePixel[2];
      if ((sw & 8) != 0) mask &= ~planePixel[3];
      gc->setPlaneMask (mask);
#endif
    }
}

void IOXWrd::execGMOVE (uint32 x1, uint32 y1, uint32 x2, uint32 y2, uint32 xd, uint32 yd, uint32 vs, uint32 vd, uint32 sw)
{
  assert (0 <= x1);
  assert (x1 < static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= y1);
  assert (y1 < static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (0 <= x2);
  assert (x2 < static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= y2);
  assert (y2 < static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (vs == 0 || vs == 1);
  assert (vd == 0 || vd == 1);
  assert (sw == 0 || sw == 1);
  
  //8dotñ̤Ǵݤ롣???
  //x1 &= 0xfff8;
  //x1 = (x1 & 0xfff8) + ((x1 & 0x7)? 8: 0);
  //x2 = (x2 & 0xfff8) + ((x2 & 0x7)? 8: 0);
  //xd = (xd & 0xfff8) + ((xd & 0x7)? 8: 0);
  x1 |= ((x1 & 0x7)? 0x7: 0);
  x2 |= ((x2 & 0x7)? 0x7: 0);
  xd |= ((xd & 0x7)? 0x7: 0);

  
  const uint32 left = min (x1, x2);
  const uint32 top = min (y1, y2);
  const uint32 right = max (x1,x2);
  const uint32 bottom = max (y1, y2);
  assert (right-left+1 < static_cast<uint32>(GRAPHIC_SIZE_X)+8);
  const uint32 width = min (right-left+1, static_cast<uint32> (GRAPHIC_SIZE_X));
  const uint32 height = bottom-top+1;

  assert (0 <= xd);
  assert (xd < static_cast<uint32> (GRAPHIC_SIZE_X));
  assert (0 <= yd);
  assert (yd < static_cast<uint32> (GRAPHIC_SIZE_Y));
  assert (xd+width <= static_cast<uint32> (GRAPHIC_SIZE_X)+8);
  assert (yd+height <= static_cast<uint32> (GRAPHIC_SIZE_Y)+8);
  
  switch (sw)
    {
    case 0:
      pixmap[vs]->copyArea (pixmap[vd], gc, left, top, width, height, xd, yd);
      if (vd == displayPage)
	{
	  pixmap[displayPage]->copyArea (window, copyGc, xd, yd, width, height,
					 xd, yd);
	  redrawText (xd, yd, width, height, textScreen);
	  XFlush ();
	}
      break;
    case 1:
      {
	WdkPixmap tmp (pixmap[vd], width, height);
	pixmap[vd]->copyArea (&tmp, copyGc, xd, yd, width, height, 0, 0);
	pixmap[vs]->copyArea (pixmap[vd], gc, left, top, width, height, xd, yd);
	tmp.copyArea (pixmap[vd], gc, 0, 0, width, height, left, top);
	if (vs == displayPage)
	  {
	    pixmap[displayPage]->copyArea (window, copyGc,
					   left, top, width, height,left, top);
	    redrawText (left, top, width, height, textScreen);
	  }
	if (vd == displayPage)
	  {
	    pixmap[displayPage]->copyArea (window, copyGc,
					   xd, yd, width, height, xd, yd);
	    redrawText (xd,   yd,  width, height, textScreen);
	  }
	if (vs == displayPage || vd == displayPage) XFlush ();
	break;
      }
    default:
      cerr << "ERROR:IOXWrd::execGMOVE sw is invalied !!!" << endl;
      break;
    }
}

void IOXWrd::execGON (bool oldIsGraphicOn_, bool isGraphicOn_)
{
  if (isGraphicOn_ == oldIsGraphicOn_)
    return;//ѹʤä
  
  if (isGraphicOn_)
    pixmap[displayPage]->copyArea (window, copyGc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, 0, 0);
  else
    window->clear ();//graphic off

  redrawText (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, textScreen);
  XFlush ();
}

void IOXWrd::execGSCREEN (uint32 oldActivePage_, uint32 oldDisplayPage_, uint32 activePage_,uint32 displayPage_)
{
  if (displayPage_ == oldDisplayPage_)
    return;//ɬ̵
  
  pixmap[displayPage_]->copyArea (window, copyGc, 0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, 0, 0);
  redrawText (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y, textScreen);
  XFlush ();
}

void IOXWrd::execINKEY (uint32 barWaitFor_)
{
  //XXX
}

void IOXWrd::execLOCATE (uint32 x_, uint32 y_)
{
  //do nothing
}


void IOXWrd::execLOOP (uint32 loop)
{
  //XXX
}


void IOXWrd::execMAG
(MagFile& mag, int32 x, int32 y, uint32 s, uint32 p)
{
  assert (0 <= x);
  assert (x < GRAPHIC_SIZE_X);
  assert (0 <= y);
  assert (y < GRAPHIC_SIZE_Y);
  assert (s == 0 || s == 1);
  assert (p == 0 || p == 1 || p == 2);
  
  if (p == 0 || p == 2)//paletteѹ
    realUpdatePal ();

  if (p == 0 || p == 1)
    {//graphicɽ
      const int width=mag.width/s;
      int height=mag.height/s;
      if (y+height > GRAPHIC_SIZE_Y)
	height = GRAPHIC_SIZE_Y - y;//¦ˤϤ߽Ф
      
      //magdataximage
      WdkXImage ximage (display, visual, depth, ZPixmap, 0, width, height);
      for(int j = 0; j < height; j++)
	{
	  for(int i = 0; i < width; i++)
	    {
	      ximage.putPixel (i, j,
			       graphicPixel[mag.Screen[i*s + j*s*mag.width]]);
	    }
	}

      if (x+width > GRAPHIC_SIZE_X)
	{//¦ˤϤ߽ФΤǡ¦ޤ֤
	  const int rightWidth = GRAPHIC_SIZE_X-x;
	  const int leftWidth = width-rightWidth;
	  int leftHeight = height;
	  if (y+leftHeight == GRAPHIC_SIZE_Y) leftHeight -= 1;
	  
	  ximage.putImage (pixmap[activePage], copyGc,
			    0, 0, x, y, rightWidth, height);
	  ximage.putImage (pixmap[activePage], copyGc,
			    rightWidth, 0, 0, y+1, leftWidth, leftHeight);
	}
      else
	{
	  ximage.putImage (pixmap[activePage], copyGc,
			    0, 0, x, y, width, height);
	  checkCopyWindow (x, y, width, height);
	}
    }
  
  XFlush ();
}

void IOXWrd::execOFFSET (uint32 offsetBarWaitFor_)
{
  //do nothing
}

void IOXWrd::realUpdatePal ()
{
  XColor xcolor[MAX_GRAPHIC_COLORS];
  for (int i = 0; i < MAX_GRAPHIC_COLORS; i++)
    {//NOTE .red,.green,.blue16bit,12=16-4
      xcolor[i].pixel = graphicPixel[i];
#if 0
      xcolor[i].red   = static_cast<unsigned short>(palette[0].rgb[i].getR ()) << 12;
      xcolor[i].green = static_cast<unsigned short>(palette[0].rgb[i].getG ()) << 12;
      xcolor[i].blue  = static_cast<unsigned short>(palette[0].rgb[i].getB ()) << 12;
#elif 0
      //NOTE
      // 0<= xcolor value <= 65535
      // 0<= mag color value <= 15
      // 65535 = 15*3855
      xcolor[i].red   =
	static_cast<unsigned short>(palette[0].rgb[i].getR ()) * 3855;
      xcolor[i].green =
	static_cast<unsigned short>(palette[0].rgb[i].getG ()) * 3855;
      xcolor[i].blue  =
	static_cast<unsigned short>(palette[0].rgb[i].getB ()) * 3855;
#else
      //NOTE xcolor value 16 bit <= 65536
      //mag color value 4 bit <= 15
      const unsigned short r = palette[0].rgb[i].getR ();
      const unsigned short g = palette[0].rgb[i].getG ();
      const unsigned short b = palette[0].rgb[i].getB ();
      xcolor[i].red   = (r << 12) | (r << 8) | (r << 4) | r;
      xcolor[i].green = (g << 12) | (g << 8) | (g << 4) | g;
      xcolor[i].blue  = (b << 12) | (b << 8) | (b << 4) | b;
#endif
      xcolor[i].flags = DoRed | DoGreen | DoBlue;
    }
  XStoreColors (display, colormap, xcolor, MAX_GRAPHIC_COLORS);

}


void IOXWrd::updatePal (uint32 page)
{
  if (page == 0)
    {
      realUpdatePal ();
      XFlush ();
    }
}

void IOXWrd::execPAL (uint32 page, int32 rgb[16])
{
  updatePal (page);
}

void IOXWrd::execPALREV (uint32 p)
{
  updatePal (p);
}

void IOXWrd::execPLOAD (PH0* ph0)
{
  assert (ph0 != NULL);
  
  const uchar* const b=ph0->getB ();
  const uchar* const r=ph0->getB ();
  const uchar* const g=ph0->getB ();
  const uchar* const e=ph0->getB ();
  for(int y_ = 0; y_ < 400; y_++)
    {
      for(int x_ = 0; x_ < 640; x_++)
	{
	  unsigned short c = 
	    ((((b[(y_*640+x_)/8] & (1<<(7-x_%8))) != 0) << 3) |
	     (((r[(y_*640+x_)/8] & (1<<(7-x_%8))) != 0) << 2) |
	     (((g[(y_*640+x_)/8] & (1<<(7-x_%8))) != 0) << 1) |
	     ((e == NULL) ? 0 :
	      ((e[(y_*640+x_)/8] & (1<<(7-x_%8))) != 0)));
	  copyGc->setForeground (graphicPixel[c]);
	  pixmap[activePage]->drawPoint (copyGc, x_, y_);
	}
    }

  //redrawText
  checkCopyWindow (0, 0, GRAPHIC_SIZE_X, GRAPHIC_SIZE_Y);
}

void IOXWrd::execREMARK ()
{
  //do nothing
}

void IOXWrd::execREST (uint32 barOffsetWaitFor, uint32 stepsWaitFor)
{
  //do nothing
}

void IOXWrd::execSCROLL (uint32 x1, uint32 y1, uint32 x2, uint32 y2, uint32 mode, uint32 color, uint32 chr)
{
  //XXX
}

void IOXWrd::execSTARTUP (uint32 versoin_)
{
  //do nothing
}

void IOXWrd::execSTOP ()
{
  //XXX
}

void IOXWrd::execTCLS
(uint32 x1, uint32 y1, uint32 x2, uint32 y2, uint32 color, uint32 chr)
{
  //do nothing
}

void IOXWrd::execTON (uint32 textMode_)
{
  //XXX
}

void IOXWrd::execWAIT (uint32 barWaitFor_, uint32 steptsWaitFor_)
{
  //do nothing
}

void IOXWrd::execWMODE (uint32 deltaSteps_, uint32 wmode_)
{
  //do nothing
}

void IOXWrd::execFONTM (int32 fontMode_)
{
  //XXX
}

void IOXWrd::execFONTP (uint32 num, uint32 x_, uint32 y_, uint32 attribute)
{
  //XXX
}

void IOXWrd::execFONTR (int32 fontNumber, int32 fontData[16])
{
  //XXX
}

void IOXWrd::execGSC (int32 mode)
{
  //XXX
}

void IOXWrd::execLINE (int32 mode)
{
  //XXX
  
  //redrawText
}

void IOXWrd::execPAL_ (int32 paletteNumber, int32 rgb)
{
  //XXX
}

void IOXWrd::execREGSAVE (int32 num, int32 value[WRD_REGSAVE_SIZE_MAX])
{
  //XXX
}

void IOXWrd::execSCROLL_ (int32 x_, int32 y_)
{
  //XXX
}

void IOXWrd::execTEXTDOT (int32 textDot_)
{
  //XXX
}

void IOXWrd::execTMODE (int32 tmode_)
{
  //XXX
}

void IOXWrd::execTSCRL (int32 lines)
{
  //XXX
}

void IOXWrd::execVCOPY (int32 sx1, int32 sy1, int32 sx2, int32 sy2, int32 tx, int32 ty, int32 ss, int32 ts, int32 mode)
{
  //XXX
}

void IOXWrd::execVSGET (int32 page[4])
{
  //XXX
}
//XXX 4Ƚ񤯤ΤϷʤ
void IOXWrd::execVSRES ()
{
  //XXX
}

void IOXWrd::execXCOPY (int32 sx1, int32 sy1, int32 sx2, int32 sy2, int32 tx, int32 ty, int32 ss, int32 ts, int32 method, int32 opt[5])
{
  //XXX
}

void IOXWrd::printChar (uint x_, uint y_, char c, int color_)
{
  assert (1 <= x_);
  assert (x_ < TEXT_SIZE_X);
  assert (1 <= y_);
  assert (y_ <= TEXT_SIZE_Y);
  assert (isiso (c));
  TextScreen::checkColor (color_);
  //do nothing
}

void IOXWrd::printChar2b (uint x_, uint y_, char c1, char c2, int color_)
{
  assert (1 <= x_);
  assert (x_ < TEXT_SIZE_X);
  assert (1 <= y_);
  assert (y_ <= TEXT_SIZE_Y);
  assert (iskanji (c1) || iskana (c1));
  assert (TextScreen::checkColor (color_));
  //do nothing
}

void IOXWrd::doEscapeSequence ()
{
  //do nothing
}

void IOXWrd::doRefreshTextScreen (TextScreen& textScreen_)
{
  if (textScreen_.getTextMode () == TextScreen::TEXT_MODE_OFF)
    return;
  drawText (textScreen_, false, 1, 1, TEXT_SIZE_X, TEXT_SIZE_Y, isGraphicOn);
  XFlush ();
}


void IOXWrd::checkCopyWindow (int x_, int y_, int width, int height)
{
  assert (0 <= x_);
  assert (x_ < GRAPHIC_SIZE_X);
  assert (0 <= y_);
  assert (y_ < GRAPHIC_SIZE_Y);
  assert (0 <= width);
  assert (0 <= height);
  assert (x_ + width <= GRAPHIC_SIZE_X);
  assert (y_ + height <= GRAPHIC_SIZE_Y);
  
  if (activePage == displayPage)
    {
      pixmap[displayPage]->copyArea (window, copyGc, x_, y_, width, height, x_, y_);
      redrawText (x_, y_, width, height, textScreen);
      XFlush ();
    }
}


void IOXWrd::redrawText (int x_, int y_, int width, int height, 
			 TextScreen& textScreen_)
{
  assert (0 <= x_);
  assert (x_ < GRAPHIC_SIZE_X);
  assert (0 <= y_);
  assert (y_ < GRAPHIC_SIZE_Y);
  assert (0 <= width);
  assert (0 <= height);
  assert (x_ + width <= GRAPHIC_SIZE_X);
  assert (y_ + height <= GRAPHIC_SIZE_Y);
  
  if (textScreen_.getTextMode () == TextScreen::TEXT_MODE_OFF)
    return;
  
  //ɤtextľΤ롣
  const int leftText = x_/CHAR_SIZE_X+1;
  const int rightText = (x_ + width + CHAR_SIZE_X-1)/CHAR_SIZE_X;
  const int topText = y_/CHAR_SIZE_Y+1;
  const int downText = (y_ + height + CHAR_SIZE_Y-1)/CHAR_SIZE_Y;
  assert (1 <= leftText);
  assert (rightText <= static_cast<int> (TEXT_SIZE_X));
  assert (1 <= topText);
  assert (downText <= static_cast <int> (TEXT_SIZE_Y));

  if (isGraphicOn)
    pixmap [displayPage]->copyArea (window, copyGc,
				    x_, y_, width, height, x_, y_);
  //
  drawText (textScreen_, true, leftText, topText, rightText, downText, false);
}

void IOXWrd::drawText (TextScreen& textScreen_, bool doesForceRedraw = false, 
		       uint leftText = 1, uint topText = 1, 
		       uint rightText = TEXT_SIZE_X, 
		       uint downText = TEXT_SIZE_Y,
		       bool doesNeedRestoreBackground = true)
{
  assert (1 <= leftText);
  assert (leftText <= rightText);
  assert (rightText <= TEXT_SIZE_X);
  assert (1 <= topText);
  assert (topText <= downText);
  assert (downText <= TEXT_SIZE_Y);

  for (uint y_ = topText; y_ <= downText; y_++)
    {
      for (uint x_ = leftText; x_ <= rightText; x_++)
	{
	  const TextScreen::OneChar& oneChar=textScreen_.getText (x_, y_);
	  if (doesForceRedraw == false && oneChar.isChanged () == false)
	    continue;//ľɬפ̵ä

	  switch (oneChar.getCharSet ())
	    {
	    case TextScreen::CHARSET_ISO://iso
	      //ʤ' '̤˽ľ礬ФƤ
	      //¹ǽ꤬Ф롣
	      if (doesForceRedraw == true &&
		  doesNeedRestoreBackground == false &&
		  oneChar.getChar () == ' ' &&
		  isReversed (oneChar.getColor ()) == false)
		break;
	      
	      assert (isiso (oneChar.getChar ()));
	      char str[1];
	      str[0]=oneChar.getChar ();
	      drawString (x_,y_,str,1,
			  getTColor (oneChar.getColor ()),
			  isReversed (oneChar.getColor ()),
			  doesNeedRestoreBackground);
	      break;
	    case TextScreen::CHARSET_KANJI1:
	      {//EUC
		assert (x_ <= TEXT_SIZE_X-1);//2byteܤ뤫顣
		const TextScreen::OneChar& oneChar2 = textScreen_.getText (x_+1, y_);
		assert (iskanji (oneChar.getChar ()));
		//XXX assertfail뤳Ȥ롣???
		//assert (oneChar2.getCharSet () == TextScreen::CHARSET_KANJI2);
		//assert (oneChar.getColor () == oneChar2.getColor ());
		char str[2];
		str[0] = oneChar.getChar ();
		str[1] = oneChar2.getChar ();
		drawString (x_,y_,str, 2,
			    getTColor (oneChar.getColor ()),
			    isReversed (oneChar.getColor ()),
			    doesNeedRestoreBackground);
		x_++;
		break;
	      }
	    case TextScreen::CHARSET_KANJI2:
	      {//EUC
		assert (1 < x_);//1byteܤ뤫
		const TextScreen::OneChar& oneChar1 = textScreen_.getText (x_-1,y_);
		assert (iskanji (oneChar1.getChar ()));
		assert (oneChar1.getCharSet () == TextScreen::CHARSET_KANJI1);
		assert (oneChar1.getColor () == oneChar.getColor ());
		char str[2];
		str[0] = oneChar1.getChar ();
		str[1] = oneChar.getChar ();
		drawString (x_-1,y_,str,2,
			    getTColor (oneChar.getColor ()),
			    isReversed (oneChar.getColor ()),
			    doesNeedRestoreBackground);
		break;
	      }
	    case TextScreen::CHARSET_KANA:
	      {//kana
		char str[2];
		str[0] = SSO;
		str[1] = oneChar.getChar ();
		drawString (x_,y_,str,2,
			    getTColor (oneChar.getColor ()),
			    isReversed (oneChar.getColor ()),
			    doesNeedRestoreBackground);
		break;
	      }
	    }
	}
    }
}

void IOXWrd::drawString
(uint x_, uint y_, char* str, int numBytes, WdkColor& color_, bool isReverse,
 bool doesNeedRestore = true)
{
#ifdef DEBUG
  //cerr << "IOXWrd::drawString called:" << hex;
  //for (int i = 0; i < numBytes; i++)
  //cerr << static_cast<int>(*(str+i)) << " ";
  //cerr << dec << endl;
  //cerr << "IOXWrd::drawString called:";
  //cerr.write (str,numBytes);
#endif

  assert (1 <= x_);
  assert (x_ <= TEXT_SIZE_X);
  assert (1 <= y_);
  assert (y_ <= TEXT_SIZE_Y);
  assert (str != NULL);
  assert (numBytes > 0);
  
  const int left = (x_-1)*CHAR_SIZE_X;
  const int top = (y_-1)*CHAR_SIZE_Y;
  assert (0 <= left);
  assert (left <= GRAPHIC_SIZE_X-CHAR_SIZE_X);
  assert (0 <= top);
  assert (top <= GRAPHIC_SIZE_Y-CHAR_SIZE_Y);
  
  XRectangle ink;
  XRectangle logical;
  XmbTextExtents (fontSet, str, numBytes, &ink, &logical);
  //CHAR_SIZE_Xñ̤Ƿ夲롣
  const int width = ((logical.width+CHAR_SIZE_X-1)/CHAR_SIZE_X)*CHAR_SIZE_X;
  const int height = CHAR_SIZE_Y;
  //assert (left+width <= GRAPHIC_SIZE_X);//this does not true???
  assert (top+height <= GRAPHIC_SIZE_Y);
  
  WdkGC drawGc (window);
  if (doesNeedRestore)
    pixmap [displayPage]->copyArea (window, drawGc,
				    left, top, width, height, left, top);
  if (isReverse)
    {//ȿžɽ
      WdkPixmap clipMask (window, width, height, 1);
      WdkGC clipGc (&clipMask);
      clipGc.setForeground (1UL);
      clipMask.fillRectangle (clipGc, 0, 0, width, height);
      clipGc.setForeground (0UL);
      //clipMask.drawString (fontSet, clipGc,
      //left-logical.x, top-logical.y, str, numBytes);
      clipMask.drawString (fontSet, clipGc,
			   -logical.x, -logical.y, str, numBytes);

      drawGc.setForeground (color_);
      drawGc.setClipOrigin (left,top);
      drawGc.setClipMask (clipMask);
      window->fillRectangle (drawGc, left, top, width, height);
    }
  else
    {
      drawGc.setForeground (color_);
      window->drawString (fontSet, drawGc, left-logical.x, top-logical.y, str, numBytes);
    }
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////for dynamic loading
///////////////////////////////////////////////////////////////newIO<T>
///////////////////////////////////////////////////////////////newMainIO<T>
///////////////////////////////////////////////////////////////infoIO<T>
////////////////////////////////////////////////////////////////////////////

#include "IO/InfoIO.h"
static IO* newIOXWrd (State* state, WidiController* widiController, 
		      Option* option)
{
  return new IOXWrd (state, widiController, option);
}

InfoIO infoIOXWrd=
{
  newIOXWrd,
  NULL,
  false
};
