/*
    Part.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <cassert>

#include "Widi/State.h"
#include "instrument.h"
#include "Part.h"

using namespace std;

pthread_once_t Part::isInstrumentInited = PTHREAD_ONCE_INIT;

const int Part::MAX_NOTES;
const bool Part::ON;
const bool Part::OFF;
const uchar Part::MAP_OFF;
const uchar Part::MAP_1;
const uchar Part::MAP_2;
const int32 Part::MAX_PITCH_BEND;
const int32 Part::CENTER_PITCH_BEND;
const uint32 Part::MAX_MODULATION;
const uint32 Part::MAX_CHANNEL_VOLUME;
const uint32 Part::MAX_PAN;
const uint32 Part::CENTER_PAN;
const uint32 Part::MAX_EXPRESSION;
const uchar Part::MAX_REVERB_SEND_LEVEL;
const int Part::PARAMETER_NULL;
const int Part::PARAMETER_RPN;
const int Part::PARAMETER_NRPN;

Part::Part ()
{
  pthread_once (&isInstrumentInited, init_inst_name);
  reset (MIDI::PORT_A, 0);
}

void Part::reset (uchar defaultPort, uchar receivedChannel)
{
  //assert (0<=receivedChannel);
  assert (receivedChannel < MIDI::MAX_CHANNEL);
#if 0
  assert (defaultPort == SequencerStatus::PORT_A ||
	  defaultPort == SequencerStatus::PORT_B);
#else
  //assert (0 <= defaultPort);
  assert (defaultPort < MIDI::MAX_PORT);
#endif
  RxChannel = receivedChannel;
  if (receivedChannel == 10-1)
    useForRhythmPart = MAP_1;
  else
    useForRhythmPart = MAP_OFF;
  port = defaultPort;

  reset ();
}

void Part::reset ()
{
  //XXX sc88_pro$B$G$OBLL\!#@_Dj$K$h$jJQ$($k$3$H!#(B
  if (useForRhythmPart == MAP_OFF)
    instrument = inst_sc88_pro[0][0];
  else
    instrument = drum_sc88_pro[0];
  isMuted=false;
  //XXX:default$B$NCM$OE,Ev!#8e$G!"$A$c$s$H$7$?CM$KD>$9$3$H!#(B
  for (int i = 0; i < MAX_NOTES; i++)
    {
      lastNote[i] = OFF;
      note[i] = OFF;
      velocity[i] = 0;
      polyphonicKeyPressure[i] = 0;
      noteOnMachineTicks[i] = 0;
    }
  
  maxVelocity = 0;
  lastMaxVelocity = 0;
  lastMachineTicks = 0;
  notesOnNumber = 0;
  channelPressure = 0;
  program = 0;
  bankMSB = 0;
  bankLSB = 0;
  
  pitchBend = 0;
  modulationWheel = 0;
  channelVolume = MAX_CHANNEL_VOLUME;
  pan = CENTER_PAN;
  expression = MAX_EXPRESSION;  
  damperPedal = 0;
  portamentSwitch = 0;
  sostenute = 0;
  softPedal = 0;

  reverbSendLevel = 0;
  chorusSendLevel = 0;
  delaySendLevel = 0;

  parameterKind = PARAMETER_NULL;
  rpnMSB = 0;
  rpnLSB = 0;
  nrpnMSB = 0;
  nrpnLSB = 0;
  vibrateRate = 0;
  vibrateDepth = 0;
  vibrateDelay = 0;
  tvfCutOff = 0;
  tvfResonance = 0;
  envelopeAttackTime = 0;
  envelopeDecayTime = 0;
  envelopeReleaseTime = 0;

#if 0
  toneNumber = ;
  RxPitchBend = ;
  RxChPressure = ;
  RxProgramChange = ;
  RxCongrolChange = ;
  RxPolyPressure = ;
  RxNoteMessage = ;
  RxRPN = ;
  RxNRPN = ;
  RxModulation = ;
  RxVolume = ;
  RxPanpot = ;
  RxExpression = ;
  RxHold1 = ;
  RxPortamento = ;
  RxSostenute = ;
  RxSoft = ;
  MonoPolyMode = ;
  assignMode = ;

  RxBankSelect = ;
  RxBandSelectLSB = ;
#endif
}


uchar Part::getNoteOnVelocity (const SongStatus& nowSongStatus) const
{
  uchar retVelocity = 0;
  for (int note = 0; note < MAX_NOTES; note++)
    {
      if (retVelocity < velocity[note])
        retVelocity = velocity[note];
    }
#define DELAY (nowSongStatus.rate*120/100)//1.2 second
#define DECAY_RATE 40/100 //40%$BA4$/:,5r$J$7!#;n9T:x8m$G7h$a$k!#(B
  if (nowSongStatus.machineTicks-lastMachineTicks >= DELAY)
    return retVelocity*DECAY_RATE;
  else
    return retVelocity;
#undef DECAY_RATE
#undef DELAY
}

uchar Part::getMaxVelocity (const SongStatus& nowSongStatus) const
{
#define DELAY (nowSongStatus.rate*15/100)//0.15 second
  if (nowSongStatus.machineTicks-lastMachineTicks < DELAY)
    return lastMaxVelocity;
  else
    return 0;
#undef DELAY
}


uchar Part::getNoteOnVelocity2 (const SongStatus& nowSongStatus) const
{
  uchar retVelocity = 0;
  for (int note = 0; note < MAX_NOTES; note++)
    {
      if (retVelocity < velocity[note])
        retVelocity = velocity[note];
    }
#define DELAY (nowSongStatus.rate*40/100)//0.4 second
  if (nowSongStatus.machineTicks-lastMachineTicks >= DELAY)
    return 0;
  else
    return retVelocity;
#undef DELAY
}

uchar Part::getMaxVelocity2 (const SongStatus& nowSongStatus) const
{
#define DELAY (nowSongStatus.rate*15/100)//0.15 second
  if (nowSongStatus.machineTicks-lastMachineTicks < DELAY)
    return lastMaxVelocity;
  else
    return 0;
#undef DELAY
}
