#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "SerialDevice.h"
#include "util/schedYield.h"

using namespace std;

#if defined (__hpux)
const char* const SerialDevice::DEFAULT_SERIAL_DEVICE = "/dev/tty00";
#elif defined(__linux__)
const char* const SerialDevice::DEFAULT_SERIAL_DEVICE = "/dev/ttyS0";
//const char* const SerialDevice::DEFAULT_SERIAL_DEVICE = "/dev/ttyS1";
//const char* const SerialDevice::DEFAULT_SERIAL_DEVICE = "/dev/cua0";
#else
const char* const SerialDevice::DEFAULT_SERIAL_DEVICE = "/dev/ttya";
#endif

const int SerialDevice::DEFAULT_BPS = 38400;
const speed_t SerialDevice::B31250;

SerialDevice::SerialDevice ()
{
  open (DEFAULT_SERIAL_DEVICE, DEFAULT_BPS);
}

SerialDevice::SerialDevice (const string& device)
{
  if (device == "")
    open (DEFAULT_SERIAL_DEVICE, DEFAULT_BPS);
  else
    open (device, DEFAULT_BPS);
}

SerialDevice::SerialDevice (const string& device, int bps)
{
  open (device, bps);
}

SerialDevice::~SerialDevice ()
{
  close (serialfd);
}

void SerialDevice::open (const string& device, int bps)
{
  //these codes is stolen from midiplay
  speed_t speed;
  
  switch(bps)
    {
    case 38400:
      speed = B38400;
      break;
    case 31250:
      speed = B31250;
      break;
    default:
      cerr << "SerialDevice::SerialDevice: Bogus speed " << bps
	   << " is given." << endl;
      throw BogusBPS ();
      break;
    }
  
  serialfd = std::open(device.c_str (), O_RDWR | O_NDELAY);
  if (serialfd == -1)
    {
      cerr << "SerialDevice::SerialDevice: Can't open device "
	   << device << "." << endl;
      throw CanNotOpenDevice ();
    }

  if (isatty (serialfd))
    {
      termios tio;
      if (tcgetattr (serialfd, &tio) < 0)
	{
	  cerr << "SerialDevice::SerialDevice:Can't get attribute of"
	       <<  device << "." << endl;
	  throw CanNotGetDeviceAttribute ();
	}
      tio.c_iflag = 0;
      tio.c_oflag = 0;
      tio.c_cflag = CS8 | CREAD | CLOCAL;
      tio.c_lflag = 0;
      cfsetispeed (&tio, speed);
      cfsetospeed (&tio, speed);
      if (tcsetattr (serialfd, TCSANOW, &tio) < 0)
	{
	  cerr << "SerialDevice::SerialDevice: Can't set attribute of"
	       << device << "." << endl;
	  throw CanNotSetDeviceAttribute ();
	}
    }
}



void SerialDevice::write (const uchar* data, ssize_t length)
{
#ifdef DEBUG
  //cerr << "SerialDevice::write called:length=" << length << endl;
#endif

  ssize_t wroteSize = 0;
  while (wroteSize < length)
    {
      const int ret = std::write (serialfd, data+wroteSize, length-wroteSize);
      if (ret == -1)
	{
	  if (errno == EAGAIN)
	    {//XXX$B$7$P$i$/BT$F$P$$$$$N$@$,!D!#(B
#if 0
	      if (fsync (serialfd) == -1)
		{
		  perror ("SerialDevice::write:fsync Can not fsync.fatal error.");
		  exit (1);
		}
#else
	      schedYield ();
#endif
	      continue;
	    }
	  
	  if (errno != EINTR)
	    {
	      //XXX $B$b$C$H$^$7$J%(%i!<=hM}$NJ}K!$,$"$kH&!#(B
	      perror ("SerialDevice::write: Can not write.fatal error.");
	      exit (1);
	    }
	}
      wroteSize += ret;
    }
}
