#undef DEBUG

#include "TimedSerial.h"
#include "util/PThreadStartRoutine.h"
#include "util/compareTime.h"

////////////////////////////////////////////////////TimedSerial::TimedDataNode
const int TimedSerial::TimedDataNode::BUFFER_UNIT_SIZE;

TimedSerial::TimedDataNode::TimedDataNode (ulong deltaTicks__,
					   ulong totalTicks__,
					   timespec startTime, int ctrlRate):
  deltaTicks (deltaTicks__), totalTicks (totalTicks__),
  data (new uchar[BUFFER_UNIT_SIZE]),
  length (0), bufferLength (BUFFER_UNIT_SIZE),
  next (NULL)
{
  const time_t sec = totalTicks/ctrlRate;
  const long nsec = static_cast<long> ((totalTicks-ctrlRate*sec)*1000000000.0/ctrlRate);
  
  absTimeSpec.tv_sec = startTime.tv_sec + sec;
  absTimeSpec.tv_nsec = startTime.tv_nsec + nsec;
  if (absTimeSpec.tv_nsec > 999999999)
    {//夬
      absTimeSpec.tv_sec += 1;
      absTimeSpec.tv_nsec -= 1000000000;
    }
#ifdef DEBUG
  cerr << "deltaTicks=" << deltaTicks
       << ":totalTicks=" << totalTicks
       << endl
       << ":startTime"
       << ":sec=" << startTime.tv_sec
       << ":nsec=" << startTime.tv_nsec
       << endl
       << ":TimedSerial::TimedDataNode::TimedDataNode"
       << ":sec=" << sec
       << ":nsec=" << nsec
       << endl
       << ":absTimeSpec sec=" << absTimeSpec.tv_sec
       << ":nsec=" << absTimeSpec.tv_nsec
       << endl;
#endif
}

TimedSerial::TimedDataNode::~TimedDataNode ()
{
  delete[] data;
}

void TimedSerial::TimedDataNode::addData (const uchar* data__, int length__)
{
  if (length+length__ > bufferLength )
    {
      bufferLength += BUFFER_UNIT_SIZE*((length+length__-bufferLength+BUFFER_UNIT_SIZE-1)/BUFFER_UNIT_SIZE);
      uchar* tmp = new uchar[bufferLength];
      memcpy (tmp, data, length);
      delete[] data;
      data = tmp;
    }

  memcpy (data+length, data__, length__*sizeof (uchar));
  length += length__;
}

///////////////////////////////////////////////////////TimedSerial::TimedData

TimedSerial::TimedData::TimedData (int ctrlRate__):
  ctrlRate (ctrlRate__), totalTicks (0), first (NULL), last (NULL)
{
  
}

TimedSerial::TimedData::~TimedData ()
{
  assert (first == NULL);
  assert (last == NULL);
}


void TimedSerial::TimedData::resetNode ()
{
  DEBUG_MESSAGE ("TimedSerial::TimedData::resetNode called.");
  TimedDataNode* next = first;
  while (next != NULL)
    {
      TimedDataNode* tmp = next->getNext ();
      delete next;
      next = tmp;
    }
  first = NULL;
  last = NULL;
  totalTicks = 0;
  DEBUG_MESSAGE ("TimedSerial::TimedData::resetNode done.");
}

void TimedSerial::TimedData::setStartTime (timeval& startTime__)
{
  assert (first == NULL);
  assert (last == NULL);
  
  startTime.tv_sec = startTime__.tv_sec;
  startTime.tv_nsec = startTime__.tv_usec * 1000;
  first = new TimedDataNode (0, 0, startTime, ctrlRate);
  last = first;
}

void TimedSerial::TimedData::addData (const uchar* data__, int length__)
{
#if 0
  assert (last != NULL);
#else
  if (last == NULL)
    {
      assert (first == NULL);
      first = new TimedDataNode (0, 0, startTime, ctrlRate);
      last = first;
    }
#endif

  last->addData (data__, length__);
}

void TimedSerial::TimedData::totalTime (ulong totalTicks__)
{
  if (totalTicks < totalTicks__)
    deltaTime (totalTicks__-totalTicks);
}

void TimedSerial::TimedData::deltaTime (ulong deltaTicks__)
{
  totalTicks += deltaTicks__;

  if (last == NULL)
    {
      assert (first == NULL);
      first = new TimedDataNode (deltaTicks__, totalTicks,
				 startTime, ctrlRate);
      last = first;
    }
  else
    {//last != NULL
      assert (first != NULL);
      last->setNext (new TimedDataNode (deltaTicks__, totalTicks,
					startTime, ctrlRate));
      last = last->getNext ();
    }
}

TimedSerial::TimedDataNode* TimedSerial::TimedData::popFront ()
{
  assert (first != NULL);
  TimedDataNode* ret = first;
  first = first->getNext ();
  if (first == NULL)
    last = NULL;
  return ret;
}


////////////////////////////////////////////////////////////TimedSerial
const int TimedSerial::CTRL_RATE;
const int TimedSerial::BUFFER_UNIT_SIZE;

TimedSerial::TimedSerial (const string& device = SerialDevice::DEFAULT_SERIAL_DEVICE,
			  int bps = SerialDevice::DEFAULT_BPS):
  timedData (CTRL_RATE), isTimerLoopRunning (false), isTimerWorking (false),
  buffer (new uchar[BUFFER_UNIT_SIZE]),bufferLength (BUFFER_UNIT_SIZE),
  bufferedSize (0)
{
  try
    {
      serialDevice = new SerialDevice (device);
    }
  catch (SerialDevice::Error&)
    {
      throw CanNotOpenDevice ();
    }
}

TimedSerial::~TimedSerial ()
{
  sync ();
  if (isTimerWorking)
    stopTimer ();
  delete serialDevice;
  delete[] buffer;
}

void TimedSerial::sleepUntil (const timespec& absTimeSpec)
{
  timeval timeVal;
  gettimeofday (&timeVal, NULL);

  timespec restTimeSpec;
  if (diffTime (absTimeSpec, timeVal, restTimeSpec) > 0)
    {//Ԥɬפä
      timespec remainTimeSpec;
      while (nanosleep (&restTimeSpec, &remainTimeSpec) != 0)
	{
	  if (errno == EINTR)
	    {
	      //signalnanosleepǤ줿Τǡnanosleepľ
#if 1
	      restTimeSpec = remainTimeSpec;
#else
	      if (diffTime (absTimeSpec, timeVal, restTimeSpec) <= 0)
		break;
#endif
	    }
	  else
	    {//XXX
	      assert (errno != EINVAL);
	      DEBUG_MESSAGE ("TimedSerial::sleepUntil should not be reached.");
	      break;
	    }
	}
    }
}


void TimedSerial::timerLoop ()
{
  DEBUG_MESSAGE ("TimedSerial::timerLoop called.");

  timedData.lock ();
  assert (isTimerWorking == true);//???

  //timerLoopäȤ󤻤롣
  timerLoopBeginMutex.lock ();
  isTimerLoopRunning = true;
  timerLoopBeginMutex.unlock ();
  timerLoopBegin.signal ();
  
  DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locking 0.");
  timeval startTime;
  gettimeofday (&startTime, NULL);
  timedData.setStartTime (startTime);

  DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locked 0.");
  while (true)
    {
    CONTINUE_TIMER:
      DEBUG_MESSAGE ("TimedSerial::timerLoop looping.");
      while (timedData.getFirst () == NULL)
	{
	  DEBUG_MESSAGE ("TimedSerial::timerLoop timedData waiting.");
	  if (isTimerWorking == false)
	    {
	      DEBUG_MESSAGE ("TimedSerial::timerLoop timer finishing.");
	      goto TIMER_FINISHED;//timerߤäƤΥ롼פɬפʤ
	    }
	  timedData.wait ();
	}
      
      
      if (timedData.getFirst ()->getNext () == NULL)
	{//⤦nodeǤޤʤΤǤФ
	  DEBUG_MESSAGE ("TimedSerial::timerLoop getNext == NULL.");
	  timespec absTimeSpec = timedData.getFirst ()->getAbsTimeSpec ();
	  timespec restTimeSpec;
	  timeval timeVal;
	  gettimeofday (&timeVal, NULL);
	  if (diffTime (absTimeSpec, timeVal, restTimeSpec) > 0)
	    {
	      timedData.unlock ();
	      DEBUG_MESSAGE ("TimedSerial::timerLoop sleeping.");
#ifdef DEBUG
	      cerr << "absTimeSpec sec=" << absTimeSpec.tv_sec
		   << ":nsec=" << absTimeSpec.tv_nsec
		   << endl;
	      cerr << "timeVal     sec=" << timeVal.tv_sec
		   << ":nsec=" << timeVal.tv_usec*1000
		   << endl;
	      cerr << "nanosleep   sec=" << restTimeSpec.tv_sec
		   << ":nsec=" << restTimeSpec.tv_nsec
		   << endl;
#endif
	      nanosleep (&restTimeSpec, NULL);
	      DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locking 1.");
	      timedData.lock ();//ΥåϤФʤ
	      DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locked 1.");
	    }
	  
	  while (timedData.getFirst ()->getNext () == NULL)
	    {//Υ롼פˤϤȤϱդ٤Ƥ뤳Ȥ̣롣
	      //ˤˤΥ롼פˤϤ褦ʤФ
	      DEBUG_MESSAGE ("TimedSerial::timerLoop critical loop.");
	      if (timedData.getFirst ()->getDataLength () > 0)
		{
		  DEBUG_MESSAGE ("TimedSerial::timerLoop length > 0.");
		  serialDevice->write (timedData.getFirst ()->getData (),
				       timedData.getFirst ()->getDataLength ());
		  timedData.getFirst ()->trancateData ();
		}
	      else if (!isTimerWorking)
		{//timerνλ
		  delete timedData.popFront ();
		  goto CONTINUE_TIMER;
		}
	      DEBUG_MESSAGE ("TimedSerial::timerLoop critical waiting.");
	      timedData.signal ();//for sync
	      timedData.wait ();
	      if (timedData.getFirst () == NULL)
		{
		  goto CONTINUE_TIMER;
		}
	    }
	  if (timedData.getFirst ()->getDataLength () > 0)
	    serialDevice->write (timedData.getFirst ()->getData (),
				 timedData.getFirst ()->getDataLength ());
	  delete timedData.popFront ();
	}
      else
	{//̤nodeԤäƤ롣
	  //XXX nodeޤĴ٤٤
	  DEBUG_MESSAGE ("TimedSerial::timerLoop getNext != NULL.");
	  TimedDataNode* now = timedData.popFront ();
	  timedData.unlock ();
	  if (now->getDataLength () > 0)
	    {
	      DEBUG_MESSAGE ("TimedSerial::timerLoop sleepUntil.");
	      sleepUntil (now->getAbsTimeSpec ());
	      serialDevice->write (now->getData (), now->getDataLength ());
	    }
	  delete now;
	  DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locking 2.");
	  timedData.lock ();
	  DEBUG_MESSAGE ("TimedSerial::timerLoop timedData.locked 2.");
	}
    }

 TIMER_FINISHED:
  timedData.resetNode ();
  timedData.unlock ();
  DEBUG_MESSAGE ("TimedSerial::timerLoop finished.");
}


void TimedSerial::write (const uchar* data, ssize_t length)
{
  DEBUG_MESSAGE ("TimedSerial::write called.");
#if 0
  //lock򾯤Ǥ⾯ʤ롣
  if (bufferedSize + length > bufferLength)
    {
      bufferLength = bufferedSize + length + BUFFER_UNIT_SIZE;
      uchar* tmp = new uchar[bufferLength];
      memcpy (tmp, buffer, bufferedSize);
      delete[] buffer;
      buffer = tmp;
    }
  memcpy (buffer+bufferedSize, data, length);
  bufferedSize += length;
#else

#if 0
  if (!isTimerWorking)
    {
      serialDevice->write (data, length);
      return;
    }
#endif
  
  
  //lock򾯤Ǥ⾯ʤ롣
  if (bufferedSize + length > BUFFER_UNIT_SIZE)
    {
      timedData.lock ();
      timedData.addData (buffer, bufferedSize);
      timedData.addData (data, length);
      timedData.unlock ();
      timedData.signal ();
      bufferedSize = 0;
    }
  else
    {
      memcpy (buffer+bufferedSize, data, length);
      bufferedSize += length;
    }
#endif
}


void TimedSerial::waitTimer (ulong totalTicks)
{
  DEBUG_MESSAGE ("TimedSerial::waitTimer called.");
  if (isTimerWorking)
    {
      DEBUG_MESSAGE ("TimedSerial::waitTimer locking.");
      timedData.lock ();
      DEBUG_MESSAGE ("TimedSerial::waitTimer locked.");
      if (bufferedSize > 0)
	{
	  timedData.addData (buffer, bufferedSize);
	  bufferedSize = 0;
	}
      timedData.totalTime (totalTicks);
      timedData.unlock ();
      DEBUG_MESSAGE ("TimedSerial::waitTimer unlocked.");
      timedData.signal ();
    }
}

void TimedSerial::deltaTimer (ulong deltaTicks)
{
  DEBUG_MESSAGE ("TimedSerial::deltaTimer called.");
  if (isTimerWorking)
    {
      DEBUG_MESSAGE ("TimedSerial::deltaTimer locking.");
      timedData.lock ();
      DEBUG_MESSAGE ("TimedSerial::deltaTimer locked.");
      if (bufferedSize > 0)
	{
	  timedData.addData (buffer, bufferedSize);
	  bufferedSize = 0;
	}
      timedData.deltaTime (deltaTicks);
      timedData.unlock ();
      DEBUG_MESSAGE ("TimedSerial::deltaTimer unlocked.");
      timedData.signal ();
    }
}

void TimedSerial::startTimer ()
{
  DEBUG_MESSAGE ("TimedSerial::startTimer called.");
  DEBUG_MESSAGE ("TimedSerial::startTimer locking 0.");
  timedData.lock ();
  DEBUG_MESSAGE ("TimedSerial::startTimer locked 0.");
  if (isTimerWorking)
    {//timerưƤΤǡTimedSerial::timerLoopưƤΤ
      //ߤ롣
      timedData.resetNode ();
      isTimerWorking = false;
      timedData.unlock ();
      DEBUG_MESSAGE ("TimedSerial::startTimer unlocked.");
      timedData.signal ();
      DEBUG_MESSAGE ("TimedSerial::startTimer joining.");
      pthread_join (timerLoopThread, NULL);
      DEBUG_MESSAGE ("TimedSerial::startTimer joined.");
      DEBUG_MESSAGE ("TimedSerial::startTimer locking 1.");
      timedData.lock ();
      DEBUG_MESSAGE ("TimedSerial::startTimer locked 1.");
      
    }

  isTimerWorking = true;
  timedData.unlock ();
  DEBUG_MESSAGE ("TimedSerial::startTimer unlocked.");

  DEBUG_MESSAGE ("TimedSerial::startTimer thread creating.");

  
  //timerLoop threadcreate
  timerLoopBeginMutex.lock ();
  isTimerLoopRunning = false;
  const int ret = pthread_create (&timerLoopThread, NULL,
				  &PThreadStartRoutine<TimedSerial, &TimedSerial::timerLoop>,
				  static_cast<void*> (this));
  if (ret != 0)
    {
      //XXX error
      cerr << "TimedSerial::startTimer can not create timerThread" << endl;
      exit (1);
    }
  //timerLoop threadäȤγǧ
  while (isTimerLoopRunning == false)
    timerLoopBegin.wait (timerLoopBeginMutex);
  timerLoopBeginMutex.unlock ();
  

  DEBUG_MESSAGE ("TimedSerial::startTimer done.");
}

void TimedSerial::stopTimer ()
{
  sync ();
  DEBUG_MESSAGE ("TimedSerial::stopTimer called.");
  DEBUG_MESSAGE ("TimedSerial::stopTimer locking.");
  timedData.lock ();
  DEBUG_MESSAGE ("TimedSerial::stopTimer locked.");
  assert (isTimerWorking == true);
  //timedData.resetNode ();
  isTimerWorking = false;
  timedData.unlock ();
  DEBUG_MESSAGE ("TimedSerial::stopTimer unlocked.");
  timedData.signal ();
  
  //timerLoop threadjoin
  pthread_join (timerLoopThread, NULL);
#if 0
  timedData.lock ();
  timedData.resetNode ();
  timedData.unlock ();
#endif
}


//???
void TimedSerial::sync ()
{
  DEBUG_MESSAGE ("TimedSerial::sync called.");
  if (!isTimerWorking)
    {//timerưƤʤ
      timedData.lock ();
      while (timedData.getFirst () != NULL)
	{
	  assert (timedData.getFirst () == timedData.getLast ());
	  serialDevice->write (timedData.getFirst ()->getData (),
			       timedData.getFirst ()->getDataLength ());
	  timedData.popFront ();
	}
      serialDevice->write (buffer, bufferedSize);
      bufferedSize = 0;
      timedData.unlock ();
      return;
    }
  

#if 0
  //XXXϸ©? ⤦礤ޤˡ?
  //Ԥ֤׻롣
  DEBUG_MESSAGE ("TimedSerial::sync locking.");
  timedData.lock ();
  DEBUG_MESSAGE ("TimedSerial::sync locked.");
  if (timedData.getLast () == NULL)
    {
      timedData.unlock ();
      DEBUG_MESSAGE ("TimedSerial::sync unlocked.");
      return;
    }
  
  timespec absTime = timedData.getLast ()->getAbsTimeSpec ();
  timedData.unlock ();
  DEBUG_MESSAGE ("TimedSerial::sync unlocked.");

  //Ԥġ
  DEBUG_MESSAGE ("TimedSerial::sync sleepUntil.");
  sleepUntil (absTime);
#else
  timedData.lock ();
  while (timedData.getFirst () != NULL &&
	 timedData.getFirst ()->getDataLength () != 0)
    timedData.wait ();
  timedData.unlock ();
#endif
}
