#ifndef TimedSerial_H
#define TimedSerial_H


#include <ctime>
#include <sys/time.h>
#include <list>

#include "SerialDevice.h"
#include "util/CondVar.h"

using namespace std;

class TimedSerial
{
public:
  //1000$B$H$$$&?t;z<+BN$K$O?<$$0UL#$,L5$$!#(B1000Hz$B$GF0$/(B
  static const int CTRL_RATE = 1000;
private:
  struct TimedDataNode
  {
  private:
    static const int BUFFER_UNIT_SIZE = 256;//$B4v$D$K$9$k$N$,E,Ev$@$i$&!#(B
    timespec absTimeSpec;//$B$3$N;~4V$K$J$C$?$i(Bwrite$B$9$k!#(B
    const ulong deltaTicks;//$BA0$N(Bnode$B$+$i$3$l$@$1N%$l$F$$$k!#(B
    const ulong totalTicks;
    uchar* data;
    int length;//$B<B:]$KCV$+$l$F$$$k(Bdata$B$N(Blength
    int bufferLength;//data$B$r3NJ]$7$?D9$5(B
    TimedDataNode* next;//single linked list
    
  public:
    TimedDataNode (ulong deltaTicks__, ulong totalTicks__,
		   timespec startTime, int ctrlRrate);
    ~TimedDataNode ();
    
    const timespec& getAbsTimeSpec () const {return absTimeSpec;};
    const ulong getDelatTicks () const {return deltaTicks;};
    const ulong getTotalTicks () const {return totalTicks;};
    void addData (const uchar* data__, int length__);
    void setNext (TimedDataNode* next__)
      {
	assert (next == NULL);
	assert (next__ != NULL);
	next = next__;
      };
    TimedDataNode* getNext () {return next;};
    const uchar* getData () const {return data;};
    const int getDataLength () const {return length;};
    void trancateData () {length = 0;};
  };

  struct TimedData
  {
  private:
    const int ctrlRate;
    Mutex mutex;
    CondVar condVar;
    ulong totalTicks;
    timespec startTime;

    TimedDataNode* first;
    TimedDataNode* last;

  public:
    
    TimedData (int ctrlRate__);
    ~TimedData ();

    void resetNode ();
    
    //$B:G=i$K$3$l$r8F$s$G$*$/I,MW$,$"$k!#(B
    void setStartTime (timeval& startTime__);
    
    void addData (const uchar* data__, int length__);
    void totalTime (ulong totalTicks__);
    void deltaTime (ulong deltaTicks__);

    TimedDataNode* getFirst () {return first;};
    TimedDataNode* getLast () {return last;};
    
    //$B8F$S=P$7B&$OJV$jCM$N(BTimedDataNode$B$r(Bdelete$B$9$kI,MW$,$"$k!#(B
    TimedDataNode* popFront ();

    
    //thread stuff
    void lock () {mutex.lock ();};
    void unlock () {mutex.unlock ();};
    void signal () {condVar.signal ();};
    void wait () {condVar.wait (mutex);};
  };
  TimedData timedData;

  pthread_t timerLoopThread;
  
  //timer loop$B$,Av$C$?$3$H$r3NG'$9$k$?$a$NJ*!#(B
  Mutex timerLoopBeginMutex;
  CondVar timerLoopBegin;
  bool isTimerLoopRunning;
  
  //absTimeSpec$B$^$GBT$D!#$b$&;~4V$,2a$.$F$$$?$i$9$0$K(Breturn$B$9$k!#(B
  void sleepUntil (const timespec& absTimeSpec);
  void timerLoop ();

  //$B$3$NJQ?t$O(BtimedData.mutex$B$GJ]8n$5$l$k!#(B
  bool isTimerWorking;
  //timeval startTime;
  //ulong extraTime;

  

  SerialDevice* serialDevice;
  
  //timedData$B$N(Block$B$r>/$7$G$b>/$J$/$9$k0Y$K(Bbuffering$B$9$k!#(B
  static const int BUFFER_UNIT_SIZE = 256;
  uchar* buffer;
  int bufferLength;//$B8=:_3NJ]$5$l$F$$$k(Bbuffer$B$ND9$5!#(B
  int bufferedSize;//$B8=:_(Bbuffer$B$5$l$F$$$k%G!<%?$NNL(B
public:
  TimedSerial (const string& device = SerialDevice::DEFAULT_SERIAL_DEVICE,
	       int bps = SerialDevice::DEFAULT_BPS);
  ~TimedSerial ();

  void write (const uchar* data, ssize_t length);

  void waitTimer (ulong ticks);
  void deltaTimer (ulong ticks);
  
  void startTimer ();
  void stopTimer ();
  
  void sync ();

  bool isTimerRunning () const {return isTimerWorking;};
  
  ///////////////////////////////////////////////////////exceptional classes
  class Error {};
  class CanNotOpenDevice:public Error {};
};

#endif// TimedSerial_H
