/*
    SMF.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <strstream>

#include "util/KanjiConvert.h"
#include "SMF.h"
#include "Sequencer/Sequencer.h"

using namespace std;

//for playing
SMF::SMF (State* state_, SongCommand* songCommand_,
	  MainIO& userIO, Option* option_):
  Song (state_, songCommand_, userIO, option_),
  length (0),smfTrack (NULL)
{
  //do nothing
}


SMF::SMF (const string& fileName, State* state_, SongCommand* songCommand_,
	  MainIO& userIO, Option* option_):
  Song (state_, songCommand_, userIO, option_),
  length (0),smfTrack (NULL)
{
  loadFile (fileName);
}

SMF::SMF (SongInputStream& songInputStream,
	  State* state_, SongCommand* songCommand_, MainIO& userIO,
	  Option* option_):
  Song (state_, songCommand_, userIO, option_),
  length (0),smfTrack (NULL)
{
  loadFile (songInputStream);
}

#if 0
//for converting
SMF::SMF (const string& fileName, State* state_, SongCommand* songCommand_,
	  MainIOConvert2Dumb& userIO, Option* option_,
	  DumbEventList& dumbEventList_):
  Song (state_, songCommand_, userIO,option_, dumbEventList_),
  length (0),smfTrack (NULL)
{
  loadFile (fileName);
}
#endif

SMF::~SMF ()
{
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    delete smfTrack[i];
  delete[] smfTrack;
}

bool SMF::isSupported (const string& fileName)
{
  try
    {
      SongInputStream songInputStream (fileName);
      char MThd[5];
      MThd[4] = '\0';
      songInputStream.read (MThd,4);
      songInputStream.close ();
      if (strcmp (MThd, "MThd") == 0)
	return true;
      else
	return false;
    }
  catch (Error& error)
    {
      cerr << error.getMessage ();
      return false;
    }

  return false;
}


void SMF::doLoadBuffer (const char* buffer, int bufferSize)// throw (Error)
{
#ifdef DEBUG
  //cerr << "loadBuffer bufferSize=" << bufferSize << endl;
#endif
  SMFBufferStream smfBufferStream (buffer, bufferSize);
  readHeader (smfBufferStream);
  smfTrack=new SMFTrack*[cacheSongStatus->getNTracks ()];
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    {
#ifdef DEBUG
      //cerr << "read " << i << "-th track" << endl;
#endif
      smfTrack[i]=new SMFTrack (smfBufferStream);
    }
}

void SMF::readHeader (SMFBufferStream& smfBufferStream) throw (Error)
{

  //////////////////////////////////////////check magic number
  char MThd[5];
  MThd[4] = '\0';
  smfBufferStream.read (MThd, 4);
  if (strcmp (MThd,"MThd") != 0)
    {
#ifdef DEBUG
      //cerr << "MThd=" << MThd << endl;
#endif//DEBUG
      Error error ("this file is not midi file.magic=", MThd);
      throw error;
    }
  
  ///////////////////////////////////////////////////////read head
  length = smfBufferStream.get32B ();
  const int format = smfBufferStream.get16B ();
  if (format == 0)
    cacheSongStatus->setFormat (SongStatus::SMF_0);
  else if (format == 1)
    cacheSongStatus->setFormat (SongStatus::SMF_1);
  else if (format == 2)
    cacheSongStatus->setFormat (SongStatus::SMF_2);
  else
    cacheSongStatus->setFormat (SongStatus::FILE_FORMAT_UNKNOWN);
  cacheSongStatus->setNTracks (smfBufferStream.get16B ());
  cacheSongStatus->setDivision (smfBufferStream.get16B ());
  
#ifdef DEBUG
  //cout << "length=" << length << "byte" << endl;
  //cout << "format=" << format << endl;
  //cout << "number of tracks=" << nTracks << endl;
  //cout << "division=" << division << endl;
#endif//DEBUG

  if (length < 0)
    {
      Error error ("MThd lenght is negative.file is borken?");
      throw (error);
    }

  if (length > 6)
    {//$BG0$N$?$a!#(B
      smfBufferStream.seekgCur (length-6);
      io.errorMessage ("unknwonw SMF header size detected.");
#ifdef DEBUG
      //cerr << "header size=" << length << endl;
#endif
    }

  if (cacheSongStatus->getNTracks () <= 0)
    {
      Error error ("no track found.");
      throw (error);
    }
}

void SMF::doPlay (Sequencer& sequencer)
{
  //DEBUG_MESSAGE ("DEBUG:SMF::doPlay");
  uint32 lastTicks = 0;//$BA02s(BwaitTime$B$7$?$H$-$N@dBP(Btick$B?t!#(B

  double machineTicks = 0;//$B8=:_1iAU$7$F$$$k@dBP(Bmachine ticks
  double dMachineTicks = 0;//$B7W;;MQ!#(B
  double lastDumpTicks = 0;//$BA02s(BdumpBuf$B$7$F$+$i$N(Bmachine tick$B?t!#(B
  
  double logicalTicks = 0;//$B8=:_1iAU$7$F$$$k@dBP(Blogical ticks
  double dLogicalTicks = 0;//$B7W;;MQ!#(B

  cacheSongStatus->lastMeasureTicks = 0;//$BA02s>.@a$N=*$j$@$C$?;~$N(Bticks$B?t!#(B
  cacheSongStatus->lastMeasureMachineTicks = 0;//$BA02s>.@a$N=*$j$@$C$?;~$N(BmachineTicks$B?t!#(B
  cacheSongStatus->lastMeasureLogicalTicks = 0;//$BA02s>.@a$N=*$j$@$C$?;~$N(BlogicalTicks$B?t!#(B
  cacheSongStatus->ticksFromLastMeasure = 0;//$B>.@aFb$N(Bticks$B?t!#(B
  uint32 setTempoTicks = 0;//set tempo$B$7$?$H$-$N;~4V7W;;$NJd@5MQ!#(B
  uint32 oldTempo = cacheSongStatus->getTempo ();//set tempo$B$7$?$H$-$N;~4V7W;;$NJd@5MQ!#(B

  cacheSongStatus->events = 0;//midi event$B$N?t(B
  
  uint16 nEndOfTracks = 0;//$B1iAU$N=*N;$7$?%H%i%C%/?t(B
  int playTrack = 0;//$B8=:_CmL\$7$F$$$k%H%i%C%/(B
  
  /////////$B<B:]$N1iAU(B
  while (nEndOfTracks < cacheSongStatus->getNTracks ())
    {
      //DEBUG_MESSAGE ("DEBUG:now playing.");
      /////$B$I$N%H%i%C%/$,:G=i$K1iAU$5$l$k$+H=Dj!#(B
      for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
	{
	  if (smfTrack[i]->getABSTicks () <
	      smfTrack[playTrack]->getABSTicks () &&
	      !smfTrack[i]->isEnd ())
	    playTrack = i;
	}

      //$B;~4VBT$A$,H/@8$7$?$i(Bbuffer$B$r(Bdump$B$9$k!#(B
      //$BBT$A;~4V7W;;!#(BMIDI ticks$B?t$+$i(Bmachine ticks$B?t$r7W;;!#(B
      dMachineTicks = 0;
      dLogicalTicks = 0;
      const uint32 playTicks = smfTrack[playTrack]->getABSTicks ();
      if (lastTicks<playTicks)
	{
	  dMachineTicks =
	    calcDMachineTicks (playTicks, lastTicks, setTempoTicks,
			       cacheSongStatus->getTempo (), oldTempo,
			       cacheSongStatus->rate,
			       cacheSongStatus->getDivision (),
			       cacheSongStatus->getSkew ());
	  dLogicalTicks =
	    calcDLogicalTicks (playTicks, lastTicks, setTempoTicks,
			       cacheSongStatus->getTempo (), oldTempo,
			       cacheSongStatus->getDivision ());

	  //$B0l>.@aJU$j$N(Bticks$B?t(B=division*4*nn/dd$B$J$N$G!#(B
	  const int measureAdd = //$B2?>.@a?J$s$@$+(B
	    (playTicks-cacheSongStatus->lastMeasureTicks)*
	    cacheSongStatus->getDd ()/(uint)cacheSongStatus->getDivision()/
	    cacheSongStatus->getNn ()/4;
	  if (measureAdd>0)
	    {
	      cacheSongStatus->lastMeasureTicks += measureAdd*
		cacheSongStatus->getDivision ()*cacheSongStatus->getNn ()*4/
		cacheSongStatus->getDd ();
	      cacheSongStatus->lastMeasureMachineTicks = machineTicks;
	      cacheSongStatus->lastMeasureLogicalTicks = logicalTicks;
	      cacheSongStatus->measures += measureAdd;
	    }
	  cacheSongStatus->ticksFromLastMeasure =
	    playTicks-cacheSongStatus->lastMeasureTicks;

	  assert (dMachineTicks >= 0);
	  assert (dLogicalTicks >= 0);
      
	  if (static_cast<ulong> (machineTicks+dMachineTicks)-static_cast<ulong>(machineTicks)>0)
	    {//$BBT$DI,MW$,$"$C$?!#(B
	      //buffer$B$r<B:]$K=q$-=P$7$F!"1iAU$N>uBV$r99?7!#(B
	      try
		{
		  dumpBuffer (sequencer, lastTicks, playTicks,
			      machineTicks, dMachineTicks,
			      logicalTicks, dLogicalTicks,
			      lastDumpTicks);
		}
	      catch (SongCommand& command_)
		{
		  if (command_.peekCommand () == SongCommand::REWIND)
		    {//$B:G=i$KLa$C$F1iAU$7D>$9!#(B
		      realRewind ();
		      lastTicks = 0;
		      lastDumpTicks = 0;
		      machineTicks = 0;
		      dMachineTicks = 0;
		      logicalTicks = 0;
		      dLogicalTicks = 0;
		      setTempoTicks = 0;
		      oldTempo = cacheSongStatus->getTempo ();
		      nEndOfTracks = 0;
		      playTrack = 0;
		      sequencer.startTimer ();
		      continue;
		    }
		  else if (command_.peekCommand () == SongCommand::ABORT)
		    {
		      //XXX note off
		      cacheSongStatus->state = SongStatus::PLAYING;
		      break;//$B6/@)=*N;!#(B
		    }
		  else
		    {
		      //XXX$B8=:_$N=j(BSET$B$O(Bsupport$B$7$J$$!#(B
		      assert (command_.peekCommand () == SongCommand::SET);
		      cacheSongStatus->state = SongStatus::PLAYING;
		    }
		}
	      lastTicks = playTicks;
	      setTempoTicks = playTicks;
	      oldTempo = cacheSongStatus->getTempo ();
	    }
	}
      
      //event$B$r2r@O$9$k!#(B
      SMFTrack* track = smfTrack[playTrack];
      uchar tmpCommand = track->peekUChar ();
#ifdef DEBUG
      //cerr << "DEBUG:track=" << playTrack
      //<< " command=" << (int) ((uchar) tmpCommand) << endl;
#endif
      if ((tmpCommand&0xF0) != 0xF0)
	//normal MIDI Event
	doMidiEvent (sequencer, track, tmpCommand);
      else
	{//sys exclusive
	  track->seekgCur (1);
	  if (tmpCommand == 0xF0 || tmpCommand == 0xF7)
	    {//SysEx Event
	      //DEBUG_MESSAGE ("DEBUG:SysEx Message found.");
	      //off running status
	      track->offRunningStatus ();
	      uint32 length = track->getVariableLength ();
	      const uchar* tmp = (uchar*) track->getCurrentBufferPtr ();
	      track->seekgCur (length);
	      uchar* data = new uchar[length+1];
	      data[0] = 0xF0;
	      memcpy (data+1, tmp, length);
	      sequencer.exclusive (data, length+1);
	      delete[] data;
	    }
	  else if (tmpCommand == 0xFF)
	    {//Meta Event
	      doMetaEvent (playTrack, oldTempo, setTempoTicks, sequencer);
	      if (track->isEnd ()) nEndOfTracks++;
	      //end of track event$B$,$"$k$+$bCN$l$J$$$N$G!#(B
	    }
	  else
	    {
	      DEBUG_MESSAGE ("DEBUG:SMF::doPlay:crash!!!!!!!!!!!!!!!");
	      cerr << "track=" << playTrack 
		   << " tmpCommand=" << hex << (int) tmpCommand << dec << endl;
	    }
	}

      cacheSongStatus->events++;
      ///////////////////////////////////$B<!$N(Btick$B$r7W;;(B
      if (smfTrack[playTrack]->isEnd ())
	smfTrack[playTrack]->setEnd ();
      else
	smfTrack[playTrack]->getABSDeltaTicks ();
    }//while (nEndOfTracks<nTracks)
}

////////////////////////////////////////////////////////////event$B$N=hM}(B
void SMF::doMidiEvent (Sequencer& sequencer, SMFTrack* track, uchar tmpCommand)
{
  assert ((tmpCommand&0xF0) != 0xF0);
  const MIDI::PORT port = track->getPort ();
  if ((tmpCommand&0x80) != 0)
    {//midi event$B$@$C$?!#(B
      //running status$B$r3P$($F$*$/(B
      track->setRunningStatus (tmpCommand);
      uchar data[3];
#define CHANNEL (data[0] & 0x0F)
      switch (tmpCommand & 0xF0)
	{
	case 0x80:
	  track->read (data, 3);
	  sequencer.noteOff (port, CHANNEL, data[1], data[2]);
	  break;
	case 0x90:
	  track->read (data, 3);
	  sequencer.noteOn(port, CHANNEL, data[1], data[2]);
	  break;
	case 0xA0:
	  track->read (data, 3);
	  sequencer.polyphonicKeyPressure (port, CHANNEL, data[1], data[2]);
	  break;
	case 0xB0:
	  track->read (data, 3);
	  sequencer.controlChange (port, CHANNEL, data[1], data[2]);
#if 0
	  cerr << "controlChange " << hex
	       << (int) data[0] << ","
	       << (int) data[1] << "," << (int) data[2] << dec << endl;
#endif
	  break;
	case 0xC0:
	  track->read (data, 2);
	  sequencer.programChange (port, CHANNEL, data[1]);
#if 0
	  cerr << "programChange " << hex
	       << (int) data[0] << "," << (int) data[1] << dec << endl;
#endif
	  break;
	case 0xD0:
	  track->read (data, 2);
	  sequencer.channelPressure (port, CHANNEL, data[1]);
	  break;
	case 0xE0:
	  track->read (data, 3);
	  sequencer.pitchBendChange (port, CHANNEL, data[1], data[2]);
	  break;
	default:
	  DEBUG_MESSAGE ("DEBUG:SMF::doEvent There's Error!!!!!!!!!!!!");
	  cerr << " tmpCommand=" << hex << (int) tmpCommand << dec << endl;
	  break;
	}
    }
  else
    {//running status$B$@$C$?!#(B
      //DEBUG_MESSAGE ("DEBUG::running status");
      if (!track->isSetRunningStatus ())
	{
	  cerr << "unkown event type event=" << hex << track->get8 () << dec
	       << endl;
	}
      else
	{
	  char data[3];
	  data[0]=track->getRunningStatus ();//running status
	  switch (data[0]&0xF0)
	    {
	    case 0x80:
	      track->read (data+1, 2);
	      sequencer.noteOff (port, CHANNEL, data[1], data[2]);
	      break;
	    case 0x90:
	      track->read (data+1, 2);
	      sequencer.noteOn(port, CHANNEL, data[1], data[2]);
	      break;
	    case 0xA0:
	      track->read (data+1, 2);
	      sequencer.polyphonicKeyPressure (port, CHANNEL, data[1], data[2]);
	      break;
	    case 0xB0:
	      track->read (data+1, 2);
	      sequencer.controlChange (port, CHANNEL, data[1], data[2]);
	      break;
	    case 0xC0:
	      track->read (data+1, 1);
	      sequencer.programChange (port, CHANNEL, data[1]);
	      break;
	    case 0xD0:
	      track->read (data+1, 1);
	      sequencer.channelPressure (port, CHANNEL, data[1]);
	      break;
	    case 0xE0:
	      track->read (data+1, 2);
	      sequencer.pitchBendChange (port, CHANNEL, data[1], data[2]);
	      break;
	    default:
	      DEBUG_MESSAGE ("DEBUG:SMF::doEvent Running Status "
			     "There's Error!!!!!!!!!!!!");
	      break;
	    }//switch (data[0]&0xF0)
	}
    }
#undef CHANNEL
}


/////////////////////////////////////////////////////////meta event$B$N=hM}!#(B
void SMF::doMetaEvent (int playTrack, uint32& oldTempo, uint32& setTempoTicks,
		       Sequencer& sequencer)
{
  switch (smfTrack[playTrack]->peekChar ())
    {
    case 0x00:
      sequencerNumber (playTrack);
      break;
    case 0x01:
    case 0x02:
    case 0x03:            
    case 0x04:
    case 0x05:
    case 0x06:
    case 0x07:
      text (playTrack);
      break;
    case 0x20:
      channelPrefix (playTrack);
      break;
    case 0x21:
      portChange (playTrack,sequencer);
      break;
    case 0x2F:
      endOfTrack (playTrack);
      break;
    case 0x51:
      oldTempo=cacheSongStatus->getTempo ();
      setTempoTicks=smfTrack[playTrack]->getABSTicks ();
      setTempo (playTrack);
      break;
      //SMPTE Offset
    case 0x54:
      smpteOffset (playTrack);
      break;
      //Time Signature
    case 0x58:
      timeSignature (playTrack);
      break;
      //Key Signature
    case 0x059:
      keySignature (playTrack);
      break;
      //Sequencer Specific
    case 0x7F:
      sequencerSpecific (playTrack);
      break;
    default:
      {
	ostrstream message;
	message << "unkown meta event:"
		<< "track=" << playTrack
		<< " event type="
		<< hex << smfTrack[playTrack]->get8 () << dec;
	uint32 length = smfTrack[playTrack]->getVariableLength ();
	message << " length=" << length
		<< " contents=";
	message.write (smfTrack[playTrack]->getCurrentBufferPtr (), length);
	message << ends;
	smfTrack[playTrack]->seekgCur (length);
	io.errorMessage (message.str ());
	message.freeze (0);
      }
      break;
    }//Meta Event
}

void SMF::sequencerNumber (int playTrack)
{
  smfTrack[playTrack]->seekgCur (2);
  ostrstream message;
  message << "seq num meta event=" << smfTrack[playTrack]->get16B ()
	  << ends;
  io.normalMessage (message.str ());
  message.freeze (0);
}

void SMF::text (int playTrack)
{
  const char* const messageType[] =
  {
    NULL,
    "Text:",
    "Copyright Notice:",
    "Sequence/Track Name:",
    "Instrument Name:",
    "Lyric:",
    "Marker:",
    "Cue Point:"
  };
  int type = smfTrack[playTrack]->get8 ();
  assert (1 <= type && type <= 7);
  ostrstream message;
  message << "track=" << playTrack
	  << ":" << messageType[type];
  int length = smfTrack[playTrack]->getVariableLength ();
  message.write (smfTrack[playTrack]->getCurrentBufferPtr (), length);
  message << ends;
  smfTrack[playTrack]->seekgCur (length);
  char* eucMessage=new char[message.tellp ()*2+1];//+256$B$O(B"track..."$B$NJ,!#(B
  convertSjis2Euc (message.str (), eucMessage);
  io.normalMessage (eucMessage);
  delete[] eucMessage;
  message.freeze (0);
}

void SMF::channelPrefix (int playTrack)
{
  DEBUG_MESSAGE ("DEBUG:channel prefix found.");
  smfTrack[playTrack]->seekgCur (2);
  ostrstream message;
  message << "midi channel prefix="
	  << smfTrack[playTrack]->get8 ()
	  << ":ignoring." << ends;
  io.normalMessage (message.str ());
  message.freeze (0);
}

void SMF::endOfTrack (int playTrack)
{
  smfTrack[playTrack]->seekgCur (2);
  smfTrack[playTrack]->setEnd ();
  /*
  ostrstream message;
  message << "end of track=" << playTrack << ends;
  io.normalMessage (message.str ());
  message.freeze (0);
  */
}

void SMF::setTempo (int playTrack)
{
  smfTrack[playTrack]->seekgCur (2);
  cacheSongStatus->setTempo (smfTrack[playTrack]->get24B ());
#ifdef DEBUG
  ostrstream message;
  message << "track=" << playTrack
	  << " set tempo=" << cacheSongStatus->getTempo () << ends;
  io.normalMessage (message.str ());
  message.freeze (0);
#endif
}

void SMF::smpteOffset (int playTrack)
{
  //XXX
  ///$BL$BP1~(B!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  smfTrack[playTrack]->seekgCur (2);
#ifdef DEBUG
  cout << "!!!!!!SMPTE offset"
       << " hr=" << smfTrack[playTrack]->get8 ()
       << " mn=" << smfTrack[playTrack]->get8 ()
       << " se=" << smfTrack[playTrack]->get8 ()
       << " fr=" << smfTrack[playTrack]->get8 ()
       << " ff=" << smfTrack[playTrack]->get8 ()
       << endl;
#else
  smfTrack[playTrack]->get8 ();
  smfTrack[playTrack]->get8 ();
  smfTrack[playTrack]->get8 ();
  smfTrack[playTrack]->get8 ();
  smfTrack[playTrack]->get8 ();
#endif
}

void SMF::timeSignature (int playTrack)
{
  smfTrack[playTrack]->seekgCur (2);
  const int8 nn_ = smfTrack[playTrack]->get8 ();
  const int8 dd_ = (1 << smfTrack[playTrack]->get8 ());
  const int8 cc_ = smfTrack[playTrack]->get8 ();
  const int8 bb_ = smfTrack[playTrack]->get8 ();
  cacheSongStatus->setTimeSignature (nn_, dd_, cc_, bb_);
  //$B$3$3$G$O>.@a$N;O$^$j$@$H2>Dj$7$F$$$k!#(B
#if DEBUG
  /*
  cout << "DEBUG:Time Signature"
       << " nn=" << cacheSongStatus->nn
       << " dd=" << cacheSongStatus->dd
       << " cc=" << cacheSongStatus->cc
       << " bb=" << cacheSongStatus->bb
       << endl;
       */
#endif
}

void SMF::keySignature (int playTrack)
{
  smfTrack[playTrack]->seekgCur (2);
  const int8 sf_ = smfTrack[playTrack]->get8 ();
  const int8 mi_ = smfTrack[playTrack]->get8 ();
  cacheSongStatus->setKeySignature (sf_, mi_);
#ifdef DEBUG
  /*
  cout << "DEBUG:key signature"
       << " sf=" << cacheSongStatus->sf
       << " mi=" << cacheSongStatus->mi
       << endl;
       */
#endif
}

void SMF::sequencerSpecific (int playTrack)
{
  smfTrack[playTrack]->seekgCur (1);
  const uint32 length = smfTrack[playTrack]->getVariableLength ();
  smfTrack[playTrack]->seekgCur (length);
  io.errorMessage ("sequencer specific meta event found.ignoring.");
}

//0xFF 0x21 0x01 pp(pp=port)
//this is NOT authorized smf meta event.
void SMF::portChange (int playTrack, Sequencer& sequencer)
{
  smfTrack[playTrack]->seekgCur (1);
  const uint32 length = smfTrack[playTrack]->getVariableLength ();
  const MIDI::PORT port = smfTrack[playTrack]->get8 ();
  smfTrack[playTrack]->setPort (port);
  smfTrack[playTrack]->seekgCur (length-1);
  io.errorMessage ("WARING:SMF::port change meta event found.");
}

void SMF::realRewind ()
{
  /////////////////////////////////////////track$B$r:G=i$N0LCV$K$9$k!#(B
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    {
      smfTrack[i]->offRunningStatus ();
      smfTrack[i]->rewind ();
      smfTrack[i]->getABSDeltaTicks ();
    }
}
