/*
    Song.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include "Song.h"
#if 0
#include "Song/dumb/SongStatusConvert2Dumb.h"
#endif
#ifdef LOCALLY_SCHEDULED_THREAD
# include "util/schedYield.h"
#endif

//for playing
Song::Song (State* state_, SongCommand* songCommand_,
	    MainIO& userIO, Option* option__):
  option (option__), state (state_), songCommand (songCommand_), io (userIO),
  isSeeking_ (false)
{
  assert (option != NULL);
  assert (state != NULL);
  assert (songCommand != NULL);
  state->lock ();
  state->fullReset ();
  state->unlock ();
  cacheSongStatus = new SongStatus ();
}

#if 0
//for converting
Song::Song (State* state_, SongCommand* songCommand_,
	    MainIOConvert2Dumb& userIO, Option* midiOption,
	    DumbEventList& dumbEventList_):
  option (midiOption), state (state_), songCommand (songCommand_), io (userIO),
  isSeeking_ (false)
{
  assert (option != NULL);
  assert (state != NULL);
  assert (songCommand != NULL);
  state->lock ();
  state->fullReset ();
  state->unlock ();
  cacheSongStatus = new SongStatusConvert2Dumb (&dumbEventList_);
}
#endif

Song::~Song ()
{
  delete cacheSongStatus;
}


void Song::loadFile (const string& fileName)// throw (Error)
{
  cacheSongStatus->setFileName (fileName);
  state->lock ();
  state->updateSongStatus (cacheSongStatus);
  state->unlock ();
  SongInputStream songInputStream(fileName);
  if (!songInputStream)
    {
      Error error ("can not open ", fileName);
      throw error;
    }

  loadFile (songInputStream);
}

void Song::loadFile (SongInputStream& songInputStream) // throw (Error)
{
  const int fileSize = songInputStream.getFileSize ();
#ifdef DEBUG
  //cerr << "file size = " << fileSize << endl;
#endif      

  char* buffer = new char[fileSize];
  songInputStream.read (buffer, fileSize);
  loadBuffer (buffer, fileSize);
  delete[] buffer;
}

void Song::resetDevice ()//???
{
}

//////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////play the data
//////////////////////////////////////////////////////////////////////////

void Song::play (Sequencer& sequencer) // throw (Error)
{
  DEBUG_MESSAGE ("Song::play.");
  //$B1iAU0LCV$r:G=i$K0LCV$5$;$k!#(B
  realRewind ();
  
  ///////////////////////////////////////device$B$N=i4|2=(B
  sequencer.reset ();
  sequencer.startTimer ();
  //////////////////////////////////////////////$B>uBVJQ?t$N=i4|2=(B
  cacheSongStatus->reset ();
  cacheSongStatus->rate = sequencer.getCtrlRate ();//machine$B$N(Btimer rate
  cacheSongStatus->state = SongStatus::PLAYING;
  timeval startTime;
  startTime.tv_sec = 0;
  startTime.tv_usec = 0;
  cacheSongStatus->addTime = startTime;
  gettimeofday (&startTime, NULL);
  cacheSongStatus->baseTime = startTime;
  state->lock ();
  state->updateSongStatus (cacheSongStatus);
  state->unlock ();
  state->broadcast ();

  DEBUG_MESSAGE ("Song::play doPlaying.");
  ////////////////////////////////////////////$B<B:]$K1iAU!#(B
  doPlay (sequencer);
  DEBUG_MESSAGE ("Song::play doPlayed");
  
  ////////////////////////////////////////////////////$B8e;OKv!#(B
  sequencer.dumpBuf ();//$B:G8e$K;D$C$?(Bbuffer$B$N(Bflush
  sequencer.sync ();//sync
  cacheSongStatus->state = SongStatus::DONE;
  updateStatus (sequencer);
  DEBUG_MESSAGE ("Song::play all done.");
}

void Song::realPause (Sequencer& sequencer, uint32 ticksOffset,
		      double machineTicksOffset,double logicalTicksOffset)
{
  DEBUG_MESSAGE ("Song::realPause");
  //XXX all note off$B$9$k!#(B
  sequencer.allNoteOffSoundOff ();
  sequencer.dumpBuf ();
  sequencer.stopTimer ();
  cacheSongStatus->state = SongStatus::PAUSING;
  cacheSongStatus->ticksOffset = ticksOffset;
  cacheSongStatus->machineTicksOffset = machineTicksOffset;
  cacheSongStatus->logicalTicksOffset = logicalTicksOffset;
  DEBUG_MESSAGE ("Song::realPause state updating");
  updateStatus (sequencer);
  DEBUG_MESSAGE ("Song::realPause songCommand updated");
  songCommand->lock ();
  songCommand->getSongCommand (&cacheSongCommand);
  while (cacheSongCommand.peekCommand () == SongCommand::PAUSE ||
	 cacheSongCommand.peekCommand () == SongCommand::NONE)
    {
      songCommand->waitCond ();
      songCommand->getSongCommand (&cacheSongCommand);
    }
  DEBUG_MESSAGE ("Song::realPause pausing done");
  songCommand->unlock ();
  DEBUG_MESSAGE ("Song::realPause pausing songCommand unlocked");
  songCommand->broadcast ();
  DEBUG_MESSAGE ("Song::realPause pausing all done");
}


void Song::realContinuePlay (Sequencer& sequencer)
{
  DEBUG_MESSAGE ("Song::realContinuePlay");
  assert (cacheSongStatus->state == SongStatus::PAUSING);
  cacheSongStatus->state = SongStatus::PLAYING;
  DEBUG_MESSAGE ("Song::realContinuePlay state updating");
  updateStatus (sequencer);
  DEBUG_MESSAGE ("Song::realContinuePlay state updated");
  //sequencer.continueTimer ();//this seems not to work
  sequencer.startTimer ();
}


void Song::dumpBuffer (Sequencer& sequencer,
		       uint32 lastTicks, uint32 playTicks,
		       double& machineTicks, double dMachineTicks,
		       double& logicalTicks, double dLogicalTicks,
		       double& lastDumpTicks)
{
  sequencer.dumpBuf ();
#define DUMP_TIME 5/100 //0.05sec
  if (isSeeking () == false &&
      option->isRealTime && lastDumpTicks>cacheSongStatus->rate*DUMP_TIME)
    {
#ifdef LOCALLY_SCHEDULED_THREAD
      schedYield ();
#endif
      sequencer.sync ();
      lastDumpTicks = 0;
    }
#undef DUMP_TIME
  if (isSeeking () == false)
    tryUpdateStatus (sequencer);
  cacheSongStatus->lastTicks = lastTicks;
  cacheSongStatus->ticks = playTicks;
  cacheSongStatus->machineTicks = machineTicks;
  cacheSongStatus->logicalTicks = logicalTicks;
  machineTicks += dMachineTicks;
  logicalTicks += dLogicalTicks;
  lastDumpTicks += dMachineTicks;
  if (isSeeking () == false)
    sequencer.waitTimer
      ((ulong)(machineTicks-cacheSongStatus->machineTicksOffset));
  else
    {//seeking$B;~$NF0:n(B
      //$B=jDj$N0LCV$^$GMh$F$$$l$P(B,seeking$B$r;_$a$FDL>o:F@8$KLa$k!#(B
      //not implemented yet
      if (cacheSongStatus->logicalTicks > getSeekPosition ())
	{//$BL\E*$N=j$^$G(Bseek$B$7$?!#(B
	  cacheSongStatus->ticksOffset = playTicks;
	  cacheSongStatus->machineTicksOffset = machineTicks;
	  cacheSongStatus->logicalTicksOffset = logicalTicks;
	  leaveSeeking (&sequencer);
	  sequencer.sync ();
	  cacheSongStatus->state = SongStatus::PLAYING;
	  sequencer.stopTimer ();
	  sequencer.startTimer ();
	  DEBUG_MESSAGE ("Song::dumpBuffer seeking done.");
	}
    }

  //PAUSE,RESTART$B$O$3$3$G=hM}$9$k(B
  //$B$=$l0J30$O(Bthrow$B$7$F$=$l$>$l$N=hM}$K$^$+$;$k!#(B
  //throw$B$5$l$k$N$O(BSET,REWIND,ABORT
  if (songCommand->trylock () == 0)
    {
      //DEBUG_MESSAGE ("Song::dumpBuffer songCommand locked");
      songCommand->getSongCommand (&cacheSongCommand);
      //DEBUG_MESSAGE ("Song::dumpBuffer songCommand gotten");
      songCommand->unlock ();
      //DEBUG_MESSAGE ("Song::dumpBuffer songCommand unlocked");
      songCommand->broadcast ();
      if (isSeeking () == true)
	return;
    EXECUTE_COMMAND:
      switch (cacheSongCommand.peekCommand ())
	{
	case SongCommand::NONE:
	  //do nothing
	  //assert (cacheSongStatus->state != SongStatus::PAUSING &&
	  //  cacheSongStatus->state != SongStatus::DOING_COMMAND) ;
	  break;
	case SongCommand::PAUSE:
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand PAUSE");
	  realPause (sequencer, playTicks, machineTicks, logicalTicks);
	  goto EXECUTE_COMMAND;
	  break;
	case SongCommand::CONTINUE:
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand CONTINUE");
	  realContinuePlay (sequencer);
	  break;
	case SongCommand::FORWARD://falling through
	case SongCommand::BACK:
	  {
	    DEBUG_MESSAGE ("Song::dumpBuffer songCommand::FORWARD,BACK");
	    ulong position = static_cast<ulong> (cacheSongStatus->logicalTicks);
#define SEEKING_UNIT 5 //5$BIC(B
	    if (cacheSongCommand.peekCommand () == SongCommand::FORWARD)
	      {
		position += SEEKING_UNIT*SongStatus::SONG_CTRL_RATE;
	      }
	    else
	      {//cacheSongCommand.peekCommand () == SongCommand::BACK
		position -= SEEKING_UNIT*SongStatus::SONG_CTRL_RATE;
	      }
#undef SEEKING_UNIT
	    cacheSongCommand.reset ();
	    cacheSongCommand.setCommandSET (position);
	    //falling throungh
	  }
	case SongCommand::SET:
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand SET");
#ifdef DEBUG
	  cerr << "now position=" << cacheSongStatus->logicalTicks
	       << " setTo=" << cacheSongCommand.getPosition ()
	       << endl;
#endif
	  if (cacheSongCommand.getPosition () == cacheSongStatus->logicalTicks)
	    break;//no need seeking
	  if (cacheSongCommand.getPosition () > cacheSongStatus->totalLogicalTicks)
	    {//$B6J$N=*$j$^$G(Bseek$B$9$k$N$G!":F@8$r=*$o$l$PNI$$!#(B
	      cacheSongCommand.reset ();
	      cacheSongCommand.setCommand (SongCommand::ABORT);
	      throw cacheSongCommand;
	    }

	  //$B$A$c$s$H(Bseek$B$9$kI,MW$,$"$k!#(B
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand SET real seeking.");
	  sequencer.allNoteOffSoundOff ();
	  setSeekPosition (cacheSongCommand.getPosition ());
	  enterSeeking (&sequencer);
	  if (getSeekPosition () < cacheSongStatus->logicalTicks)
	    {//$B4,$-La$7$J$N$G(B,$B:G=i$KLa$C$F(Bseek$B$9$k!#(B
	      DEBUG_MESSAGE ("Song::dumpBuffer songCommand SET rewinding.");
	      sequencer.reset ();
	      cacheSongStatus->reset ();
	      cacheSongCommand.reset ();
	      cacheSongCommand.setCommand (SongCommand::REWIND);
	      throw cacheSongCommand;
	    }
	  break;
	case SongCommand::REWIND:
	  cacheSongStatus->reset ();
	  cacheSongStatus->state = SongStatus::PLAYING;
	  //falling through
	case SongCommand::ABORT:
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand ABORT");
	  sequencer.reset ();
	  state->lock ();
	  state->reset ();
	  state->unlock ();
	  state->broadcast ();
	  throw (cacheSongCommand);
	  break;
	case SongCommand::SET_SKEW:
#ifdef DEBUG
	  DEBUG_MESSAGE ("Song::dumpBuffer songCommand::SET_SKEW");
	  cerr << "new skew = " << cacheSongCommand.getSkew () << endl;
#endif
	  cacheSongStatus->setSkew (cacheSongCommand.getSkew ());
	  break;


	  //$B$3$l$i$O(Binstance$B$,$I$N(BSong$B$N(Bsubclass$B$G$"$k$+$I$&$+$K4X$o$i$:(B
	  //$B=hM}$,$G$-$k$N$G0l3g$7$F=hM}$9$k!#(B
	case SongCommand::MIDI_PORT_NOTE_ON://falling throung
	case SongCommand::MIDI_PORT_NOTE_OFF://falling througn
	case SongCommand::MIDI_PORT_POLYPHONIC_KEY_PRESSURE://falling throung
	case SongCommand::MIDI_PORT_CONTROL_CHANGE://falling through
	case SongCommand::MIDI_PORT_RPN://falling through
	case SongCommand::MIDI_PORT_NRPN://falling through
	case SongCommand::MIDI_PORT_PROGRAM_CHANGE://falling through
	case SongCommand::MIDI_PORT_CHANNEL_PRESSURE://falling through
	case SongCommand::MIDI_PORT_PITCH_BEND_CHANGE://falling through
	case SongCommand::MIDI_PORT_SYS_EX://falling through
	case SongCommand::SET_MUTE://falling through
	case SongCommand::TOGGLE_MUTE:
	  cacheSongCommand.executeCommand (sequencer);
	  break;
	}
    }
}


///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////thread stuff
///////////////////////////////////////////////////////////////////////
Song::PlayInfo Song::playInfo;
Mutex Song::playInfoLock;

void* Song::startPlay (PlayInfo* playInfo)
{
  //DEBUG_MESSAGE ("DEBUG:Song::startPlay");
  assert (playInfo->seq != NULL);
  assert (playInfo->song != NULL);
  (playInfo->song)->play (*(playInfo->seq));
  return NULL;
}

int Song::createPlayThread (pthread_t* thread, pthread_attr_t* attr,
			    Sequencer& sequencer)
{
  //DEBUG_MESSAGE ("Song::createPlayThread now starting play thread");
  playInfoLock.lock ();
  playInfo.song = this;
  playInfo.seq = &sequencer;
  const int ret = pthread_create (thread, attr, startPlay, &playInfo);
  playInfoLock.unlock ();
  return ret;
}
