/*
    SongCommand.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include "SongCommand.h"

SongCommand::SongCommand ():
  sysExData (NULL), muting (NULL)
{
  reset ();
}

SongCommand::SongCommand (const SongCommand& songCommand_):
  sysExData (NULL), muting (NULL)
{
  command = songCommand_.command;
  position = songCommand_.position;
  skew = songCommand_.skew;
  memcpy (midiParameter, songCommand_.midiParameter,
	  sizeof (midiParameter));
  sysExSize = songCommand_.sysExSize;
  if (sysExSize > 0)
    {
      sysExData = new uchar[sysExSize];
      memcpy (sysExData,songCommand_.sysExData, sysExSize);
    }
  muteSize = songCommand_.muteSize;
  if (muteSize > 0)
    {
      assert (songCommand_.muting != NULL);
      muting = Sequencer::Muting::newCopy (songCommand_.muting, muteSize);
    }
}


SongCommand::~SongCommand ()
{
  delete[] sysExData;
  delete[] muting;
}


void SongCommand::executeCommand (Sequencer& sequencer)
{
  switch (peekCommand ())
    {
    case SongCommand::NONE:
    case SongCommand::PAUSE:
    case SongCommand::CONTINUE:
    case SongCommand::FORWARD://falling through
    case SongCommand::BACK:
    case SongCommand::SET:
    case SongCommand::REWIND:
    case SongCommand::ABORT:
    case SongCommand::SET_SKEW:
      //these command depends on Song class.
      //Should not be handled in this method.
      assert (false);
      break;
      
    case SongCommand::MIDI_PORT_NOTE_ON:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_NOTE_ON");
      sequencer.noteOn (midiParameter[0], midiParameter[1],
			midiParameter[2], midiParameter[3]);
      break;
    case SongCommand::MIDI_PORT_NOTE_OFF:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_NOTE_OFF");
      sequencer.noteOff (midiParameter[0], midiParameter[1],
			 midiParameter[2], midiParameter[3]);
      break;
    case SongCommand::MIDI_PORT_POLYPHONIC_KEY_PRESSURE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_POLYPHONIC_KEY_PRESSURE");
      sequencer.polyphonicKeyPressure (midiParameter[0],
				       midiParameter[1],
				       midiParameter[2],
				       midiParameter[3]);
      break;
    case SongCommand::MIDI_PORT_CONTROL_CHANGE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_CONTROL_CHANGE");
      //RPN,NRPN¸ϤʤΤ
      sequencer.controlChange (midiParameter[0], midiParameter[1],
			       midiParameter[2], midiParameter[3]);
      sequencer.controlChange (midiParameter[0], midiParameter[4],
			       midiParameter[5], midiParameter[6]);
      break;
    case SongCommand::MIDI_PORT_RPN:
      sequencer.changeRPN_NRPN (true,
				midiParameter[0], midiParameter[1],
				midiParameter[2], midiParameter[3],
				midiParameter[4], midiParameter[5],
				midiParameter[6], midiParameter[7]);
      break;
    case SongCommand::MIDI_PORT_NRPN:
      sequencer.changeRPN_NRPN (false,
				midiParameter[0], midiParameter[1],
				midiParameter[2], midiParameter[3],
				midiParameter[4], midiParameter[5],
				midiParameter[6], midiParameter[7]);
      break;
    case SongCommand::MIDI_PORT_PROGRAM_CHANGE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_PROGRAM_CHANGE");
      sequencer.programChange (midiParameter[0], midiParameter[1],
			       midiParameter[2]);
      break;
    case SongCommand::MIDI_PORT_CHANNEL_PRESSURE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_CHANNEL_PRESSURE");
      sequencer.channelPressure (midiParameter[0], midiParameter[1],
				 midiParameter[2]);
      break;
    case SongCommand::MIDI_PORT_PITCH_BEND_CHANGE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_PITCH_BEND_CHANGE");
      sequencer.pitchBendChange (midiParameter[0], midiParameter[1],
				 midiParameter[2], midiParameter[3]);
      break;
    case SongCommand::MIDI_PORT_SYS_EX:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::MIDI_PORT_SYS_EX");
      sequencer.exclusive (getSysExData (), getSysExSize ());
      break;
    case SongCommand::SET_MUTE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::SET_MUTE");
      sequencer.setMute (getMuting (), getMuteSize ());
      break;
    case SongCommand::TOGGLE_MUTE:
      DEBUG_MESSAGE ("SongCommand::executeCommand songCommand::TOGGLE_MUTE");
      sequencer.toggleMute (getMuting (), getMuteSize ());
      break;
    }
}
