/*
    MidiEvent.h
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#ifndef MidiEvent_H
#define MidiEvent_H

#include <string>
#include <cassert>

#include "Song/SongStatus.h"

using namespace std;

class MidiEventNode
{
public:
  enum EventType
  {
    //Song EVENT
    FORMAT,
    NTRACKS,
    DIVISION,
    TEMPO,
    BPM,
    TIME_SIGNATURE,
    TIME_SIGNATURE2,
    KEY_SIGNATURE,
    
    //MIDI EVENT
    MIDI_NOTE_OFF,
    MIDI_NOTE_ON,
    MIDI_KEY_PRESSURE,
    MIDI_CONTROL_CHANGE,
    MIDI_PROGRAM_CHANGE,
    MIDI_CHANNEL_PRESSURE,
    MIDI_PITCH_BEND,
    MIDI_SYS_EX,
    MIDI_PORT_CHANGE,
    
    //meta
    TEXT_OUT,
    END,
    //
    NOT_SET,//$B$^$@@_Dj$5$l$F$$$J$$!#(Bfor inner use only
  };
  EventType eventType;
  MidiEventNode ():eventType (NOT_SET) {};
  virtual ~MidiEventNode () {};
};

//Song Event
class MidiFormat:public MidiEventNode
{
public:
  SongStatus::Format format;
};

class MidiNTracks:public MidiEventNode
{
public:
  int16 nTracks;
};

class MidiDivision:public MidiEventNode
{
public:
  int16 division;
};

class MidiTempo:public MidiEventNode
{
public:
  uint32 tempo;
};

class MidiBPM:public MidiEventNode
{
public:
  uint32 bpm;
};

class MidiTimeSignature:public MidiEventNode
{
public:
  int nn;
  int dd;
  int cc;
  int bb;
};

class MidiTimeSignature2:public MidiEventNode
{
public:
  int nn;
  int dd;
};

class MidiKeySignature:public MidiEventNode
{
public:
  int sf;
  int mi;
};

class MidiNoteOff:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar note;
  uchar velocity;
};

class MidiNoteOn:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar note;
  uchar velocity;
};

class MidiKeyPressure:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar note;
  uchar keyPressure;
};

class MidiControlChange:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar control;
  uchar value;
};

class MidiProgramChange:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar  program;
};

class MidiChannelPressure:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar pressure;
};

class MidiPitchBend:public MidiEventNode
{
public:
  int port;
  uchar channel;
  uchar lsb;
  uchar msb;
};

class MidiSysEx:public  MidiEventNode
{
public:
  uint32 size;
  uchar* udata;
  //port, channel$B$OK\Mh$OITMW$@$,%]!<%H;XDj$N0Y$K$"$k!#(B-1$B$N;~$O%]!<%H;XDjL5$7!#(B
  int port;
  int channel;
};

class MidiPortChange:public  MidiEventNode
{
public:
  int port;
};

class MidiWaitTime:public MidiEventNode
{
public:
  ulong waitMachineTicks;
};

class MidiTextOut:public  MidiEventNode
{
public:
  string text;
};

class MidiEnd:public MidiEventNode
{
public:
};


class MidiEvent
{
protected:
  MidiEvent* next;
  MidiEventNode* event;
  uint32 ticks;//midi ticks
  double logicalTicks;//song ticks(1ms$BC10L(B)
  void setTicks (uint32 ticks_, double logicalTicks_)
    {
      ticks = ticks_;
      logicalTicks = logicalTicks_;
    };
public:
  MidiEvent ();
  virtual ~MidiEvent ();

  void setNext (MidiEvent* next_);
  const MidiEvent* getNext () const {return next;};
  const MidiEventNode* getEventNode () const
    {assert (event != NULL); return event;};
  uint32 getTicks () const {return ticks;};
  double getLogicalTicks () const {return logicalTicks;};
  
  //Song Event
  void setEventFormat (SongStatus::Format format,
		       uint32 ticks_, double logicalTicks_);
  void setEventNTracks (int16 nTracks, uint32 ticks_, double logicalTicks_);
  void setEventDivision (int16 division, uint32 ticks_, double logicalTicks_);
  void setEventTempo (uint32 tempo, uint32 ticks_, double logicalTicks_);
  void setEventBpm (uint32 bpm, uint32 ticks_, double logicalTicks_);
  void setEventTimeSignature (int nn, int dd, int cc, int dd,
			      uint32 ticks_, double logicalTicks_);
  void setEventTimeSignature2 (int nn, int dd,
			       uint32 ticks_, double logicalTicks_);
  void setEventKeySignature (int sf, int mi,
			     uint32 ticks_, double logicalTicks_);

  //midi event
  void setEventMidiNoteOff (int port, uchar channel,
			    uchar note, uchar velocity,
			    uint32 ticks_, double logicalTicks_);
  void setEventMidiNoteOn (int port, uchar channel, uchar note, uchar velocity,
			   uint32 ticks_, double logicalTicks_);
  void setEventMidiKeyPressure (int port, uchar channel,
				uchar note, uchar  keyPressure,
				uint32 ticks_, double logicalTicks_);
  void setEventMidiControlChange (int port, uchar channel,
				  uchar control, uchar value,
				  uint32 ticks_, double logicalTicks_);
  void setEventMidiProgramChange (int port, uchar channel, uchar program,
				  uint32 ticks_, double logicalTicks_);
  void setEventMidiChannelPressure (int port, uchar channel, uchar pressure,
				    uint32 ticks_, double logicalTicks_);
  void setEventMidiPitchBend (int port, uchar channel, uchar lsb, uchar msb,
			      uint32 ticks_, double logicalTicks_);
  void setEventMidiSysEx (const uchar* data, uint32 size,
			  uint32 ticks_, double logicalTicks_,
			  int port = -1, int channel=-1);
  void setEventMidiPortChange (int port, uint32 ticks_, double logicalTicks_);
  //meta
  void setEventTextOut (const char* data, int size,
			uint32 ticks_, double logicalTicks_);
  void setEventEnd (uint32 ticks_, double logicalTicks_);

  //for debug
  void printEvent () const;
};

#endif //MidiEvent_H
