/*
    G36.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <strstream>

#include "G36.h"
#include "G36Track.h"

using namespace std;

const char G36::G36_MAGIC[];
const int G36::G36_RYTHMS_NUMBER;

//for playing
G36::G36 (State* state_, SongCommand*  songCommand_, MainIO& userIO, Option* option_):
  RCP (state_, songCommand_, userIO, option_)
{
  //do nothing
}


G36::G36 (const string& fileName, State* state_, SongCommand* songCommand_,
	  MainIO& userIO, Option* option_):
  RCP (state_, songCommand_, userIO, option_)
{
  loadFile (fileName);
}

G36::G36 (SongInputStream& songInputStream,
	  State* state_, SongCommand* songCommand_, MainIO& userIO,
	  Option* option_):
  RCP (state_, songCommand_, userIO, option_)
{
  loadFile (songInputStream);
}

#if 0
//for converting
G36::G36 (const string& fileName, State* state_, SongCommand* songCommand_,
	  MainIOConvert2Dumb& userIO, Option* option_,
	  DumbEventList& dumbEventList_):
  RCP (fileName, state_, songCommand_, userIO, option_, dumbEventList_)
{
  //do nothing
}
#endif



G36::~G36 ()
{
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    delete rcpTrack[i];
  delete[] rcpTrack;
}

bool G36::isSupported (const string& fileName)
{
  try
    {
      SongInputStream songInputStream (fileName);
      char magic[G36_MAGIC_LENGTH+1];
      magic[G36_MAGIC_LENGTH] = '\0';
      songInputStream.read (magic, G36_MAGIC_LENGTH);
      songInputStream.close ();
      if (memcmp (magic, G36_MAGIC, G36_MAGIC_LENGTH) == 0)
	return true;
      else
	return false;
    }
  catch (Error& error)
    {
      cerr << error.getMessage ();
      return false;
    }

  return false;
}

RCPTrack* G36::newTrack ()
{
  return new G36Track ();
}

void G36::readHeader (RCPBufferStream& rcpBufferStream) throw (Error)
{
  //check magic number
  char magic[G36_MAGIC_LENGTH+1];
  magic[G36_MAGIC_LENGTH] = '\0';
  rcpBufferStream.read (magic, G36_MAGIC_LENGTH);
  if (memcmp (magic, G36_MAGIC, G36_MAGIC_LENGTH) != 0)
    {
      Error error ("this file is not .g18 or g36 file.magic=", magic);
      throw (error);
    }
#ifdef DEBUG
  cerr << "DEBUG:magic=" << magic << endl;
  cerr << "DEBUG:G36::readHeader title position=" << hex
       << rcpBufferStream.tellg () << dec << endl;
#endif
  rcpBufferStream.read (title, 64);//read title
  title[64] = '\0';
#ifdef DEBUG
  cerr << "DEBUG:titile=" << title << endl;
#endif
  rcpBufferStream.skipReservedArea (64, 0x20);
  rcpBufferStream.read (memo, 360);//read memo
  memo[360] = '\0';
#ifdef DEBUG
  cerr << "DEBUG:memo=" << memo << endl;
#endif

  //read track count
  //cacheSongStatus->nTracks = rcpBufferStream.get16L ();
  cacheSongStatus->setNTracks (rcpBufferStream.get16L ());
#ifdef DEBUG
  cerr << "DEBUG:tracks=" << cacheSongStatus->getNTracks () << endl;
#endif
#if 0
  if (cacheSongStatus->nTracks != 18 && cacheSongStatus->nTracks != 36)
    {
      ostrstream message;
      message << "track number is invalid? # of track=" << cacheSongStatus->nTracks;
      message << ends;
      io.errorMessage (message.str ());
      message.freeze (0);
    }
#endif

  //read time base 
  //cacheSongStatus->division = rcpBufferStream.get16L ();
  cacheSongStatus->setDivision (rcpBufferStream.get16L ());
  if (cacheSongStatus->getDivision () != 48  &&
      cacheSongStatus->getDivision () != 60  &&
      cacheSongStatus->getDivision () != 80  &&
      cacheSongStatus->getDivision () != 96  &&
      cacheSongStatus->getDivision () != 120 &&
      cacheSongStatus->getDivision () != 160 &&
      cacheSongStatus->getDivision () != 240 &&
      cacheSongStatus->getDivision () != 480)
    {
      io.errorMessage ("time base is invalid.assumig timebase is 48.");
      cacheSongStatus->setDivision (48);
    }
#ifdef DEBUG
  cerr << "DEBUG:division=" << cacheSongStatus->getDivision () << endl;
#endif

  //read bpm
  cacheSongStatus->setBpm (rcpBufferStream.get16L ());
  if (cacheSongStatus->getBpm ()<8 || 250<cacheSongStatus->getBpm ())
    {
      io.errorMessage ("invalid bpm.assuming bpm=120.");
      cacheSongStatus->setBpm (120);
    }
#ifdef DEBUG
  cerr << "DEBUG:bpm=" << cacheSongStatus->getBpm () << endl;
#endif

  const int nn_ = rcpBufferStream.get8 ();
  const int dd_ = rcpBufferStream.get8 ();
  cacheSongStatus->setTimeSignature (nn_, dd_);
  
#ifdef DEBUG
  cerr << "DEBUG:nn/dd=" << cacheSongStatus->getNn () << "/" << cacheSongStatus->getDd ()
       << endl;
#endif
  //read key
  setKey(rcpBufferStream.get8 ());
#ifdef DEBUG
  cerr << "DEBUG:key=" << cacheSongStatus->getKey () << endl;
#endif
  
  //read play bias
  playBias = rcpBufferStream.get8 ()-128;
#ifdef DEBUG
  cerr << "DEBUG:playBias=" << (int) playBias << endl;
#endif
  if (playBias < -36 || 36 < playBias)
    {
      io.errorMessage ("invalid play bias.assuming play bias=0.");
      playBias = 0;
    }
  
  rcpBufferStream.seekgCur (6);
  rcpBufferStream.skipReservedArea (16, 0x40);
  rcpBufferStream.skipReservedArea (112, 0x00);

  //read control file name(GSD)
  rcpBufferStream.read (controlFileNameGSD, RCP_MAX_FILENAME_LENGTH);
  controlFileNameGSD[RCP_MAX_FILENAME_LENGTH] = '\0';
  //read control file name(GSD1)
  rcpBufferStream.read (controlFileNameGSD1, RCP_MAX_FILENAME_LENGTH);
  controlFileNameGSD1[RCP_MAX_FILENAME_LENGTH] = '\0';
  //read control file name(CM6)
  rcpBufferStream.read (controlFileNameCM6, RCP_MAX_FILENAME_LENGTH);
  controlFileNameCM6[RCP_MAX_FILENAME_LENGTH] = '\0';
#ifdef DEBUG
  cerr << "DEBUG:GSD=" << controlFileNameGSD  << endl;
  cerr << "DEBUG:GSD1=" << controlFileNameGSD1  << endl;
  cerr << "DEBUG:CM6=" << controlFileNameCM6  << endl;
#endif

  rcpBufferStream.skipReservedArea (80, 0x00);

#ifdef DEBUG
  cerr << "DEBUG:G36::brefore read Rythm readHeader track position=" << hex
       << rcpBufferStream.tellg () << dec << endl;
#endif
  //read rythm area
  rythm = new Rythm*();
  for (int i = 0; i < G36_RYTHMS_NUMBER; i++)
    {
      rythm[i] = new Rythm ();
      rythm[i]->load (rcpBufferStream);
    }
#ifdef DEBUG
  cerr << "DEBUG:G36::brefore read user exclusive readHeader track position=" << hex
       << rcpBufferStream.tellg () << dec << endl;
#endif

  //read user defining exclusive data
  for (int i = 0; i < RCP_MAX_USER_EXCLUSIVE; i++)
    userExclusive[i].load (rcpBufferStream, true);

#ifdef DEBUG
  cerr << "DEBUG:G36::readHeader track position=" << hex
       << rcpBufferStream.tellg () << dec << endl;
#endif
}

void G36::getParameter (int playTrack, uchar& note,
			uint32& stepTime, uchar& gateTime, uchar& velocity)
{
  note = rcpTrack[playTrack]->get8 ();
  velocity = rcpTrack[playTrack]->get8 ();
  stepTime = rcpTrack[playTrack]->get16L ();
  gateTime = rcpTrack[playTrack]->get8 ();
  rcpTrack[playTrack]->seekgCur (1);
}



void G36::channelExclusive (Sequencer& sequencer, int playTrack,
			    const uchar command,
			    const uchar gateTime, const uchar velocity)
{
#ifdef DEBUG
  //cerr << "Channel excusive is called." << endl;
#endif
#define MAX_EXCLUSIVE_LENGTH 128
  uchar exclusive[MAX_EXCLUSIVE_LENGTH];
  int length = 0;
  exclusive[length++] = 0xF0;//start exclusive event
  uchar checkSum = 0;
  while (rcpTrack[playTrack]->peekUChar () == 0xF7)
    {
      rcpTrack[playTrack]->seekgCur (1);//skip (0xF0 or 0xF7)
      uchar data[5];
      rcpTrack[playTrack]->read(data, 5);
      for (int i = 0; i < 2 ;i++)
	{
	  switch (data[i])
	    {
	    case 0x80://gate data
	      checkSum += exclusive[length++] = gateTime;
	      break;
	    case 0x81://velocity
	      checkSum += exclusive[length++] = velocity;
	      break;
	    case 0x82://midi channel
	      checkSum += exclusive[length++] = rcpTrack[playTrack]->getRCPMidiChannel ();
	      break;
	    case 0x83://clear check sum
	      checkSum = 0;
	      break;
	    case 0x84://check sum
	      exclusive[length++] = 128-(checkSum & 0x7F);
	      break;
	    case 0xF7://end of exclusive
	      goto END_OF_EXCLUSIVE;
	      break;
	    default:
	      checkSum += exclusive[length++] = data[i];
	      break;
	    }
	}
    }
 END_OF_EXCLUSIVE:
  exclusive[length++] = 0xF7;//end of exclusive event
#ifdef DEBUG
  /*
  cerr << "channel ex=" << hex << endl;
  for (int i=0;i<length;i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
  */
#endif
  assert (length <= MAX_EXCLUSIVE_LENGTH);
  if (length >= 5)
    sequencer.exclusive (exclusive, length);
#undef MAX_EXCLUSIVE_LENGTH
}

void G36::externalCommand (int playTrack, uchar gateTime, uchar velocity)
{
  io.errorMessage ("G36::external command found.There's something wrong.");
}

void G36::bankAndProgramLSB
(Sequencer& sequencer, int playTrack, uchar gateTime, uchar velocity)
{
  ///////////////////////////bank  change
  sequencer.controlChange (rcpTrack[playTrack]->getMidiPort (),
			   rcpTrack[playTrack]->getMidiChannel (),
			   0x20, velocity);
  /////////////////////program change
  sequencer.programChange (rcpTrack[playTrack]->getMidiPort (),
			   rcpTrack[playTrack]->getMidiChannel (),
			   gateTime);
}


void G36::tempoChange (int playTrack, uint32& oldTempo, uint32& setTempoTicks,
		       uchar gateTime, uchar velocity)//XXX
{
#ifdef DEBUG
  //cerr << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!tempo change is called." << endl;
#endif
  if (velocity != 0)
    {
      cerr << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!tempo graduation is not supported yet." << endl;
      cerr << "ignoring.sorry." << endl;
    }
  setTempoTicks = rcpTrack[playTrack]->getABSTicks ();
  oldTempo = cacheSongStatus->getTempo ();
  //cacheSongStatus->tempo=(uint32) ((60.0*1000000)/(cacheSongStatus->bpm*gateTime/64));
  cacheSongStatus->setTempo ((uint32) ((60.0*1000000)/(cacheSongStatus->getBpm ()*gateTime/64)));
}


void G36::commentStart (int playTrack)//XXX
{
#ifdef DEBUG
  //cerr << "comment is called." << endl;
#endif
  rcpTrack[playTrack]->seekgCur (-6);
  ostrstream comment;
  comment << "G36 Comment:";
#if 0
  for (int i = 0; i < 5; i++)
    {
      char data[6];
      rcpTrack[playTrack]->read (data, 6);
      assert (data[0] == 0xF6||data[0] == 0xF7);
      comment.write (data+1, 5);
    }
#endif
  do
    {
      uchar command = rcpTrack[playTrack]->get8 ();
      assert (command == 0xF6 || command == 0xF7);
      char data[5];
      comment.write (data, 5);
    }while (rcpTrack[playTrack]->peekUChar () == 0xF7);
  comment << ends;
  io.normalMessage (comment.str ());
  comment.freeze (0);
}


void G36::sameMeasure (int playTrack)
{
  rcpTrack[playTrack]->seekgCur (-4);
  uchar data[4];
  rcpTrack[playTrack]->read (data, 4);
  const uint16 measureNumber = data[0]+(((int) (data[1])) << 8);
  const uint16 offsetAddress = (data[2]+(data[3] << 8))*6-0xF2;
  assert (measureNumber <= 999);
  assert (offsetAddress >= G36Track::G36_TRACK_HEADER_SIZE);
#ifdef DEBUG
  /*
  cerr << "same measure is called." << endl;
  cerr << "playTrack=" << playTrack << endl;
  cerr << "measureNumber=" << measureNumber << endl;
  cerr << "offsetAddress=" << offsetAddress << endl;
  cerr << "tellg=" << rcpTrack[playTrack]->tellg () << endl;
  cerr << "same measure nesting="
       << rcpTrack[playTrack]->measureStack.length () << endl;
       */
#endif
  rcpTrack[playTrack]->setSameMeasure (measureNumber, offsetAddress);
}
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////for debug
////////////////////////////////////////////////////////////////////////////
void G36::dumpTrack ()
{
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++ )
    {
      rcpTrack[i]->rewind ();
      cout << i << "-th track " << endl;
      while (rcpTrack[i]->getLeftLength () >= 4)
	{
	  uchar buff[6];
	  rcpTrack[i]->read (buff, 6);
	  cout << hex
	       << (int)buff[0] << " "
	       << (int)buff[1] << " "
	       << (int)buff[2] << " "
	       << (int)buff[3] << " "
	       << (int)buff[4] << " "
	       << (int)buff[5] << " "
	       << dec
	       << ":" << rcpTrack[i]->tellg ()
	       << endl;
	}
    }
}
