/*
    RCP.cc  
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include <strstream>

#include "RCP.h"
#include "util/KanjiConvert.h"
#include "Sequencer/MIDI.h"

using namespace std;

const int RCP::RCP_MAGIC_LENGTH;
const char RCP::RCP_MAGIC[RCP_MAGIC_LENGTH+1];
const int RCP::RCP_MAX_FILENAME_LENGTH;
const int RCP::MAX_RYTHMS;
const int RCP::RCP_RYTHMS_NUMBER;
const int RCP::RCP_MAX_USER_EXCLUSIVE;


//for playing
RCP::RCP (State* state_, SongCommand* songCommand_, MainIO& userIO,
	  Option* option_):
  Song (state_, songCommand, userIO, option_),
  rythm (NULL), rcpTrack (NULL)
{
  //do nothing
}


RCP::RCP (const string& fileName_, State* state_, SongCommand* songCommand_,
	  MainIO& userIO, Option* option_):
  Song (state_, songCommand_, userIO, option_), 
  rythm (NULL), rcpTrack (NULL)
{
  loadFile (fileName_);
}

RCP::RCP (SongInputStream& songInputStream,
	  State* state_, SongCommand* songCommand_, MainIO& userIO, 
	  Option* option_):
  Song (state_, songCommand_, userIO, option_), 
  rythm (NULL), rcpTrack (NULL)
{
  loadFile (songInputStream);
}

#if 0
//for converting
RCP::RCP (const string& fileName_, State* state_, SongCommand* songCommand_, 
	  MainIOConvert2Dumb& userIO, 
	  Option* option_, DumbEventList& dumbEventList_):
  Song (state_, songCommand_, userIO, option_, dumbEventList_), 
  rythm (NULL), rcpTrack (NULL)
{
  loadFile (fileName_);
}
#endif

RCP::~RCP ()
{
  for (int i = 0; i < RCP_RYTHMS_NUMBER; i++)
    delete rythm[i];
  delete[] rythm;
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    delete rcpTrack[i];
  delete[] rcpTrack;
}

bool RCP::isSupported (const string& fileName)
{
  try
    {
      SongInputStream songInputStream (fileName);
      if (songInputStream.getFileSize () <= RCP_LEAST_FILE_LENGTH)
	return false;
      char magic[RCP_MAGIC_LENGTH+1];
      memset (magic, 0, RCP_MAGIC_LENGTH+1);
      songInputStream.read (magic, RCP_MAGIC_LENGTH);
      songInputStream.close ();
      if (memcmp (magic, RCP_MAGIC, RCP_MAGIC_LENGTH) == 0)
	return true;
      else
	return false;
    }
  catch (Error& error)
    {
      cerr << error.getMessage ();
      return false;
    }

  return false;
}

RCPTrack* RCP::newTrack ()
{
  return new RCPTrack ();
}

void RCP::doLoadBuffer (const char* buffer, int bufferSize)// throw (Error)
{
#ifdef DEBUG
  //cerr << "loadBuffer bufferSize=" << bufferSize << endl;
#endif
  RCPBufferStream rcpBufferStream (buffer, bufferSize);
  readHeader (rcpBufferStream);//read header
  //read track data($B$D$^$k$H$3$m%G!<%?K\BN(B)
  rcpTrack = new RCPTrack*[cacheSongStatus->getNTracks ()];
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    {
#ifdef DEBUG
      //cerr << "read " << i << "-th track" << endl;
#endif
      rcpTrack[i] = newTrack ();
#if 0
      try
#endif
	{
	  rcpTrack[i]->readTrackData (rcpBufferStream);
	}
#if 0
      catch ()
	{
	  io.errorMessage ("RCP::doLoadBuffer unexpected end of file");
	  throw UnexpectedEndOfData ();//XXX
	}
#endif
      //???
      if (i < cacheSongStatus->getNTracks () -1 &&
	  rcpBufferStream.getLeftLength () == 0)
	{
	  //XXX
	  //io.errorMessage ();$B$r;H$&$Y$-!#(B
	  cerr << "RCP::doLoadBuffer Warning:r.rcp header says have "
	       << cacheSongStatus->getNTracks () << " tracks."
	       << "but having " << i+1 << " tracks." << endl;
	  cacheSongStatus->setNTracks (i+1);
	  break;
	}
    }
}

void RCP::readHeader (RCPBufferStream& rcpBufferStream) throw (Error)
{
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader file length=" << hex
  //       << rcpBufferStream.getBufferSize () << dec << endl;
#endif
#if 0
  if (rcpBufferStream.getLeftLength () <= 0x5B1)
    {
      cerr << "RCP::readHeader file is too short." << endl;
      //throw TooShortFiles();//XXX
    }
#endif
  
  //check magic number
  char magic[RCP_MAGIC_LENGTH+1];
  magic[RCP_MAGIC_LENGTH] = '\0';
  rcpBufferStream.read (magic, RCP_MAGIC_LENGTH);
  if (memcmp (magic, RCP_MAGIC, RCP_MAGIC_LENGTH) != 0)
    {
      Error error ("this file is not .rcp file.magic=", magic);
      throw (error);
    }
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader title position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  rcpBufferStream.read (title, 64);//read title
  title[64] = '\0';
  char eucTitle[128+1];
  convertSjis2Euc (title, eucTitle);
  assert (cacheSongStatus->isTitleSet () == false);
  cacheSongStatus->setTitle (eucTitle);
  rcpBufferStream.read (memo, 336);//read memo
  memo[336] = '\0';
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader reserved 0x40 X 16 position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  for (int i = 0;i<16;i++)
    {
      if (rcpBufferStream.get8 () != 0x40)
	{
#if 0
	  io.errorMessage ("RCP::readHeader unexpected data header.");
	  throw (error);
#endif
	}
    }
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader time base lsb position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  //read time base LSB
  uchar timeBase = rcpBufferStream.get8 ();
  if (timeBase != 48 && timeBase != 60  && timeBase != 80  &&
      timeBase != 96 && timeBase != 120 && timeBase != 160 &&
      timeBase != 240)
    {
      io.errorMessage ("time base is invalid.assumig timebase is 48.");
      timeBase = 48;
    }
#ifdef DEBUG
  //  cerr << "RCP::readHeader::cacheSongStatus->division="
  //       << cacheSongStatus->division << endl;
#endif
  
  //read bpm
  cacheSongStatus->setBpm (rcpBufferStream.get8 ());
  if (cacheSongStatus->getBpm ()<8 || 250<cacheSongStatus->getBpm ())
    {
      io.errorMessage ("invalid bpm.assuming bpm=120.");
      cacheSongStatus->setBpm (120);
    }

  const int nn_ = rcpBufferStream.get8 ();//read $BGo;R$NJ,;R(B
  const int dd_ = rcpBufferStream.get8 ();//read $BGo;R$NJ,Jl(B
  cacheSongStatus->setTimeSignature (nn_, dd_);
  
  //read key
  setKey(rcpBufferStream.get8 ());

  //read play bias
  playBias = rcpBufferStream.get8 ()-128;
  //???
  if (playBias < -36 || 36 < playBias)
    {
      io.errorMessage ("invalid play bias.assuming play bias=0.");
      playBias = 0;
    }
  
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader control file position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  rcpBufferStream.read (controlFileNameCM6, RCP_MAX_FILENAME_LENGTH);
  //read control file name(CM6)
  controlFileNameCM6[RCP_MAX_FILENAME_LENGTH] = '\0';
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader control file(GSD) position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  rcpBufferStream.read (controlFileNameGSD, RCP_MAX_FILENAME_LENGTH);
  //read control file name(GSD)
  controlFileNameGSD[RCP_MAX_FILENAME_LENGTH] = '\0';
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader reserved position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  //read # of track 
  cacheSongStatus->setNTracks (rcpBufferStream.get8 ());
  if (cacheSongStatus->getNTracks ()  == 0)//RCM v2.5 or earlyer
    cacheSongStatus->setNTracks (18);//XXX
  if (cacheSongStatus->getNTracks () != 18 && cacheSongStatus->getNTracks () != 36)
    {
      io.errorMessage ("track number is invalid.assuming # of track=18");
      cacheSongStatus->setNTracks (18);//XXX
    }
  
  uchar timeBaseMSB = rcpBufferStream.get8 ();//read time base MSB
  if (timeBaseMSB == 0x00||timeBaseMSB == 0x01)
    {
      io.errorMessage ("time base MSB is invalid.assuming 0x00");
      timeBaseMSB = 0;
    }
  timeBase+=(timeBaseMSB << 8);
  cacheSongStatus->setDivision (timeBase);
  rcpBufferStream.skipReservedArea (14);//skip reserved area
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader tone name tb position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  ////tone name TB?
  //???
  for (int i = 0; i < 32/2; i++)
    {
      uchar tmp;
      tmp = rcpBufferStream.get8 ();
      toneNameTB[i] = (tmp&0x0F);
      toneNameTB[i+32/2] = (tmp >> 4);
    }
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader rythm position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
  //read rythm area
  rythm = new Rythm*[RCP_RYTHMS_NUMBER];
  for (int j = 0; j < RCP_RYTHMS_NUMBER; j++)
    {
      rythm[j] = new Rythm ();
      rythm[j]->load (rcpBufferStream);
    }
  
  
#ifdef DEBUG
  //cerr << "DEBUG:RCP::readHeader user exclusive position=" << hex
  //<< rcpBufferStream.tellg () << dec << endl;
#endif
  //read user defining exclusive data
  for (int i = 0; i < RCP_MAX_USER_EXCLUSIVE; i++)
    userExclusive[i].load (rcpBufferStream, false);
#ifdef DEBUG
  //  cerr << "DEBUG:RCP::readHeader track position=" << hex
  //       << rcpBufferStream.tellg () << dec << endl;
#endif
}

void RCP::setKey (uchar keyData)
{
  key = keyData;
  if (key>0x1F)
    {
      io.errorMessage ("invalid key found.assuming key is C Major.");
      key = 0;
    }
  const int8 mi_ = (key >= 10)?1:0;
  const char convertToSMFsf[] = 
  {
    0, 1, 2, 3, 4, 5, 6, 7, //major #
    0, -1, -2, -3, -4, -5, -6, -7, -8, //major b
    0, //dummy
    0, 1, 2, 3, 4, 5, 6, 7, //minor #
    0, -1, -2, -3, -4, -5, -6, -7, -8, //minor b
  };
  const int8 sf_ = convertToSMFsf[key];
  cacheSongStatus->setKeySignature (sf_, mi_);
}

void RCP::doPlay (Sequencer& sequencer)
{
  //DEBUG_MESSAGE ("RCP::doPlay.");
  uint32 lastTicks = 0;//$BA02s(BwaitTime$B$7$?$H$-$N@dBP(Btick$B?t!#(B
  double lastDumpTicks = 0;//$BA02s(BdumpBuf$B$7$F$+$i$N(Bmachine tick$B?t!#(B
  double machineTicks = 0;//$B8=:_1iAU$7$F$$$k@dBP(Bmachine ticks
  double dMachineTicks = 0;//$B7W;;MQ!#(B
  double logicalTicks = 0;//$B8=:_1iAU$7$F$$$k@dBP(Blogical ticks
  double dLogicalTicks = 0;//$B7W;;MQ!#(B
  uint32 setTempoTicks = 0;//set tempo$B$7$?$H$-$N;~4V7W;;$NJd@5MQ!#(B
  uint32 oldTempo = cacheSongStatus->getTempo ();//set tempo$B$7$?$H$-$N;~4V7W;;$NJd@5MQ!#(B
  cacheSongStatus->events = 0;//midi event$B$N?t(B
  uint16 nEndOfTracks = 0;//$B1iAU$N=*N;$7$?%H%i%C%/?t(B
  int playTrack = 0;//$B8=:_CmL\$7$F$$$k%H%i%C%/(B
  bool isNoteOff = false;//RCP$B$N(Bevent$B$h$j$b(Bnote off$B$,@h$K$"$k$+(B
  
  /////////$B<B:]$N1iAU(B  
  while (nEndOfTracks<cacheSongStatus->getNTracks () || isNoteOff)//isNoteOff$B$,$"$k$N$O:G8e$N(Bnote off$B$^$G$A$c$s$H1iAU$9$k0Y!#(B
    {
      //DEBUG_MESSAGE ("DEBUG:now playing.");

      /////$B$I$N%H%i%C%/$,:G=i$K1iAU$5$l$k$+0?$O!"(Bnote off$B$+H=Dj!#(B
      for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
	{
	  if ((rcpTrack[i]->getABSTicks ()<rcpTrack[playTrack]->getABSTicks ())
	      && (!rcpTrack[i]->isEnd ()))
	    playTrack = i;
	}
      uint32 playTicks = rcpTrack[playTrack]->getABSTicks ();//$B<!$N%$%Y%s%H$,5/$-$k(Bticks
      //note off$B$N0Y$NA0=hM}!#(B
      uint32 nextNoteOffTicks = doNoteOff (sequencer, 0);//$B<!$N(Bnote off$B$N(Bticks$B$r=P$9!#(B
      isNoteOff = false;
      if (nextNoteOffTicks <= playTicks-lastTicks)
      //if (nextNoteOffTicks<playTicks-lastTicks)
	{//note off event$B$NJ}$,@h$K5/$-$k!#(B
	  playTicks = lastTicks+nextNoteOffTicks;
	  isNoteOff = true;
	}

      
      //$B;~4VBT$A$N=hM}!#(B
      //$B;~4VBT$A$,H/@8$7$?$i(Bbuffer$B$r(Bdump$B$9$k!#(B
      //$BBT$A;~4V7W;;!#(BMIDI ticks$B?t$+$i(Bmachine ticks$B?t$r7W;;!#(B
      dMachineTicks = 0;
      dLogicalTicks = 0;
      assert (lastTicks <= playTicks);
      if (lastTicks<playTicks)
	{
	  dMachineTicks =
	    calcDMachineTicks (playTicks, lastTicks, setTempoTicks,
			       cacheSongStatus->getTempo (), oldTempo,
			       cacheSongStatus->rate,
			       cacheSongStatus->getDivision (),
			       cacheSongStatus->getSkew ());
	  dLogicalTicks =
	    calcDLogicalTicks (playTicks, lastTicks, setTempoTicks,
			       cacheSongStatus->getTempo (), oldTempo,
			       cacheSongStatus->getDivision ());

	  
	  cacheSongStatus->ticksFromLastMeasure = 
	    playTicks-cacheSongStatus->lastMeasureTicks;
	  if (static_cast<ulong> (machineTicks+dMachineTicks)-static_cast<ulong>(machineTicks)>0)
	    {//$BBT$DI,MW$,$"$C$?!#(B
	      //buffer$B$r<B:]$K=q$-=P$7$F!"1iAU$N>uBV$r99?7!#(B
	      try
		{
		  dumpBuffer (sequencer, lastTicks, playTicks, 
			      machineTicks, dMachineTicks,
			      logicalTicks, dLogicalTicks,
			      lastDumpTicks);
		}
	      catch (SongCommand& command_)
		{
		  if (command_.peekCommand () == SongCommand::REWIND)
		    {//$B:G=i$KLa$C$F1iAU$7D>$9!#(B
		      realRewind ();
		      lastTicks = 0;
		      lastDumpTicks = 0;
		      machineTicks = 0;
		      dMachineTicks = 0;
		      logicalTicks = 0;
		      dLogicalTicks = 0;
		      setTempoTicks = 0;
		      oldTempo = cacheSongStatus->getTempo ();
		      nEndOfTracks = 0;
		      playTrack = 0;
		      isNoteOff = false;
		      cacheSongStatus->state = SongStatus::PLAYING;
		      sequencer.startTimer ();
		      continue;
		    }
		  else if (command_.peekCommand () == SongCommand::ABORT)
		    {
		      //XXX note off
		      cacheSongStatus->state = SongStatus::PLAYING;
		      break;//$B6/@)=*N;!#(B
		    }
		  else
		    {
		      //XXX$B8=:_$N=j(BSET$B$O(Bsupport$B$7$J$$!#(B
		      assert (command_.peekCommand () == SongCommand::SET);
		      cacheSongStatus->state = SongStatus::PLAYING;
		    }
		}
	      for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
		{
		  for (int note = 0; note < Part::MAX_NOTES; note++)
		    {
		      if (rcpTrack[i]->gateTimes[note] >= 0)
			{
			  rcpTrack[i]->gateTimes[note] -= (playTicks-lastTicks);
			  assert (rcpTrack[i]->gateTimes[note] >= 0);
			}
		    }
		}
	      lastTicks = playTicks;
	      setTempoTicks = playTicks;
	      oldTempo = cacheSongStatus->getTempo ();
	      nextNoteOffTicks = 0;
	    }
	}

      //event$B$r2r@O$9$k!#(B
      if (isNoteOff)
	{
	  (void) doNoteOff (sequencer, nextNoteOffTicks);//note off$B$@$C$?!#(B
	}
      else
	{
	  doEvent (sequencer, playTrack, oldTempo, setTempoTicks,
		   nEndOfTracks, machineTicks, logicalTicks);
	  cacheSongStatus->events++;
	}
     }//while (nEndOfTracks<nTracks)
}


uint32 RCP::doNoteOff (Sequencer& sequencer, uint32 noteOffOffset)
{
  ////note off$B$N=hM}!#(B
  //noteOffOffset$B$h$j>.$5$$(Bgate time$B$r;}$DJ*$r(Bnote off$B$9$k!#(B
  //$BJV$jCM$O(BnoteOffOffset$B$h$jBg$-$$(Bgate time$B$NCf$G:GBg$NJ*!#(B
  //$B$3$3$O(BrcpTrack[i]->keyPlus$B$OL54X78!#(B
  uint32 nextNoteOffTicks = ~static_cast<uint32> (0);
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    {
      for (int note = 0; note < Part::MAX_NOTES; note++)
	{
	  if (rcpTrack[i]->gateTimes[note] != RCPTrack::RCP_NOTE_OFF)
	    {
	      if ((uint32) rcpTrack[i]->gateTimes[note] <= noteOffOffset)
		{
		  //note off$B$N=hM}!#(B
		  sequencer.noteOff (rcpTrack[i]->getMidiPort (),
				     rcpTrack[i]->getMidiChannel (),
				     note, 0x40);
		  rcpTrack[i]->gateTimes[note] = RCPTrack::RCP_NOTE_OFF;
		}
	      else
		{
		  //$B<!$N(B0$B0J>e$N(Bnote off$B$N;~4V$r7W;;(B
		  if ((uint32) rcpTrack[i]->gateTimes[note]<nextNoteOffTicks)
		    nextNoteOffTicks = rcpTrack[i]->gateTimes[note];
		}
	    }
	}
    }
  return nextNoteOffTicks;
}

void RCP::getParameter (int playTrack, uchar& note,
			uint32& stepTime, uchar& gateTime, uchar& velocity)
{
  note = rcpTrack[playTrack]->get8 ();
  stepTime = rcpTrack[playTrack]->get8 ();
  gateTime = rcpTrack[playTrack]->get8 ();
  velocity = rcpTrack[playTrack]->get8 ();
}


void RCP::doEvent (Sequencer& sequencer, int playTrack,
		   uint32& oldTempo, uint32& setTempoTicks,
		   uint16& nEndOfTracks,
		   double machineTicks, double logicalTicks)
{
  //DEBUG_MESSAGE ("RCP::doEvent");
  if (rcpTrack[playTrack]->peekUChar () >= 0x80)
    {//rcp$B$N(Bcommand$B$@$C$?!#(B
      doRCPCommand (sequencer, playTrack, oldTempo, setTempoTicks,
		    nEndOfTracks, machineTicks, logicalTicks);
      return;
    }
  
  ////////////////////////////////////////////////////////////note on
  uchar note;
  uint32 stepTime;
  uchar gateTime;
  uchar velocity;
  getParameter (playTrack, note, stepTime, gateTime, velocity);
  if (rcpTrack[playTrack]->keyPlus != RCPTrack::RCP_KEYPLUS_OFF)
    {
      note += rcpTrack[playTrack]->keyPlus+playBias;
      //note %= 128;
      note &= 0x7F;
    }
  
  //assert (velocity>0);
  assert (velocity <= 127);
  //assert (note >= 0);
  assert (note <= 127);

  if (rcpTrack[playTrack]->gateTimes[note] == RCPTrack::RCP_NOTE_OFF && velocity>0)
    {
      sequencer.noteOn (rcpTrack[playTrack]->getMidiPort (),
			rcpTrack[playTrack]->getMidiChannel (),
			note, velocity);
      rcpTrack[playTrack]->gateTimes[note] = gateTime;
    }
  else if (rcpTrack[playTrack]->gateTimes[note] != RCPTrack::RCP_NOTE_OFF && velocity == 0)
    {
      sequencer.noteOn (rcpTrack[playTrack]->getMidiPort (),
			rcpTrack[playTrack]->getMidiChannel (),
			note, velocity);
      rcpTrack[playTrack]->gateTimes[note] = RCPTrack::RCP_NOTE_OFF;
    }
  else if (rcpTrack[playTrack]->gateTimes[note] != RCPTrack::RCP_NOTE_OFF && velocity>0)
    {
      rcpTrack[playTrack]->gateTimes[note] = gateTime;
    }
#if 0
  else
    {
      //(rcpTrack[playTrack]->gateTimes[note] == RCPTrack::RCP_NOTE_OFF && velocity == 0)
      //XXX
      //???$B2?8N$3$N>l9g$,$"$k(B?
      sequencer.noteOn (rcpTrack[playTrack]->getMidiPort (),
			rcpTrack[playTrack]->getMidiChannel (),
			note, velocity);
      rcpTrack[playTrack]->gateTimes[note] = RCPTrack::RCP_NOTE_OFF;
    }
#endif

  rcpTrack[playTrack]->addTicks (stepTime);
}

void RCP::doRCPCommand (Sequencer& sequencer, int playTrack,
			uint32& oldTempo, uint32& setTempoTicks, 
			uint16& nEndOfTracks,
			double machineTicks, double logicalTicks)
{
  uchar command;
  uint32 stepTime;
  uchar gateTime;
  uchar velocity;
  getParameter (playTrack, command, stepTime, gateTime, velocity);
  switch (command)
    {
      ///////////////////////////////////////////////////////user definition
    case 0x90://user exclusive 1 $BD>2<$X(B
    case 0x91://user exclusive 2 $BD>2<$X(B
    case 0x92://user exclusive 3 $BD>2<$X(B
    case 0x93://user exclusive 4 $BD>2<$X(B
    case 0x94://user exclusive 5 $BD>2<$X(B
    case 0x95://user exclusive 6 $BD>2<$X(B
    case 0x96://user exclusive 7 $BD>2<$X(B
    case 0x97://user exclusive 8
      doUserExclusive (sequencer, playTrack, command, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0x98://channel exclusive
      channelExclusive (sequencer, playTrack, command, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0x99://$B30It(Bccommand$B$N<B9T(B
      //XXX
      externalCommand (playTrack, gateTime, velocity);
      break;

      ////////////////////////////////////////////////////////////MIDI Event
    case 0xE1://bank & program LSB
      bankAndProgramLSB (sequencer, playTrack, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xE2://bank & program MSB
      bankAndProgramMSB (sequencer, playTrack, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xE5://key scan
      keyScan (sequencer, playTrack, gateTime);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xE6://midi ch chnage
      midiChannelChange (playTrack, gateTime);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xE7://tempo change
      tempoChange (playTrack, oldTempo, setTempoTicks, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xEA://after touch (ch)
      afterTouch (sequencer, playTrack, gateTime);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xEB://control change
      controlChange (sequencer, playTrack, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xEC://program change
      programChange (sequencer, playTrack, gateTime);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xED://after touch poriphonic
      afterTouchPolyphonic (sequencer, playTrack, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xEE://pitch bend
      pitchBend (sequencer, playTrack, gateTime, velocity);
      rcpTrack[playTrack]->addTicks (stepTime);
      break;

      //////////////////////////////////////////////////////SysEx
    case 0xD0://YAMAHA base
      //XXX
      yamahaBase (playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xD1://YAMAHA Device#
      //XXX
      yamahaDeviceID (playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xD2://YAMAHA Addr/Para
      //XXX
      yamahaAddressAndParameter (sequencer, playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xD3://YAMAHA XG Addr/Para
      //XXX
      yamahaXG (sequencer, playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xDD://Roland Base
      //XXX
      rolandBase (playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xDE://Roland Para
      //XXX
      rolandParameter (sequencer, playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
    case 0xDF://Roland Device
      //XXX
      rolandDeviceAndModelID (playTrack, gateTime, velocity);//XXX
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
      
      //////////////////////////////////////////////////////////$B@)8f4X78!#(B
    case 0xF5://key change
      //XXX
      keyChange (playTrack, stepTime);
      break;
    case 0xF6://comment start
      commentStart (playTrack);
      break;
    case 0xF7:
#ifdef DEBUG
      cerr << "RCP::doPlay:0xF7 command found.There's error(s)." <<endl;
      cerr << "track=" << playTrack
	   << ":position=" << rcpTrack[playTrack]->tellg () << endl;
      {
	ostrstream message;
	message << (char) gateTime << (char) velocity;
	while (rcpTrack[playTrack]->peekUChar () == 0xF7)
	  {
	    uchar command_;
	    uint32 stepTime_;
	    uchar gateTime_;
	    uchar velocity_;
	    getParameter (playTrack, command_, stepTime_, gateTime_, velocity_);
	    message << (char) gateTime_ << (char) velocity_;
	  }
	message << ends;
	cerr << message.str ();
	message.freeze (0);
      }
#endif
      break;
    case 0xF8://loop end
      loopEnd (playTrack, stepTime);
      break;
    case 0xF9://loop start
      loopStart (playTrack);
      break;
    case 0xFC://same measure
      sameMeasure (playTrack, machineTicks, logicalTicks);
      break;
    case 0xFD://mesure end
      measureEnd (playTrack, machineTicks, logicalTicks);
      break;
    case 0xFE://end of track
      endOfTrack (playTrack, nEndOfTracks);
      break;

      //////////////////////////////////////////////////////////$B0J2<$O(Bunsupport
    case 0xC0://DX7 function
      //XXX
      //break;
    case 0xC1://DX parameter
      //XXX
      //break;
    case 0xC2://DX rerf
      //XXX
      //break;
    case 0xC3://TX function
      //XXX
      //break;
    case 0xC5://FB-01 P parameter
      //XXX
      //break;
    case 0xC6://FB-01 S System
      //XXX
      //break;
    case 0xC7://TX81Z V VCED
      //XXX
      //break;
    case 0xC8://TX81Z A ACED
      //XXX
      //break;
    case 0xC9://TX81Z PPCED
      //XXX
      //break;
    case 0xCA://TX81Z S System
      //XXX
      //break;
    case 0xCB://TX81Z E EFFECT
      //XXX
      //break;
    case 0xCC://DX702 R REMOTE SW
      //XXX
      //break;
    case 0xCD://DX7-2 A ACED
      //XXX
      //break;
    case 0xCE://DX7-2 P PCED
      //XXX
      //break;
    case 0xCF://TX802 P PCED
      //XXX
      //break;
      ///////////////////////////////////////////////////////////////////
    case 0xDC://MKS-7
      //XXX
      {
	uchar data[4];
	rcpTrack[playTrack]->seekgCur (-4);
	rcpTrack[playTrack]->read (data, 4);
#ifdef DEBUG
      cerr << "this command is not supported." << endl;
      cerr << "data=" << hex
	   << (int)data[0] << " "
	   << (int)data[1] << " "
	   << (int)data[2] << " "
	   << (int)data[3] << " "
	   << dec << endl;
#endif
      }
      rcpTrack[playTrack]->addTicks (stepTime);
      break;
	
    default:
      DEBUG_MESSAGE ("DEBUG:RCP::doRCPCommand:unkown command!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
      break;
    }
}

/////////////////////////////////////////////////////RCP command
void RCP::doUserExclusive (Sequencer& sequencer,
			   int playTrack, const uchar command,
			   const uchar gateTime, const uchar velocity)
{
  //DEBUG_MESSAGE ("User excusive is called.");
  const uchar userExclusiveNumber = command-0x90;
  //command 0x90-0x97$B$,(B0$BHV(B-7$BHV$N(Buser exclusive$B$KBP1~!#(B
  uchar exclusive[UserExclusive::RCP_MAX_USER_EXCLUSIVE_LENGTH+2];
  const uchar length = userExclusive[userExclusiveNumber].storeData
    (exclusive, gateTime, velocity, rcpTrack[playTrack]->getRCPMidiChannel ());
#ifdef DEBUG
  /*
  cerr << "user ex=" << hex << endl;
  for (int i=0;i<length;i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
  */
#endif
  if (length>0)
    sequencer.exclusive (exclusive, length,
			 rcpTrack[playTrack]->getMidiPort (),
			 rcpTrack[playTrack]->getMidiChannel ());
}

void RCP::channelExclusive (Sequencer& sequencer,
			    int playTrack, const uchar command,
			    const uchar gateTime, const uchar velocity)
{
  //DEBUG_MESSAGE ("Channel excusive is called.");
#define MAX_EXCLUSIVE_LENGTH 128
  uchar exclusive[MAX_EXCLUSIVE_LENGTH];
  int length = 0;
  exclusive[length++] = 0xF0;//start exclusive event
  uchar checkSum = 0;
  while (rcpTrack[playTrack]->peekUChar () == 0xF7)
    {
      rcpTrack[playTrack]->seekgCur (2);//skip (0xF0 or 0xF7) and step time
      uchar data[2];
      data[0] = rcpTrack[playTrack]->get8 ();
      data[1] = rcpTrack[playTrack]->get8 ();
      for (int i = 0; i < 2; i++)
	{
	  switch (data[i])
	    {
	    case 0x80://gate data
	      checkSum += exclusive[length++] = gateTime;
	      break;
	    case 0x81://velocity
	      checkSum += exclusive[length++] = velocity;
	      break;
	    case 0x82://midi channel
	      checkSum += exclusive[length++] = rcpTrack[playTrack]->getRCPMidiChannel ();//XXX$B$^$:$$$+$J$!!)(B
	      break;
	    case 0x83://clear check sum
	      checkSum = 0;
	      break;
	    case 0x84://check sum
	      exclusive[length++] = 128-(checkSum & 0x7F);
	      break;
	    case 0xF7://end of exclusive
	      goto END_OF_EXCLUSIVE;
	      break;
	    default:
	      checkSum+=exclusive[length++] = data[i];
	      break;
	    }
	}
    }
 END_OF_EXCLUSIVE:
  exclusive[length++] = 0xF7;//end of exclusive event
#ifdef DEBUG
  /*
  cerr << "channel ex=" << hex << endl;
  for (int i=0;i<length;i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
  */
#endif
  assert (length <=  MAX_EXCLUSIVE_LENGTH);
  sequencer.exclusive (exclusive, length,
		       rcpTrack[playTrack]->getMidiPort (),
		       rcpTrack[playTrack]->getMidiChannel ());
#undef MAX_EXCLUSIVE_LENGTH
}

void RCP::externalCommand (int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("external command is not supporeted.");
  while (rcpTrack[playTrack]->peekUChar () == 0xF7)
    {
      rcpTrack[playTrack]->seekgCur (4);
    }
}

void RCP::bankAndProgramLSB
(Sequencer& sequencer, int playTrack, uchar gateTime, uchar velocity)
{
  io.errorMessage
    ("RCP:bankAndProgramLSB command found.There's something wrong.");
}


void RCP::bankAndProgramMSB (Sequencer& sequencer, int playTrack,
			     uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("bankAndProgram is called.");
  /////////////////////bank change
  const int port =  rcpTrack[playTrack]->getMidiPort ();
  const uchar channel = rcpTrack[playTrack]->getMidiChannel ();
  sequencer.controlChange (port, channel, 0x00, velocity);
  sequencer.controlChange (port, channel, 0x20, 0x00);
  /////////////////////program change
  sequencer.programChange (port, channel, gateTime);
}

void RCP::keyScan
(Sequencer& sequencer, int playTrack, uchar gateTime)
{
#ifdef DEBUG
  cerr << "DEBUG:key scan is not supporeted." << endl;
  cerr << " scan track no=" << (int)gateTime << endl;
#endif
}

void RCP::midiChannelChange (int playTrack, uchar gateTime)
{
  if (gateTime == 0x00)
    {
      cerr << "midi channel is off.track=" << playTrack << endl;
      cerr << "is not supported yet." << endl;
      return;
    }
  if (gateTime>32)
    {
      cerr << "invalid channel.channel=" << (int) gateTime
	   << " ignored."<< endl;
      return;
    }
  
  //rcpTrack[playTrack]->midiChannel = gateTime-1;
  rcpTrack[playTrack]->setRCPMidiChannel (gateTime-1);
#ifdef DEBUG
  //cerr << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!midi channel change is called.channel=" << (int) gateTime << endl;
#endif
}

void RCP::tempoChange (int playTrack,
		       uint32& oldTempo, uint32& setTempoTicks, 
		       uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!tempo change is called.");
  if (velocity != 0)
    {
      io.errorMessage ("tempo graduation is not supported yet.");
      io.errorMessage ("ignoring.sorry.");
    }
  setTempoTicks = rcpTrack[playTrack]->getABSTicks ();
  oldTempo = cacheSongStatus->getTempo ();
  cacheSongStatus->setTempo ((uint32) ((60.0*1000000)/
				       (cacheSongStatus->getBpm ()*gateTime/64)));
}

void RCP::afterTouch (Sequencer& sequencer, int playTrack, uchar gateTime)
{
  //DEBUG_MESSAGE ("after touch is called.");
  sequencer.channelPressure (rcpTrack[playTrack]->getMidiPort (),
			     rcpTrack[playTrack]->getMidiChannel (),
			     gateTime);
}

void RCP::controlChange
(Sequencer& sequencer, int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("control change is called.");
  sequencer.controlChange (rcpTrack[playTrack]->getMidiPort (),
			   rcpTrack[playTrack]->getMidiChannel (),
			   gateTime, velocity);
}

void RCP::programChange (Sequencer& sequencer, int playTrack, uchar gateTime)
{
  //DEBUG_MESSAGE ("program change is called.");
  sequencer.programChange (rcpTrack[playTrack]->getMidiPort (),
			   rcpTrack[playTrack]->getMidiChannel (),
			   gateTime);
}

void RCP::afterTouchPolyphonic
(Sequencer& sequencer, int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("after touch polyphonic is called.");
  sequencer.polyphonicKeyPressure (rcpTrack[playTrack]->getMidiPort (),
				   rcpTrack[playTrack]->getMidiChannel (),
				   gateTime, velocity);
}

void RCP::pitchBend
(Sequencer& sequencer, int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("pitch bend polyphonic is called.");
  sequencer.pitchBendChange (rcpTrack[playTrack]->getMidiPort (),
			     rcpTrack[playTrack]->getMidiChannel (),
			     gateTime, velocity);
}

////////////////////////////////////////YAMAHA & ROLAND SysEx
void RCP::yamahaBase (int playTrack, uchar gateTime, uchar velocity)//???
{
  DEBUG_MESSAGE ("yamaha base is called.");
  rcpTrack[playTrack]->yamahaAddressHigh = gateTime;//address high
  rcpTrack[playTrack]->yamahaAddressMiddle = velocity;//address middle
  rcpTrack[playTrack]->isYamahaAddressInited = true;
}

void RCP::yamahaDeviceID (int playTrack, uchar gateTime, uchar velocity)//???
{
  DEBUG_MESSAGE ("yamaha device id is called.");
  rcpTrack[playTrack]->yamahaDeviceID = gateTime;//device ID
  rcpTrack[playTrack]->yamahaModelID = velocity;//model ID
  rcpTrack[playTrack]->isYamahaIDInited = true;
}

void RCP::yamahaAddressAndParameter (Sequencer& sequencer, int playTrack,
				     uchar gateTime, uchar velocity)//???
{
  DEBUG_MESSAGE ("yamaha address is called.");
  const uchar yamahaAddressLow = gateTime;
  const uchar parameter = velocity;
  if (!rcpTrack[playTrack]->isYamahaAddressInited)
    {
      io.errorMessage ("yamaha address is not specified.ingnoring");
      return;
    }
  if (!rcpTrack[playTrack]->isYamahaIDInited)
    {
      io.errorMessage
	("yamaha device ID and model ID are not specified.ingnoring");
      return;
    }

  uchar exclusive[9];
  exclusive[0] = 0xF0;//start of exclusive message
  exclusive[1] = MIDI::YAMAHA_ID;//yamaha
  exclusive[2] = rcpTrack[playTrack]->yamahaDeviceID;//device id
  exclusive[3] = rcpTrack[playTrack]->yamahaModelID;//model id
  exclusive[4] = rcpTrack[playTrack]->yamahaAddressHigh;//$B>e0L%"%I%l%9(B
  exclusive[5] = rcpTrack[playTrack]->yamahaAddressMiddle;//$BCf0L%"%I%l%9(B
  exclusive[6] = yamahaAddressLow;//$B2<0L%"%I%l%9(B
  exclusive[7] = parameter;
  exclusive[8] = MIDI::EOX;

#ifdef DEBUG
  cerr << "yamaha ex=" << hex << endl;
  for (int i = 0; i < 9; i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
#endif
  sequencer.exclusive (exclusive, 9,
		       rcpTrack[playTrack]->getMidiPort (),
		       rcpTrack[playTrack]->getMidiChannel ());
}

void RCP::yamahaXG (Sequencer& sequencer, int playTrack,
		    uchar gateTime, uchar velocity)//???
{
  DEBUG_MESSAGE ("yamaha XG id is called.");
  const uchar yamahaAddressLow = gateTime;
  const uchar parameter = velocity;
  if (!rcpTrack[playTrack]->isYamahaAddressInited)
    {
      io.errorMessage ("yamaha address is not specified.ingnoring");
      return;
    }
  if (!rcpTrack[playTrack]->isYamahaIDInited)
    {
      io.errorMessage
	("yamaha device ID and model ID are not specified.ingnoring");
      return;
    }

  uchar exclusive[9];
  exclusive[0] = 0xF0;//start of exclusive message
  exclusive[1] = MIDI::YAMAHA_ID;//yamaha
  exclusive[2] = rcpTrack[playTrack]->yamahaDeviceID;//device id
  exclusive[3] = rcpTrack[playTrack]->yamahaModelID;//model id
  exclusive[4] = rcpTrack[playTrack]->yamahaAddressHigh;//$B>e0L%"%I%l%9(B
  exclusive[5] = rcpTrack[playTrack]->yamahaAddressMiddle;//$BCf0L%"%I%l%9(B
  exclusive[6] = yamahaAddressLow;//$B2<0L%"%I%l%9(B
  exclusive[7] = parameter;
  exclusive[8] = MIDI::EOX;

#ifdef DEBUG
  cerr << "XG ex=" << hex << endl;
  for (int i = 0; i < 9; i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
#endif
  sequencer.exclusive (exclusive, 9,
		       rcpTrack[playTrack]->getMidiPort (),
		       rcpTrack[playTrack]->getMidiChannel ());
}

void RCP::rolandBase (int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("roland base is called.");
  rcpTrack[playTrack]->rolandAddressHigh = gateTime;//address high
  rcpTrack[playTrack]->rolandAddressMiddle = velocity;//address middle
  rcpTrack[playTrack]->isRolandAddressInited = true;
}

void RCP::rolandParameter (Sequencer& sequencer, int playTrack,
			   uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("roland parameter is called.");
  const uchar rolandAddressLow = gateTime;
  const uchar parameter = velocity;
  if (!rcpTrack[playTrack]->isRolandAddressInited)
    {
      io.errorMessage ("roland address is not specified.ingnoring");
      return;
    }
  if (!rcpTrack[playTrack]->isRolandIDInited)
    {
      io.errorMessage
	("roland device ID and model ID are not specified.ingnoring");
      return;
    }
  uchar exclusive[11];
  exclusive[0] = 0xF0;//start of exclusive message
  exclusive[1] = MIDI::ROLAND_ID;//roland
  exclusive[2] = rcpTrack[playTrack]->rolandDeviceID;//device id
  exclusive[3] = rcpTrack[playTrack]->rolandModelID;//model id
  exclusive[4] = MIDI::DT1;//command id
  uchar checkSum = 0;
  checkSum += exclusive[5] = rcpTrack[playTrack]->rolandAddressHigh;//$B>e0L%"%I%l%9(B
  checkSum+=exclusive[6] = rcpTrack[playTrack]->rolandAddressMiddle;//$BCf0L%"%I%l%9(B
  checkSum+=exclusive[7] = rolandAddressLow;//$B2<0L%"%I%l%9(B
  checkSum+=exclusive[8] = parameter;//data
  exclusive[9] = 128-(checkSum&0x7F);//check sum
  exclusive[10] = MIDI::EOX;//end of exclusive
#ifdef DEBUG
  /*
  cerr << "roland ex=" << hex << endl;
  for (int i=0;i<11;i++)
    cerr << " " << (int) exclusive[i];
  cerr << dec << endl;
  */
#endif
  sequencer.exclusive (exclusive, 11,
		       rcpTrack[playTrack]->getMidiPort (),
		       rcpTrack[playTrack]->getMidiChannel ());
}

void RCP::rolandDeviceAndModelID (int playTrack, uchar gateTime, uchar velocity)
{
  //DEBUG_MESSAGE ("roland device and moded id is called.");
  rcpTrack[playTrack]->rolandDeviceID = gateTime;
  rcpTrack[playTrack]->rolandModelID = velocity;
  rcpTrack[playTrack]->isRolandIDInited = true;
}


/////////////////////////////////////////////////////////////////$B@)8f4XO"(B
void RCP::keyChange (int playTrack, uint32 stepTime)
{
  //DEBUG_MESSAGE ("key change is called.");
  const uchar keyData = stepTime&0xFF;
  setKey (keyData);
}


void RCP::commentStart (int playTrack)
{
  //DEBUG_MESSAGE ("comment is called.");
  rcpTrack[playTrack]->seekgCur (-4);
  ostrstream comment;
  comment << "Track=" << playTrack << ":RCP Comment:";
  do
    {
      const uchar command = rcpTrack[playTrack]->get8 ();
      assert (command == 0xF6||command == 0xF7);
      rcpTrack[playTrack]->seekgCur (1);//skip step time
      comment << static_cast<char> (rcpTrack[playTrack]->get8 ());
      comment << static_cast<char> (rcpTrack[playTrack]->get8 ());
    }while (rcpTrack[playTrack]->peekUChar () == 0xF7);
  comment << ends;
  char* eucMessage = new char[comment.tellp ()*2+1];
  convertSjis2Euc (comment.str (), eucMessage);
  io.normalMessage (eucMessage);
  delete[] eucMessage;
  comment.freeze (0);
}

void RCP::loopEnd (int playTrack, uint32 loopCount)
{
  //DEBUG_MESSAGE ("loop end is called.");
  //assert (loopCount >= 1);
  //assert (loopCount>1);
  if (rcpTrack[playTrack]->loopStack.empty ())
    {
      //???
      //io.errorMessage ("RCP:loop end is found.but loop start"
      //" is not appeared.Miss mathcing loop."
      //"ignoring loop end.");
      return;
    }
  if (rcpTrack[playTrack]->getLeftLoopCount () == -1)
    rcpTrack[playTrack]->setLoopCount (loopCount);
  //$B$O$8$a$F(Bloop end$B$KMh$?$N$G%k!<%W2s?t$r@_Dj$9$k!#(B

  rcpTrack[playTrack]->decrimentLoopCount ();//loop$B2s?t$r8:$i$9!#(B
  if (rcpTrack[playTrack]->getLeftLoopCount ()>0)
    rcpTrack[playTrack]->setLoopStart ();//loop$B$,;D$C$F$$$?(B
  else
    rcpTrack[playTrack]->popLoopStack ();//loop$B=*$j(B
}

void RCP::loopStart (int playTrack)
{
  //DEBUG_MESSAGE ("loop start is called.");
  rcpTrack[playTrack]->pushLoopStack ();
}

void RCP::sameMeasure (int playTrack, double machineTicks, double logicalTicks)
{
  //same measure$B$,=*$o$C$?$H$3$m$r3P$($F$*$/!#(B
  const int curIndex = rcpTrack[playTrack]->tellg ();

  //??? same measure(0xFC)$B$ND>A0$K$O(Bmeasure end(0xFD)$BKt$O(B, same measure(0xFC)
  //$B$,F~$k$H$$$&2>Dj$O@.N)$9$k(B?
  //$B$=$&$G$J$$%G!<%?$b$"$kMM$J$N$G$3$N%A%'%C%/$rF~$l$F(B
  //$B6/@)E*$K(Bmeasure end$B$H8+$J$9!#(B
  rcpTrack[playTrack]->seekgCur (-8);//rewind
  if (rcpTrack[playTrack]->peekUChar () != 0xFC &&
      rcpTrack[playTrack]->peekUChar () != 0xFD)
    {
#ifdef DEBUG
      cerr << "WARNING:RCP::sameMeasure same measure event(0xFC) found.but last Event is not end of measure(0xFD)!!!" 
	   << ":track=" << playTrack << ":position=" << rcpTrack[playTrack]->tellg () << endl;
#endif
      if (rcpTrack[playTrack]->isInSameMeasure ())
	{//same measure$B$@$C$?$N$G!"$3$l$G=*$o$j(B
	  measureEnd (playTrack, machineTicks, logicalTicks);
	  return;
	}
      else
	measureEnd (playTrack, machineTicks, logicalTicks);
    }
  rcpTrack[playTrack]->seekgCur (8);//$B85$KLa$9!#(B

  rcpTrack[playTrack]->seekgCur (-4);
  assert (rcpTrack[playTrack]->peekUChar () == 0xFC);
  uint16 measureNumber = 0;
  uint16 offsetAddress = 0;
  //same measure$B$N@h$,(Bsame measure$B$@$C$?$i99$KAL$k!#(B
  while (rcpTrack[playTrack]->peekUChar () == 0xFC)
    {
      uchar data[4];
      rcpTrack[playTrack]->read (data, 4);
      measureNumber = data[1]+(((int) (data[2]&3)) << 8);
      offsetAddress = (data[2] & 0xFC)|(data[3] << 8);
      assert (measureNumber <= 999);
      assert (offsetAddress  >=  RCPTrack::RCP_TRACK_HEADER_SIZE);      
      rcpTrack[playTrack]->setg (offsetAddress-RCPTrack::RCP_TRACK_HEADER_SIZE);
    }
  rcpTrack[playTrack]->setg (curIndex);
#ifdef DEBUG
  /*
  cerr << "same measure is called." << endl;
  cerr << "playTrack=" << playTrack << endl;
  cerr << "measureNumber=" << measureNumber << endl;
  cerr << "offsetAddress=" << offsetAddress << endl;
  cerr << "tellg=" << rcpTrack[playTrack]->tellg () << endl;
  cerr << "same measure nesting="
       << rcpTrack[playTrack]->measureStack.length () << endl;
       */
#endif
  rcpTrack[playTrack]->setSameMeasure (measureNumber, offsetAddress);
}

void RCP::measureEnd (int playTrack, double machineTicks, double logicalTicks)
{
  //DEBUG_MESSAGE ("measure end is called.");
   cacheSongStatus->measures = ++(rcpTrack[playTrack]->measures);
   cacheSongStatus->lastMeasureTicks = rcpTrack[playTrack]->getABSTicks ();
   cacheSongStatus->lastMeasureMachineTicks = machineTicks;
   cacheSongStatus->lastMeasureLogicalTicks = logicalTicks;
   cacheSongStatus->ticksFromLastMeasure = 0;
   if (rcpTrack[playTrack]->isInSameMeasure ())
     rcpTrack[playTrack]->unsetSameMeasure ();
}

void RCP::endOfTrack (int playTrack, uint16& nEndOfTracks)
{
  //DEBUG_MESSAGE ("end of track is called.");
  rcpTrack[playTrack]->setEnd ();
  nEndOfTracks++;

#ifdef DEBUG
  ostrstream message;
  message << "end of track=" << playTrack << ends;
  io.normalMessage (message.str ());
  message.freeze (0);
#endif
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////for debug
////////////////////////////////////////////////////////////////////////////
void RCP::dumpTrack ()
{
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++ )
    {
      rcpTrack[i]->rewind ();
      cout << i << "-th track " << endl;
      while (rcpTrack[i]->getLeftLength () >= 4)
	{
	  uchar buff[4];
	  rcpTrack[i]->read (buff, 4);
	  cout << hex
	       << (int)buff[0] << " "
	       << (int)buff[1] << " "
	       << (int)buff[2] << " "
	       << (int)buff[3] << " "
	       << dec
	       << ":" << rcpTrack[i]->tellg ()
	       << endl;
	}
    }
}


void RCP::realRewind ()
{
  /////////////////////////////////////////track$B$r:G=i$N0LCV$K$9$k!#(B
  for (int i = 0; i < cacheSongStatus->getNTracks (); i++)
    {
      rcpTrack[i]->rewind ();
    }
}
