/*
    RCPTrack.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include "RCPTrack.h"

const int RCPTrack::RCP_TRACK_HEADER_SIZE;
const uchar RCPTrack::RCP_RYTHM_ON;
const uchar RCPTrack::RCP_RYTHM_OFF;
const char RCPTrack::RCP_KEYPLUS_OFF;
const uchar RCPTrack::RCP_MODE_PLAY;
const uchar RCPTrack::RCP_MODE_MUTE;
const uchar RCPTrack::RCP_MODE_MIX;
const uchar RCPTrack::RCP_MODE_REC;
const int RCPTrack::MAX_COMMENT_LENGTH;
const int32 RCPTrack::RCP_NOTE_OFF;


RCPTrack::RCPTrack ():
  SongBufferStream (),
  isYamahaAddressInited (false),isYamahaIDInited (false),
  isRolandAddressInited (false),isRolandIDInited (false)
{
  for (int i=0;i<Part::MAX_NOTES;i++)
    gateTimes[i]=RCP_NOTE_OFF;
}

RCPTrack::~RCPTrack ()
{
  //do nothing
}

void RCPTrack::readTrackData (RCPBufferStream& rcpBufferStream) throw (Error)
{
  //read data size
  const uchar lsb=rcpBufferStream.get8 ();
  const uchar msb=rcpBufferStream.get8 ();
  dataSize=((msb << 8)+lsb);
  
#ifdef DEBUG
  //cerr << "DEBUG:RCPTrack:track size=" << dataSize
  //     << " to " << hex << rcpBufferStream.tellg ()+dataSize << endl;
#endif
  trackIndex=rcpBufferStream.get8 ();//read track number
  rythm=rcpBufferStream.get8 ();//read rythm ON=0x80,OFF=0x00
  headerMidiChannel = rcpBufferStream.get8 ();//read midi channel
  
  //KEY#+$B$NCM$O(B                 (RCPTrack.cc:64$B9TL\$"$?$j(B)
  //127$B$h$jBg$-$1$l$P(B off        if (tmp&0x80) keyPlus = OFF;
  //64$B$h$j>.$5$1$l$P$=$N$^$^(B     else if(tmp<64) keyPlus = temp;
  //64$B0J>e$J$i(B-128               else keyPlus = tmp-128;
  //$B$H$J$k$H;W$$$^$9!#(B($B$?$V$s(B)
  //$B$=$l$H!"(BkeyPlus$B$,(BOFF$B$N$H$-$O(BplayBias$B$N1F6A$b<u$1$J$/$J$k$N$G!"(B
  //$B$=$NJU$b9MN8$7$F%N!<%H%*%s$9$kI,MW$,$"$k$H;W$$$^$9!#(B(RCP.cc 509$BJU$j(B)
  //thanx to $B:d0f(B@$BK!@/Bg(B(sakai@yoshi.ei.hosei.ac.jp)$B$5$s!#(B
  uchar const tmp=rcpBufferStream.get8 ();//read key+ OFF=0x80//???
  if (tmp&0x80)
    keyPlus = RCP_KEYPLUS_OFF;
  else if (tmp < 64)
    keyPlus = tmp;
  else
    keyPlus = tmp-128;
  
  stPlus=rcpBufferStream.get8 ();//read st+
  mode=rcpBufferStream.get8 ();//read mode
  rcpBufferStream.read (comment,MAX_COMMENT_LENGTH);//read comment

  //read track data
  loadBuffer
    (rcpBufferStream.getCurrentBufferPtr (),dataSize-RCP_TRACK_HEADER_SIZE);
  rcpBufferStream.seekgCur (dataSize-RCP_TRACK_HEADER_SIZE);

#ifdef DEBUG
  //cerr << "trackIndex=" << (int) trackIndex << endl;
  //cerr << "rythm=" << (int) rythm << endl;
  //cerr << "midiChannel=" << (int) midiChannel << endl;
  //cerr << "keyPlus=" << (int) keyPlus << endl;
  //cerr << "stPlus=" << (int) stPlus << endl;
  //cerr << "mode=" << (int) mode << endl;
#endif
}

uint32 RCPTrack::getNoteOffABSTicks (int* note) const
{
  *note=-1;
  if (isEnd ())
    return ~0;
  
  uint32 delta=peekStepTime ();
  for (int i=0;i<Part::MAX_NOTES;i++)
    {
      if ((uint32) gateTimes[i]<delta && gateTimes[i]!=RCP_NOTE_OFF)
	{
	  delta=gateTimes[i];
	  *note=i;
	}
    }
  return ticks+delta;
}

uint32 RCPTrack::peekStepTime () const
{
  return (uchar) buffer[index+1];
}

void RCPTrack::pushLoopStack ()
{
  Loop push;
  push.loopStart=tellg ();
  push.loopCount=-1;
  loopStack.push_back (push);
}

void RCPTrack::popLoopStack ()
{
  assert (!loopStack.empty ());
  assert ((*(loopStack.end ()-1)).loopCount==0);
  loopStack.pop_back ();
}

int RCPTrack::getLeftLoopCount ()
{
  assert (!loopStack.empty ());
  return (*(loopStack.end ()-1)).loopCount;
}


void RCPTrack::setLoopCount (uchar loopCount)
{
  assert (!loopStack.empty ());
  vector<Loop>::iterator nowLoop=loopStack.end ()-1;
  assert ((*nowLoop).loopCount==-1);
  (*nowLoop).loopCount=loopCount;
}

void RCPTrack::decrimentLoopCount ()
{
  assert (!loopStack.empty ());
  vector<Loop>::iterator nowLoop=loopStack.end ()-1;
  //assert (nowLoop->loopCount>0);//???
  if((*nowLoop).loopCount>0)//???
    (*nowLoop).loopCount--;
}

void RCPTrack::setLoopStart ()
{
  assert (!loopStack.empty ());
  vector<Loop>::iterator nowLoop=loopStack.end ()-1;
  assert ((*nowLoop).loopCount>0);
  index=(*nowLoop).loopStart;
}

void RCPTrack::setSameMeasure (uint16 measureNumber,uint16 offsetAddress)
{
  assert (RCP_TRACK_HEADER_SIZE<=offsetAddress);
  assert (offsetAddress<=dataSize);
  assert (!isInSameMeasure ());//???
  sameMeasure.fromIndex=tellg ();
  sameMeasure.sameMeasureNo=measureNumber;
  sameMeasure.offsetAddress=offsetAddress;
  sameMeasure.isInSameMeasure=true;
  setg (offsetAddress-RCP_TRACK_HEADER_SIZE);
}

void RCPTrack::unsetSameMeasure ()
{
  assert (isInSameMeasure ());
  setg (sameMeasure.fromIndex);
  sameMeasure.reset ();
}

void RCPTrack::rewind ()
{
  SongBufferStream::rewind ();
  midiChannel = headerMidiChannel;
  ticks=99+(signed char)stPlus;//???
  measures=0;
  loopStack.erase (loopStack.begin (),loopStack.end ());
  sameMeasure.reset ();
  isYamahaAddressInited=false;
  isYamahaIDInited=false;
  isRolandAddressInited=false;
  isRolandIDInited=false;
}
