/*
    SeverCommand.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include <iostream>

#include "util/mytype.h"
#include "ServerCommand.h"

using namespace std;

ServerCommand::ServerCommand ():
  commandLock (),commandCond (),command (NONE), position (0),
  skew (1.0), sysExSize (-1), sysExData (NULL), channel (-1),
  muteSize (-1),muting (NULL),
  playingMode (PlayingStyle::CONTINUOUS_PLAY),
  deviceResetType (DeviceResetType::NONE)
{
  reset ();
}

ServerCommand::~ServerCommand ()
{
  delete[] sysExData;
  delete[] muting;
}

void ServerCommand::getCommand (ServerCommand* serverCommand_)
{
  assert (serverCommand_ != NULL);
  serverCommand_->reset ();
  serverCommand_->command = command;
  serverCommand_->filename = filename;

  serverCommand_->position = position;
  serverCommand_->skew = skew;
  memcpy (serverCommand_->midiParameter, midiParameter, sizeof(midiParameter));
  serverCommand_->sysExSize = sysExSize;
  if (sysExSize > 0)
    {
      assert (sysExData != NULL);
      serverCommand_->sysExData = newCopySysEx (sysExData, sysExSize);
    }
  serverCommand_->channel = channel;
  serverCommand_->muteSize = muteSize;
  if (muting != NULL)
    {
      assert (muteSize > 0);
      serverCommand_->muting = Sequencer::Muting::newCopy (muting, muteSize);
    }

  serverCommand_->playingMode = playingMode;
  serverCommand_->deviceResetType = deviceResetType;
  
  reset ();
}

//command$B$,<u$1F~$l$i$l$k$^$GBT$D!#(B
void ServerCommand::waitLockCommand ()
{
  DEBUG_MESSAGE ("ServerCommand::waitLockCommand");
  lock ();
  DEBUG_MESSAGE ("ServerCommand::waitLockCommand locked");
  while (command != NONE)
    waitCond ();
  DEBUG_MESSAGE ("ServerCommand::waitLockCommand done");
}

void ServerCommand::setCommand (Command command_)
{
  DEBUG_MESSAGE ("ServerCommand::setCommand");
  waitLockCommand ();
  assert (command == NONE);
  command = command_;
  unlock ();
  DEBUG_MESSAGE ("ServerCommand::setCommand unlocked");
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::setCommand done");
}

void ServerCommand::play ()
{
  setCommand (PLAY);
}

void ServerCommand::pause ()
{
  setCommand (PAUSE);
}

void ServerCommand::continuePlay ()
{
  setCommand (CONTINUE_PLAY);
}

void ServerCommand::prev ()
{
  setCommand (PREV);
}

void ServerCommand::rewind ()
{
  setCommand (REWIND);
}

void ServerCommand::back ()
{
  setCommand (BACK);
}

void ServerCommand::stop ()
{
  setCommand (STOP);
}

void ServerCommand::forward ()
{
  setCommand (FORWARD);
}

void ServerCommand::next ()
{
  DEBUG_MESSAGE ("ServerCommand::next");
  setCommand (NEXT);
  DEBUG_MESSAGE ("ServerCommand:next done");
}

void ServerCommand::open ()
{
  setCommand (OPEN);
}

void ServerCommand::quit ()
{
  setCommand (QUIT);
}

void ServerCommand::setPosition (uint32 position_)
{
  waitLockCommand ();
  command = SET;
  position = position_;
  unlock ();
  broadcast ();
}

void ServerCommand::setSkew (double skew_)
{
#ifdef DEBUG
  cerr << "ServerCommand::setSkew new skew_ = " << skew_ << endl;
#endif
  waitLockCommand ();
  command = SET_SKEW;
  skew = skew_;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandNoteOn (int port, uchar channel,
				      uchar note, uchar velocity)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = note;
  midiParameter[3] = velocity;
  command = MIDI_PORT_NOTE_ON;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandNoteOff (int port, uchar channel,
				       uchar note, uchar velocity)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = note;
  midiParameter[3] = velocity;
  command = MIDI_PORT_NOTE_OFF;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandPolyphonicKeyPressure (int port, uchar channel,
						     uchar note,
						     uchar keyPressure)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = note;
  midiParameter[3] = keyPressure;
  command = MIDI_PORT_POLYPHONIC_KEY_PRESSURE;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandControlChange (int port, uchar channel,
					     uchar msbControl, uchar msbValue,
					     uchar lsbControl, uchar lsbValue)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX$B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = msbControl;
  midiParameter[3] = msbValue;
  midiParameter[4] = channel;
  midiParameter[5] = lsbControl;
  midiParameter[6] = lsbValue;
  command = MIDI_PORT_CONTROL_CHANGE;
  unlock ();
  broadcast ();
}


void ServerCommand::setCommandRPN (int port, uchar channel,
				   uchar rpnMSB, uchar rpnLSB,
				   uchar cc1, uchar cc1Value,
				   uchar cc2, uchar cc2Value)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX$B0z?t%A%'%C%/(B
  assert ((cc1 == 0x06 && cc2 == 0x26) ||
	  (cc1 == 0x60 && cc1Value == 0x00) ||
	  (cc1 == 0x61 && cc1Value == 0x00));
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = rpnMSB;
  midiParameter[3] = rpnLSB;
  midiParameter[4] = cc1;
  midiParameter[5] = cc1Value;
  midiParameter[6] = cc2;
  midiParameter[7] = cc2Value;
  command = MIDI_PORT_RPN;
  unlock ();
  broadcast ();
}


void ServerCommand::setCommandNRPN (int port, uchar channel,
				    uchar nrpnMSB, uchar nrpnLSB,
				    uchar cc1, uchar cc1Value,
				    uchar cc2, uchar cc2Value)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX$B0z?t%A%'%C%/(B
  assert ((cc1 == 0x06 && cc2 == 0x26) ||
	  (cc1 == 0x60 && cc1Value == 0x00) ||
	  (cc1 == 0x61 && cc1Value == 0x00));
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = nrpnMSB;
  midiParameter[3] = nrpnLSB;
  midiParameter[4] = cc1;
  midiParameter[5] = cc1Value;
  midiParameter[6] = cc2;
  midiParameter[7] = cc2Value;
  command = MIDI_PORT_NRPN;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandProgramChange (int port, uchar channel,
					     uchar program)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = program;
  command = MIDI_PORT_PROGRAM_CHANGE;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandChannelPressure (int port, uchar channel,
					       uchar channelPressure)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = channelPressure;
  command = MIDI_PORT_CHANNEL_PRESSURE;
  unlock ();
  broadcast ();
}

void ServerCommand::setCommandPicthBendChange (int port, uchar channel,
					       uchar lsb_, uchar msb_)
{
  waitLockCommand ();
  assert (command == NONE);
  //XXX $B0z?t%A%'%C%/(B
  midiParameter[0] = port;
  midiParameter[1] = channel;
  midiParameter[2] = lsb_;
  midiParameter[3] = msb_;
  command = MIDI_PORT_PITCH_BEND_CHANGE;
  unlock ();
  broadcast ();
}


void ServerCommand::setCommandSysEx  (const uchar* data_, int sysExSize_,
				      int port = -1, int channel_ = -1)
{
  assert (sysExSize_ > 0);
  waitLockCommand ();
  command = MIDI_PORT_SYS_EX;
  sysExSize = sysExSize_;
  delete[] sysExData;
  sysExData = newCopySysEx (data_, sysExSize_);
  channel = channel_;
  unlock ();
  broadcast ();
}

void ServerCommand::setMute (const Sequencer::Muting* muting_, int muteSize_,
			     bool isToggle = false)
{
  assert (muting_ != NULL);
  assert (muteSize_ > 0);
  waitLockCommand ();
  if (isToggle)
    command = TOGGLE_MUTE;
  else
    command = SET_MUTE;
  muteSize = muteSize_;
  delete[] muting;
  muting = Sequencer::Muting::newCopy (muting_, muteSize);
  unlock ();
  broadcast ();
}

void ServerCommand::done ()
{
  DEBUG_MESSAGE ("ServerCommand::done");
  //setCommand (DONE);//XXX this cause dead lock
  //done song$B$,(BSongStatus::DONE$B$H$7$F!"(BMainIO$B$,(BserverCommand->done ()
  //$B$r8F$VA0$K!"B>$N(Bcommand$B$,8F$P$l$k$HBLL\!#(B
  waitLockCommand ();
  command = DONE;
  unlock ();
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::done all done");
}

void ServerCommand::addFile (const string& filename_)
{
  DEBUG_MESSAGE ("ServerCommand::addFile");
  waitLockCommand ();
  assert (command == NONE);
  command = ADD_FILE;
  filename = filename_;
  unlock ();
  DEBUG_MESSAGE ("ServerCommand::addFile unlocked");
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::addFile done");
}

void ServerCommand::playWithFileName (const string& filename_)
{
  DEBUG_MESSAGE ("ServerCommand::playWithFileName");
  waitLockCommand ();
  assert (command == NONE);
  command = PLAY_WITH_FILENAME;
  filename = filename_;
  unlock ();
  DEBUG_MESSAGE ("ServerCommand::playWithFileName unlocked");
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::playWithFileName done");
}

void ServerCommand::setPlayingMode (PlayingStyle::PlayingMode playingMode__)
{
  DEBUG_MESSAGE ("ServerCommand::setPlayingMode");
  waitLockCommand ();
  command = SET_PLAYING_MODE;
  playingMode = playingMode__;
  unlock ();
  DEBUG_MESSAGE ("ServerCommand::setPlayingMode unlocked");
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::setPlayingMode done");
}

void ServerCommand::setDeviceResetType (DeviceResetType::DeviceResetType deviceResetType__)
{
  DEBUG_MESSAGE ("ServerCommand::setDeviceResetType");
  waitLockCommand ();
  command = SET_DEVICE_RESET_TYPE;
  deviceResetType = deviceResetType__;
  unlock ();
  DEBUG_MESSAGE ("ServerCommand::setDeviceResetType unlocked");
  broadcast ();
  DEBUG_MESSAGE ("ServerCommand::setDEviceResetType done");
}


void ServerCommand::getFileList ()
{//XXX
  waitLockCommand ();
  
  unlock ();
  broadcast ();
}
