/*
    WidiServer.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include "IO/MainIONull.h"
#include "Song/SongFactory.h"
#include "WidiServer.h"
#include "util/PThreadAttribute.h"
#include "Sequencer/MIDI.h"

WidiServer::WidiServer (State* state__, Option* option__):
  state (state__),
  songManager (new SongManager ()), serverCommand (),
  widiController (new WidiController (&serverCommand, songManager)),
  option (option__),
  mainIO (NULL), mainDlHandler (NULL),
  playingMode (option->playingMode),
  deviceResetType (option->deviceResetType),
  isQuitWhenFinished (true)
{
  assert (state != NULL);
  assert (option != NULL);

  //preload io
  for (list<string>::const_iterator i = option->preloadIoNames.begin ();
       i != option->preloadIoNames.end ();
       i++)
    preloadDlHandler.push_back (DlHandler::newOpenDlHandler ("", *i, option));

  //mainLoop$B$G1iAU3+;O$9$k$^$G$O(BserverCommand$B$O<u$1IU$1$J$$!#(B
  //mainLoop$B$G2r=|$9$k!#(B
  serverCommand.lock ();


  ///////////////////////////////////////////////////////load io
  if (option->isQuiet)//quiet mode
    mainIO = new MainIONull (state, widiController, option);
  
  for (list<string>::const_iterator i = option->ioNames.begin ();
       i != option->ioNames.end ();
       i++)
    {
      DlHandler* dlHandler = DlHandler::newOpenDlHandler ("IO", *i, option);
      if (dlHandler != NULL)
	{
	  InfoIO* infoIO = InfoIO::getInfoIO (dlHandler, *i);
	  if (infoIO != NULL)
	    {
	      if (mainIO == NULL && infoIO->newMainIO != NULL)
		{
		  assert (mainDlHandler == NULL);
		  mainDlHandler = dlHandler;
		  mainIO = infoIO->newMainIO (state, widiController, option);
		}
	      else
		{
		  ioDlHandler.push_back (dlHandler);
		  io.push_back (InfoIO::createIO(dlHandler, *i, state,
						 widiController, option));
		}

	      if (infoIO->isWaitUserEventWhenFinished == true)
		isQuitWhenFinished = false;
	    }
	  else
	    {//error can not resolve infoIO
	      cerr << "WidiServer::WidiServer: can not resolv InfoIO "
		   << *i << endl;
	    }
	}
      else
	{//error can not load *i
	  cerr << "WidiServer::WidiServer: can not load " << *i
	       << endl;
	}
    }

  if (mainIO == NULL)
    mainIO = new MainIO (state, widiController, option);
  nIO = io.size ();
}

WidiServer::~WidiServer ()
{
  delete mainIO;
  delete mainDlHandler;

  for (list<IO*>::reverse_iterator i = io.rbegin ();
       i != io.rend ();
       i++)
    delete *i;
  for (list<DlHandler*>::reverse_iterator i = ioDlHandler.rbegin ();
       i != ioDlHandler.rend ();
       i++)
    delete *i;
  for (list<DlHandler*>::reverse_iterator i = preloadDlHandler.rbegin ();
       i != preloadDlHandler.rend ();
       i++)
    delete *i;

  delete songManager;
  delete widiController;
}

void WidiServer::createPlayThread (Song* song, Sequencer* sequencer,
				   pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::createPlayThread io creating playLoop");
  //assert (song != NULL);
  if (song == NULL)
    {
      DEBUG_MESSAGE ("WidiServer::createPlayThread song == NULL."
		     "not creating thread");
      return;
    }
  
  
  assert (sequencer != NULL);
  assert (thread != NULL);
  assert (ioThread != NULL);
  
  state->lock ();
  state->reset ();
  state->unlock ();
  song->prePlay ();

  DEBUG_MESSAGE ("WidiServer::createPlayThread mainIO prePlaying");
  mainIO->prePlay ();
  DEBUG_MESSAGE ("WidiServer::createPlayThread io prePlaying");
  for (list<IO*>::iterator i = io.begin ();
       i != io.end ();
       i++)
    (*i)->prePlay ();


  PThreadAttribute threadAttribute;
  
  DEBUG_MESSAGE ("WidiServer::createPlayThread mainIO creating thread.");
  mainIO->createMainLoopThread (&ioThread[0],
				&threadAttribute.getPThreadAttribute ());
  DEBUG_MESSAGE ("WidiServer::createPlayThread IO creating threads.");
  int j = 0;
  for (list<IO*>::iterator i = io.begin ();
       i != io.end ();
       i++, j++)
    (*i)->createPlayLoopThread (&ioThread[j+1],
				&threadAttribute.getPThreadAttribute ());

  //send reset data
  switch (deviceResetType)
    {
    case DeviceResetType::NONE:
      //no reset data are send
      break;
    case DeviceResetType::GM_SYSTEM_ON:
      sequencer->exclusive (MIDI::GM_SYSTEM_ON_DATA,
			    MIDI::GM_SYSTEM_ON_LENGTH);
      break;
    case DeviceResetType::GS_RESET:
      sequencer->exclusive (MIDI::GS_RESET_DATA,
			    MIDI::GS_RESET_LENGTH);
      break;
    case DeviceResetType::SYSTEM_MODE_SET:
      sequencer->exclusive (MIDI::SYSTEM_MODE_SET_DATA,
			    MIDI::SYSTEM_MODE_SET_LENGTH);
      break;
    case DeviceResetType::XG_SYSTEM_ON:
      sequencer->exclusive (MIDI::XG_SYSTEM_ON_DATA,
			    MIDI::XG_SYSTEM_ON_LENGTH);
      break;
    }
  

  DEBUG_MESSAGE ("WidiServer::createPlayThread song creating playThread");
  song->createPlayThread (thread, &threadAttribute.getPThreadAttribute (),
			  *sequencer);
  DEBUG_MESSAGE ("WidiServer::createPlayThread song creating playThread done");
}

void WidiServer::joinPlayThread (pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::joinPlayThread called.");
  assert (thread != NULL);
  assert (ioThread != NULL);
  
  DEBUG_MESSAGE ("WidiServer::joinPlayThread song joining");
  pthread_join (*thread, NULL);
  DEBUG_MESSAGE ("WidiServer::joinPlayThread song joined");

  DEBUG_MESSAGE ("WidiServer::joinPlayThread io joining");
  for (int i = nIO; i > 0; i--)
    pthread_join (ioThread[i], NULL);
  pthread_join (ioThread[0], NULL);
  DEBUG_MESSAGE ("WidiServer::joinPlayThread io joined");

  //$B<!$N1iAU$N0Y!#(B
  //song$B$,1iAU$7$F$$$J$$$N$K!"L$=hM}$N(Bsong command$B$,;D$C$F$$$k$N$O(B
  //$B$H$F$b$^$:$$!#(B
  DEBUG_MESSAGE ("WidiServer::joinPlayThread songCommand locking.");
  songCommand.lock ();
  DEBUG_MESSAGE ("WidiServer::joinPlayThread songCommand locked.");
  DEBUG_MESSAGE ("WidiServer::joinPlayThread state locking.");
  state->lock ();
  DEBUG_MESSAGE ("WidiServer::joinPlayThread state locked.");
  songCommand.reset ();
  state->reset ();
  state->unlock ();
  DEBUG_MESSAGE ("WidiServer::joinPlayThread state unlocked.");
  songCommand.unlock ();
  DEBUG_MESSAGE ("WidiServer::joinPlayThread songCommand unlocked.");
  
  for (list<IO*>::reverse_iterator i = io.rbegin ();
       i != io.rend ();
       i++)
    (*i)->postPlay ();
  mainIO->postPlay ();

  DEBUG_MESSAGE ("WidiServer::joinPlayThread done.");
}

void WidiServer::mainLoop (list<string> files, Sequencer* sequencer_)
{
  //////////////////////////////////////////////////////////////setup
  Sequencer* sequencer = sequencer_;
  assert (sequencer != NULL);
  sequencer->reset ();
  sequencer->sync ();

  mainIO->initialize ();
  for (list<IO*>::iterator i = io.begin ();
       i != io.end ();
       i++)
    (*i)->initialize ();

  for (list<string>::const_iterator i = files.begin ();
       i != files.end ();
       i++)
    songManager->appendSongListItem (*i);
  
  ////////////////////////////////////////////////////////////////$B1iAU3+;O(B
  pthread_t* ioThread = new pthread_t[nIO+1];
  //list<string>::iterator i = files.begin ();
  songManager->moveToNextSong ();
  Song* song = SongFactory::newSong (songManager->getSongName (),
				     state, &songCommand, *mainIO, option);
  //assert (song != NULL);
  pthread_t thread;
  if (song != NULL)
    createPlayThread (song, sequencer, &thread, ioThread);

  ///////////////////////////////////////////serverCommand$B$r<u$1IU$1$k=`Hw(B
  serverCommand.reset ();
  serverCommand.unlock ();
  serverCommand.broadcast ();
  
  /////////////////////////////////////////////////////////command$BBT$A%k!<%W(B
  for (;;)
    {
      ServerCommand tmpCommand;
      serverCommand.lock ();
      DEBUG_MESSAGE ("WidiServer::mainLoop serverCommand locked.waiting serverCommand");
      while (serverCommand.peekCommand () == ServerCommand::NONE)
	serverCommand.waitCond ();
      DEBUG_MESSAGE ("WidiServer::mainLoop serverCommand changed");
      serverCommand.getCommand (&tmpCommand);
      DEBUG_MESSAGE ("WidiServer::mainLoop serverCommand gotten");
      serverCommand.unlock ();
      DEBUG_MESSAGE ("WidiServer::mainLoop serverCommand unlocked");
      serverCommand.broadcast ();
      switch (tmpCommand.peekCommand ())
	{
	case ServerCommand::NONE:
	  //not reached
	  cerr << "WidiServer::mainLoop:ServerCommand.command==NONE." << endl
	       << "Is there error?" << endl;
	  break;

	  
	case ServerCommand::PLAY:
	  //play (song, sequencer, files, i, &thread, ioThread);
	  play (song, sequencer, *songManager, &thread, ioThread);
	  break;
	case ServerCommand::PLAY_WITH_FILENAME:
	  playWithFileName (song, sequencer,
			    //files,
			    *songManager,
			    tmpCommand.getFileName (),
			    &thread, ioThread);
	  break;
	case ServerCommand::PAUSE:
	  lockDoCommand (&unlockPause, &thread, ioThread);
	  break;
	case ServerCommand::CONTINUE_PLAY:
	  lockDoCommand (&unlockContinuePlay, &thread, ioThread);
	  break;
	case ServerCommand::REWIND:
	  lockDoCommand (&unlockRewind, &thread, ioThread);
	  break;
	case ServerCommand::BACK:
	  lockDoCommand (&unlockBack, &thread, ioThread);
	  break;
	case ServerCommand::STOP:
	  stop (song, &thread, ioThread);
	  break;
	case ServerCommand::FORWARD:
	  lockDoCommand (&unlockForward, &thread, ioThread);
	  break;

	case ServerCommand::SET:
	  set (tmpCommand.getPosition (), &thread, ioThread);
	  break;
	  
	case ServerCommand::SET_SKEW:
	  lockDoMidiCommand (&unlockSetSkew, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_NOTE_ON:
	  lockDoMidiCommand (&unlockMidiNoteOn, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_NOTE_OFF:
	  lockDoMidiCommand (&unlockMidiNoteOff, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_POLYPHONIC_KEY_PRESSURE:
	  lockDoMidiCommand (&unlockMidiPolyphonicKeyPressure, song,
			     tmpCommand, &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_CONTROL_CHANGE:
	  lockDoMidiCommand (&unlockMidiControlChange,
			     song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_RPN:
	  lockDoMidiCommand (&unlockMidiRPN, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_NRPN:
	  lockDoMidiCommand (&unlockMidiNRPN, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_PROGRAM_CHANGE:
	  lockDoMidiCommand (&unlockMidiProgramChange,
			     song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_CHANNEL_PRESSURE:
	  lockDoMidiCommand (&unlockMidiChannelPressure,
			     song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_PITCH_BEND_CHANGE:
	  lockDoMidiCommand (&unlockMidiPitchBendChange,
			     song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::MIDI_PORT_SYS_EX:
	  lockDoMidiCommand (&unlockMidiSysEx, song, tmpCommand,
			     &thread, ioThread);
	  break;

	case ServerCommand::SET_MUTE:
	  lockDoMidiCommand (&unlockSetMute, song, tmpCommand,
			     &thread, ioThread);
	  break;
	case ServerCommand::TOGGLE_MUTE:
	  lockDoMidiCommand (&unlockToggleMute, song, tmpCommand,
			     &thread, ioThread);
	  break;
	  
	case ServerCommand::PREV:
	  //prev (song, sequencer, files, i, &thread, ioThread);
	  prev (song, sequencer, *songManager, &thread, ioThread);
	  break;
	case ServerCommand::NEXT:
	  //next (song, sequencer, files, i, &thread, ioThread);
	  next (song, sequencer, *songManager, &thread, ioThread);
	  break;
	case ServerCommand::OPEN:
	  open (&thread, ioThread);
	  break;
	case ServerCommand::DONE:
	  //if (done (song, sequencer, files, i, &thread, ioThread) == false &&
	  if (done (song, sequencer, *songManager, &thread, ioThread) == false &&
	      isQuitWhenFinished)
	    {
	      DEBUG_MESSAGE ("WidiServer::mainLoop all song finished.exiting");
	      //XXX$B8e;OKv!#(B
	      goto QUIT;
	    }
	  DEBUG_MESSAGE ("WidiServer::mainLoop all song finished.waiting event");
	  break;
	case ServerCommand::QUIT://XXX$B8=:_1iAUCf$J$i$=$l$N=*N;=hM}$,I,MW!#(B
	  goto QUIT;
	  break;
	case ServerCommand::SET_PLAYING_MODE:
	  playingMode = tmpCommand.getPlayingMode ();
	  break;
	case ServerCommand::SET_DEVICE_RESET_TYPE:
	  deviceResetType = tmpCommand.getDeviceResetType ();
	  break;
	case ServerCommand::ADD_FILE:
	  files.push_back (tmpCommand.getFileName ());
	  break;
	}
    }
 QUIT:

  //////////////////////////////////////////////////////////////clean up
  delete song;
  delete[] ioThread;
  for (list<IO*>::reverse_iterator i = io.rbegin ();
       i != io.rend ();
       i++)
    (*i)->release ();
  mainIO->release ();
}

void WidiServer::waitLockSongCommand ()
{
  songCommand.lock ();
  while (songCommand.peekCommand () != SongCommand::NONE)
    songCommand.waitCond ();
}

void WidiServer::waitLockState (pthread_t* thread, pthread_t* ioThread)
{
  state->lock ();
  while (state->songStatus.state == SongStatus::DOING_COMMAND ||
	 state->songStatus.state == SongStatus::DONE)
    {
      if (state->songStatus.state == SongStatus::DONE)
	joinAndIgnoreServerCommandDone (thread, ioThread);
      state->waitChange ();
    }
}

void WidiServer::joinAndIgnoreServerCommandDone (pthread_t* thread,
						 pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone called.");

  state->unlock ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone state unlocked.");
  songCommand.unlock ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone songCommand unlocked.");

  //ignore done command
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone serverCommand locking");
  serverCommand.lock ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone serverCommand locked");
  while (serverCommand.peekCommand () != ServerCommand::DONE)
    {
      DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone serverCommand != ServerCommand::DONE");
      serverCommand.reset ();
      serverCommand.broadcast ();
      serverCommand.waitCond ();
    }
  serverCommand.reset ();
  serverCommand.unlock ();
  serverCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone serverCommand unlocked");
  
  joinPlayThread (thread, ioThread);

  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone state locking.");
  songCommand.lock ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone state locked.");
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone state locking");
  state->lock ();
  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone state locked");

  DEBUG_MESSAGE ("WidiServer::joinAndIgnoreServerCommandDone all done");
}

void WidiServer::play (Song*& song, Sequencer* sequencer,
		       //list<string>& files, list<string>::iterator& i,
		       SongManager& songManager__,
		       pthread_t* thread, pthread_t* ioThread)
{

  DEBUG_MESSAGE ("WidiServer::play song locking");
  waitLockSongCommand ();
  DEBUG_MESSAGE ("WidiServer::play state locking");
  waitLockState (thread, ioThread);
  DEBUG_MESSAGE ("WidiServer::play state locked");
  
  if (song != NULL)
    {//song != NULL song may be playing.check the songStatus.
      DEBUG_MESSAGE ("WidiServer::play song != NULL");
      if (state->songStatus.state == SongStatus::SEEKING ||
	  state->songStatus.state == SongStatus::PLAYING)
	{
	  state->unlock ();
	  DEBUG_MESSAGE ("WidiServer::play PLAY songStatus==PLAYING or SEEKING");
	}
      else if (state->songStatus.state == SongStatus::PAUSING)
	{//pause$BCf$@$C$?$N$G1iAU:F3+!#(B
	  DEBUG_MESSAGE ("WidiServer::play PLAY continuePlay");
	  songCommand.setCommand (SongCommand::CONTINUE);
	  state->songStatus.state = SongStatus::DOING_COMMAND;
	  DEBUG_MESSAGE("WidiServer::play PLAY continuePlay done");
	  state->unlock ();
	}
      else if (state->songStatus.state == SongStatus::DONE)
	{
	  DEBUG_MESSAGE ("WidiServer::play songStatus==DONE");
	  joinAndIgnoreServerCommandDone (thread, ioThread);
	  state->unlock ();
	  songManager__.moveToNextSong ();
	  song = SongFactory::newSong (songManager__.getSongName (),
				       state, &songCommand, *mainIO, option);
	  createPlayThread (song, sequencer, thread, ioThread);
	}
      else
	{
	  DEBUG_MESSAGE ("WidiServer::play song != NULL else");
	  state->unlock ();
	  createPlayThread (song, sequencer, thread, ioThread);
	}
    }
  else
    {//song == NULL song is not playing
      DEBUG_MESSAGE ("WidiServer::play PLAY song==NULL");
      //song = SongFactory::newSong (*i, state, &songCommand, *mainIO, option);
      song = SongFactory::newSong (songManager__.getSongName (),
				   state, &songCommand, *mainIO, option);
      state->unlock ();
      createPlayThread (song, sequencer, thread, ioThread);
    }

  songCommand.unlock ();
  songCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::play PLAY all done");
}


void WidiServer::playWithFileName (Song*& song, Sequencer* sequencer,
				   //list<string>& files,
				   SongManager& songManager__,
				   const string& filename,
				   pthread_t* thread, pthread_t* ioThread)
{
  //$B$^$:!"(BsongManager$B$NCf$K(Bfilename$B$,B8:_$9$k$+D4$Y$k!#(B
  if (songManager__.doesHave (filename))
    songManager__.moveTo (filename);
  else
    songManager__.appendSongListItemAndMoveTo (filename);
  
  

  DEBUG_MESSAGE ("WidiServer::playWithFilenName song locking");
  waitLockSongCommand ();
  DEBUG_MESSAGE ("WidiServer::playWithFilenName state locking");
  waitLockState (thread, ioThread);
  DEBUG_MESSAGE ("WidiServer::playWithFilenName state locked");
  
  if (song != NULL)
    {//song != NULL song may be playing.check the songStatus.
      DEBUG_MESSAGE ("WidiServer::playWithFilenName song != NULL");

      if (state->songStatus.state == SongStatus::SEEKING ||
	  (song->getFileName () == filename &&
	   state->songStatus.state == SongStatus::PLAYING))
	{//$B4{$KL\Ev$F$N6J$,1iAUCfKt$O(Bseek$BCf$G$"$k!#(B
	  state->unlock ();
	  DEBUG_MESSAGE ("WidiServer::playWithFileName songStatus==PLAYING");
	}
      else if (song->getFileName () == filename &&
	       state->songStatus.state == SongStatus::PAUSING)
	{//$BL\Ev$F$N6J$,(Bpause$BCf$@$C$?$N$G1iAU:F3+!#(B
	  DEBUG_MESSAGE ("WidiServer::playWithFileName PLAY continuePlay");
	  songCommand.setCommand (SongCommand::CONTINUE);
	  state->songStatus.state = SongStatus::DOING_COMMAND;
	  DEBUG_MESSAGE("WidiServer::playWithFileName PLAY continuePlay done");
	  state->unlock ();
	}
      else
	{//$BL\Ev$F$N6J$r1iAU$9$k!#(B
	  if (state->songStatus.state == SongStatus::PLAYING ||
	      state->songStatus.state == SongStatus::PAUSING)
	    {//$B1iAUCf$N$b$N$r6/@)=*N;!#(B
	      abort (thread, ioThread);
	      state->unlock ();
	    }
	  else if (state->songStatus.state == SongStatus::DONE)
	    {
	      DEBUG_MESSAGE ("WidiServer::playWithFilenName songStatus==DONE");
	      joinAndIgnoreServerCommandDone (thread, ioThread);
	      state->unlock ();
	    }
	  else
	    state->unlock ();
	  
	  DEBUG_MESSAGE ("WidiServer::playWithFilenName deleting song");
	  delete song;
	  DEBUG_MESSAGE ("WidiServer::playWithFileName newing song");
	  song = SongFactory::newSong (filename, state,
				       &songCommand, *mainIO, option);
	  DEBUG_MESSAGE ("WidiServer::playWithFileName newed song");
	  createPlayThread (song, sequencer, thread, ioThread);
	}
    }
  else
    {//song == NULL song is not playing
      DEBUG_MESSAGE ("WidiServer::playWithFilenName PLAY song==NULL");
      song = SongFactory::newSong (filename, state,
				   &songCommand, *mainIO, option);
      state->unlock ();
      createPlayThread (song, sequencer, thread, ioThread);
    }

  songCommand.unlock ();
  songCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::playWithFilenName PLAY all done");
}


void WidiServer::lockDoCommand (unlockDoCommand doCommand,
				pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::lockDoCommand called");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);

  (this->*doCommand) ();
  
  state->unlock ();
  songCommand.unlock ();
  songCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::lockDoCommand all done.");
}


//$B1iAU$r%]!<%:$9$k!#(B
//$B%]!<%:Cf$J$i$=$N$^$^(B
//$B1iAUCf$G$J$1$l$P2?$b$7$J$$!#(B
void WidiServer::unlockPause ()
{
  if (state->songStatus.state == SongStatus::PLAYING)
    {//There is need to pause
      DEBUG_MESSAGE ("WidiServer::pause really pausing.song");
      songCommand.setCommand (SongCommand::PAUSE);
      state->songStatus.state = SongStatus::DOING_COMMAND;
    }
}

//$B%]!<%:$7$F$$$?1iAU$r:F3+$9$k!#(B
//$B%]!<%:$7$F$$$J$1$l$P2?$b$7$J$$!#(B
void WidiServer::unlockContinuePlay ()
{
  if (state->songStatus.state == SongStatus::PAUSING)
    {
      state->songStatus.state = SongStatus::DOING_COMMAND;
      songCommand.setCommand (SongCommand::CONTINUE);
    }
}

//$B:F@80LCV$r:G=i$KLa$9!#(B
void WidiServer::unlockRewind ()
{
  if (state->songStatus.state == SongStatus::PLAYING ||
      state->songStatus.state == SongStatus::PAUSING)
    {
      songCommand.setCommand (SongCommand::REWIND);
      state->songStatus.state = SongStatus::DOING_COMMAND;
    }
}


void WidiServer::unlockBack ()
{
  if (state->songStatus.state == SongStatus::PLAYING ||
      state->songStatus.state == SongStatus::PAUSING)
    {
      songCommand.setCommand (SongCommand::BACK);
      state->songStatus.state = SongStatus::SEEKING;
    }
}

void WidiServer::stop (Song*& song, pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::stop called.");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);
  
  if (state->songStatus.state != SongStatus::SEEKING)
    {
      DEBUG_MESSAGE ("WidiServer::stop STOP aborting");
      abort (thread, ioThread);
      DEBUG_MESSAGE ("WidiServer::stop STOP io aborted");
      state->songStatus.state = SongStatus::STOPPED;
      state->broadcast ();
    }

  state->unlock ();
  songCommand.unlock ();
  DEBUG_MESSAGE ("WidiServer::stop done.");  
}

void WidiServer::unlockForward ()
{
  if (state->songStatus.state == SongStatus::PLAYING ||
      state->songStatus.state == SongStatus::PAUSING)
    {
      songCommand.setCommand (SongCommand::FORWARD);
      state->songStatus.state = SongStatus::SEEKING;
    }
}

//position$B$N0LCV$K1iAU$N2U=j$r0\F0$5$;$k!#(B
void WidiServer::set (ulong position, pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::set called.");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);

  if (state->songStatus.state == SongStatus::PLAYING ||
      state->songStatus.state == SongStatus::PAUSING)
    {
      songCommand.setCommandSET (position);
      state->songStatus.state = SongStatus::SEEKING;
    }
  
  state->unlock ();
  songCommand.unlock ();
  songCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::set done.");
}

void WidiServer::lockDoMidiCommand (unlockDoMidiCommand doCommand,
				    Song* song,  ServerCommand& tmpCommand,
				    pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::lockDoMidi called.");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);

    if (state->songStatus.state == SongStatus::PLAYING)
    {
      (this->*doCommand) (tmpCommand);
      state->songStatus.state = SongStatus::DOING_COMMAND;
    }
  
  state->unlock ();
  songCommand.unlock ();
  songCommand.broadcast ();
  DEBUG_MESSAGE ("WidiServer::lockDoMidi done.");
}



void WidiServer::unlockSetSkew (ServerCommand& tmpCommand)
{
  songCommand.setCommandSkew (tmpCommand.getSkew ());
}

void WidiServer::unlockMidiNoteOn (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandNoteOn (midiParameter[0], midiParameter[1],
				midiParameter[2], midiParameter[3]);
}

void WidiServer::unlockMidiNoteOff (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandNoteOff (midiParameter[0], midiParameter[1],
				 midiParameter[2], midiParameter[3]);
}

void WidiServer::unlockMidiPolyphonicKeyPressure (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandPolyphonicKeyPressure (midiParameter[0],
					       midiParameter[1],
					       midiParameter[2],
					       midiParameter[3]);
}

void WidiServer::unlockMidiControlChange (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandControlChange (midiParameter[0], midiParameter[1],
				       midiParameter[2], midiParameter[3],
				       midiParameter[4], midiParameter[5]);
}

void WidiServer::unlockMidiRPN (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandRPN (midiParameter[0], midiParameter[1],
			     midiParameter[2], midiParameter[3],
			     midiParameter[4], midiParameter[5],
			     midiParameter[6], midiParameter[7]);
}

void WidiServer::unlockMidiNRPN (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandNRPN (midiParameter[0], midiParameter[1],
			      midiParameter[2], midiParameter[3],
			      midiParameter[4], midiParameter[5],
			      midiParameter[6], midiParameter[7]);
}

void WidiServer::unlockMidiProgramChange (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandProgramChange (midiParameter[0], midiParameter[1],
				       midiParameter[2]);
}

void WidiServer::unlockMidiChannelPressure (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandChannelPressure (midiParameter[0], midiParameter[1],
					 midiParameter[2]);
}

void WidiServer::unlockMidiPitchBendChange (ServerCommand& tmpCommand)
{
  const uchar* midiParameter = tmpCommand.getMidiParameter ();
  songCommand.setCommandPitchBendChange (midiParameter[0], midiParameter[1],
					 midiParameter[2], midiParameter[3]);
}

void WidiServer::unlockMidiSysEx (ServerCommand& tmpCommand)
{
  songCommand.setCommandSysEx (tmpCommand.getSysExData (),
			       tmpCommand.getSysExDataSize ());
}

void WidiServer::unlockSetMute (ServerCommand& tmpCommand)
{
  songCommand.setCommandMute (tmpCommand.getMuting (),
			      tmpCommand.getMuteSize ());
}

void WidiServer::unlockToggleMute (ServerCommand& tmpCommand)
{
  songCommand.setCommandToggleMute (tmpCommand.getMuting (),
				    tmpCommand.getMuteSize ());
}


//$B1iAU$rCf;_$9$k!#(B
void WidiServer::abort (pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::abort called.");
  assert (state->songStatus.state != SongStatus::DOING_COMMAND);
  assert (state->songStatus.state != SongStatus::SEEKING);
  
  if (state->songStatus.state == SongStatus::PLAYING ||
      state->songStatus.state == SongStatus::PAUSING)
    {//$BK\Ev$K(Babort$B$5$;$k!#(B
      DEBUG_MESSAGE ("WidiServer::abort really aborting.");
      songCommand.setCommand (SongCommand::ABORT);
      state->songStatus.state = SongStatus::DOING_COMMAND;
      songCommand.broadcast ();
      
      DEBUG_MESSAGE ("WidiServer::abort ignoring 'done' command");
      //abort$B$7$?$N$G$=$l$r3NG'$9$k!#(B
      joinAndIgnoreServerCommandDone (thread, ioThread);
    }
  
  DEBUG_MESSAGE ("WidiServer::abort all done.");
}


//$BA06J$X(B
void WidiServer::prev (Song*& song, Sequencer* sequencer,
		       //list<string>& files, list<string>::iterator& i,
		       SongManager& songManager__,
		       pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::prev called.");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);

  if (state->songStatus.state != SongStatus::SEEKING)
    {
      DEBUG_MESSAGE ("WidiServer::prev PREV aborting");
      abort (thread, ioThread);
      state->unlock ();
      songCommand.unlock ();
      
      songManager__.moveToPrevSong ();
      song = SongFactory::newSong (songManager__.getSongName (),
				   state, &songCommand, *mainIO, option);
      
      createPlayThread (song, sequencer, thread, ioThread);
    }
  else
    {
      state->unlock ();
      songCommand.unlock ();
    }

  DEBUG_MESSAGE ("WidiServer::prev PREV all done");
}

void WidiServer::next (Song*& song, Sequencer* sequencer,
		       //list<string>& files, list<string>::iterator& i,
		       SongManager& songManager__,
		       pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::prev called.");
  waitLockSongCommand ();
  waitLockState (thread, ioThread);
  
  if (state->songStatus.state != SongStatus::SEEKING)
    {
      DEBUG_MESSAGE ("WidiServer::next NEXT aborting");
      abort (thread, ioThread);
      DEBUG_MESSAGE ("WidiServer::next NEXT io aborted");
      state->unlock ();
      songCommand.unlock ();
      
      songManager__.moveToNextSong ();
      song = SongFactory::newSong (songManager__.getSongName (),
				   state, &songCommand, *mainIO, option);
      
      createPlayThread (song, sequencer, thread, ioThread);
    }
  else
    {
      state->unlock ();
      songCommand.unlock ();
    }
  
  DEBUG_MESSAGE ("WidiServer::next NEXT all done");
}


void WidiServer::open (pthread_t* thread, pthread_t* ioThread)
{
  //XXX
}

bool WidiServer::done (Song*& song, Sequencer* sequencer,
		       //list<string>& files, list<string>::iterator& i,
		       SongManager& songManager__,
		       pthread_t* thread, pthread_t* ioThread)
{
  DEBUG_MESSAGE ("WidiServer::done called.");

  joinPlayThread (thread, ioThread);
  
  DEBUG_MESSAGE ("WidiServer::done DONE song deleting");
  delete song;

  songManager__.moveToNextSong ();
  song = SongFactory::newSong (songManager__.getSongName (),
			       state, &songCommand, *mainIO, option);
  if (songManager__.isEnd ())
    {
      DEBUG_MESSAGE ("WidiServer::done all done. finished playing.");
      return false;
    }
  
  switch (playingMode)
    {
    case PlayingStyle::CONTINUOUS_PLAY:
      createPlayThread (song, sequencer, thread, ioThread);
      break;
    case PlayingStyle::STOPS_PER_SONG:
      //do nothing
      break;
    }
  
  
  DEBUG_MESSAGE ("WidiServer::done all done. continue playing.");
  return true;
}
