/*
    widi.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cstdio>
#include <cstdlib>
#include <cctype>
#include <sys/ioctl.h>
#include <iostream>
#include <unistd.h>
#include <getopt.h>

#include <list>

#include "IO/MainIO.h"
#include "IO/InfoIO.h"
#include "Sequencer/Sequencer.h"
#include "Sequencer/SequencerNull.h"
#include "WidiServer.h"
#include "util/DlHandler.h"

void printUsage (const char* progName)
{
  cerr << "Usage:" << progName  << " [options] midi-file...\n" 
       << "options:\n"
       << "--convert,-c:\tconvert mode\n"
       << "--help,-h:\tprint this message and exit.\n"
       << "--io <module name> ,-i <module name>:\n"
       << "--loop,-l:\tloop play\n"
       << "--mididevice <midi deivce number>,-m <midi device number>:\n"
       << "--norealtime,-n:\tbuffering mode\n"
       << "--preload <preload module name>,-p <preload module name>:\n"
       << "--quiet,-q:\tquiet mode\n"
       << "--random,-r:\trandom play\n"
    //<< "--roland,-R:\tuse rolland sys ex(CHANNEL MSG RX PORT) to change port\n"
       << "--serial,-s:\tserial midi mode(use 0xF5 pp to change port)\n"
       << "--version,-V:\tprint version and exit.\n"
       << flush;
}

int
main (int argc,char** argv)
{
  char* progName=argv[0];
  if (argc<=1)
    {
      printUsage (progName);
      exit (-1);
    }
  
  srand (getpid ());
  State state;
  Option option;
  option.isQuickStart=true;
  option.isRealTime=true;
  option.modulePath=new char*[2];
  option.modulePath[0]=new char[strlen (WIDI_LIB_DIR)+1];
  strcpy (option.modulePath[0],WIDI_LIB_DIR);
  option.modulePath[1]=NULL;
  option.mainIO=NULL;
#define MAX_IOS 256
  option.preloadIoNames=new char*[MAX_IOS];
  for (int i=0;i<MAX_IOS;i++)
    option.preloadIoNames[i]=NULL;
  option.ioNames=new char*[MAX_IOS];
  for (int i=0;i<MAX_IOS;i++)
    option.ioNames[i]=NULL;
  
  struct option optionDscTbl[]=
  {
    {"roland",    no_argument,      NULL,'R'},
    {"version",   no_argument,      NULL,'V'},
    {"convert",   no_argument,      NULL,'c'},
    {"help",      no_argument,      NULL,'h'},
    {"io",        required_argument,NULL,'i'},
    {"loop",      no_argument,      NULL,'l'},
    {"mididevice",required_argument,NULL,'m'},
    {"norealtime",no_argument,      NULL,'n'},
    {"preload",   required_argument,NULL,'p'},
    {"quiet",     no_argument,      NULL,'q'},
    {"random",    no_argument,      NULL,'r'},
    {"serial",    no_argument,      NULL,'s'},
    {0,0,0,0}
  };

  int optionIndex=0;
  int c=0;
  int npreloadio=0;
  int nio=0;
  while ((c=getopt_long (argc,argv,"RVchi:lm:np:qrs",
			 optionDscTbl,&optionIndex))!=EOF)
    {
      switch (c)
	{
	case 'R':
	  option.portChangeType=Option::USE_ROLAND_SYS_EX;
	  break;	  
	case 'V':
	  cout << progName << " version=" << VERSION << endl;
	  exit (0);
	  break;
	case 'c':
	  option.isQuickStart=false;
	  break;
	case 'h':
	  printUsage (progName);
	  exit (0);
	  break;
	case 'i':
	  if (option.mainIO!=NULL)
	    {
	      option.ioNames[nio++]=option.mainIO;
	      option.mainIO=NULL;
	    }
	  delete[] option.mainIO;
	  option.mainIO=new char[strlen (optarg)+1];
	  strcpy (option.mainIO,optarg);
	  break;
	case 'l':
	  option.playStyle=Option::LOOP;
	  break;
	case 'm':
	  if (isdigit (optarg[0]))
	    option.midiDeviceNum=atoi (optarg);
	  else
	    {
	      printUsage (progName);
	      exit (-1);
	    }
	  break;
	case 'n':
	  option.isRealTime=false;
	  break;
	case 'p':
	  {
	    const int length=strlen (optarg);
	  option.preloadIoNames[npreloadio]=
	    //new char[strlen (optarg)+1];
	    new char[length+1];
	  strcpy (option.preloadIoNames[npreloadio],optarg);
	  }
	  npreloadio++;
	  break;
	case 'q':
	  option.isQuiet=true;
	  break;
	case 'r':
	  option.playStyle=Option::RANDOM;
	  break;
	case 's':
	  option.portChangeType=Option::SERIAL;
	  break;
	default:
	  cerr << "ERROR:getopt returned character code=" << c << endl;
	  printUsage (progName);
	  exit (-1);
	  break;
	}
    }

  Sequencer* sequencer=new SequencerNull (&state,&option);
    
  WidiServer server (&state,&option);

  list<const char*> files;
  for (int i=optind;i<argc;i++)
    files.push_back (argv[i]);
  server.mainLoop (files,sequencer);
  
  delete sequencer;
  
  delete[] option.modulePath[1];
  delete[] option.modulePath;
  delete[] option.mainIO;
  for (int i=0;i<MAX_IOS;i++)
    delete[] option.ioNames[i];
  delete[] option.ioNames;
  for (int i=0;i<MAX_IOS;i++)
    delete[] option.preloadIoNames[i];
  delete[] option.preloadIoNames;
#undef MAX_IOS
  exit (0);
}

