/*
    widi.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cstdio>
#include <cstdlib>
#include <cctype>
#include <sys/ioctl.h>
#include <iostream>
#include <unistd.h>

#ifdef HAVE_GETOPT_LONG
# include <getopt.h>
#else
# include "getopt.h"
#endif

#include <list>

#include "IO/MainIO.h"
#include "IO/InfoIO.h"
#include "Sequencer/Sequencer.h"
#include "WidiServer.h"
#include "util/DlHandler.h"
#include "Song/SongFactory.h"

using namespace std;

void printUsage (const char* progName)
{
  cerr << "Usage:" << progName << " [options] midi-file...\n" 
       << "options:\n"
       << "--convert,-c:\tconvert mode\n"
       << "--device {none,gm,gs,88,xg},-d {none,gm,gs,88,xg}:\n"
       << "\tsend reset data before playing each song."
       << " if not specified, no reset data are send.\n"
       << "\tnone:no data are send.\n"
       << "\tgm:gm system on are send.\n"
       << "\tgs:gs reset are send.\n"
       << "\t88:system mode set are send.\n"
       << "\txg:xg system on send.\n"
       << "--format <format name>,-f <format name>:\n"
       << "--help,-h:\tprint this message and exit.\n"
       << "--io <module name> ,-i <module name>:\n"
       << "--loop,-l:\tloop play\n"
       << "--mididevice <midi deivce number>,-m <midi device number>:\n"
       << "--norealtime,-n:\tbuffering mode\n"
       << "--preload <preload module name>,-p <preload module name>:\n"
       << "--plane,-P:\toff plane mask of XWrd\n"
       << "--quiet,-q:\tquiet mode\n"
       << "--random,-r:\trandom play\n"
    //<< "--roland:\tuse rolland sys ex(CHANNEL MSG RX PORT) to change port\n"
       << "--sequencer <sequencer name>,-s <sequencer name>:\n"
       << "--serial,-S:\tserial midi mode(send 0xF5 pp to change port)\n"
       << "--version,-V:\tprint version and exit.\n"
       << "--wrddebug,-W:\tprint macro command of wrd.\n"
       << flush;

  cerr << "Usable interface modules"
       << ": NCurses"
       << ", NLevel,"
       << ", NTrace,"
       << ", NWrd"
       << ", WrdDumb"
       << ", XWrd"
       << ", Xlib"
       << ", XSherry"
       << "."
       << endl;
  cerr << "Usable sequencer modules"
       << ": ALSA"
       << ", Serial"
       << ", Null"
       << "."
       << endl;
  cerr << "Usable format modules"
       << ": (none)"
       << "."
       << endl;
  
}

void setSigAction ()
{
#ifdef SA_RESTART
  sigaction sigAction;
  sigAction.sa_handler = SIG_DFL;
  sigemptyset (&sigAction.sa_mask);
  sigAction.sa_flags = SA_RESTART;
  //sigAction.sa_restorer = NULL;
  //sigaction (SIGINT, &sigAction, NULL);
  //sigaction (SIGUSR1, &sigAction, NULL);
  //sigaction (SIGUSR2, &sigAction, NULL);
  //sigaction (SIGABORT, &sigAction, NULL);
#endif // SA_RESTAET
}


int
main (int argc,char** argv)
{
  char* progName = argv[0];
  if (argc <= 1)
    {
      printUsage (progName);
      exit (1);
    }
#ifndef __FreeBSD__
  pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL);
#endif
  //setSigAction ();
  

  State state;
  Option option;
  option.isQuickStart = true;
  option.isRealTime = true;
  option.modulePath.push_back (WIDI_LIB_DIR);
  string sequencerName;
  list<string> formatList;
  
  struct option optionDscTbl[]=
  {
    {"plane",     no_argument,      NULL,'P'},
    {"roland",    no_argument,      NULL,'R'},
    {"serial",    no_argument,      NULL,'S'},
    {"version",   no_argument,      NULL,'V'},
    {"wrddebug",  no_argument,      NULL,'W'},
    {"convert",   no_argument,      NULL,'c'},
    {"device",    required_argument,NULL,'d'},
    {"format",    required_argument,NULL,'f'},
    {"help",      no_argument,      NULL,'h'},
    {"io",        required_argument,NULL,'i'},
    {"loop",      no_argument,      NULL,'l'},
    {"mididevice",required_argument,NULL,'m'},
    {"norealtime",no_argument,      NULL,'n'},
    {"preload",   required_argument,NULL,'p'},
    {"quiet",     no_argument,      NULL,'q'},
    {"random",    no_argument,      NULL,'r'},
    {"sequencer", required_argument,NULL,'s'},
    {0,0,0,0}
  };

  int optionIndex = 0;
  int c = 0;
  while ((c = getopt_long (argc,argv,"PRSVWcd:f:hi:lm:np:qrs:",
			   optionDscTbl,&optionIndex)) != EOF)
    {
      switch (c)
	{
	case 'P':
	  option.xwrdDoesPlaneMask = false;
	  break;
	case 'R':
	  option.portChangeType = Option::USE_ROLAND_SYS_EX;
	  break;	  
	case 'S':
	  option.portChangeType = Option::SERIAL;
	  break;
	case 'V':
	  cout << progName << " version=" << VERSION << endl;
	  exit (0);
	  break;
	case 'W':
	  option.wrdDoesCerrDebug = true;
	  break;
	case 'c':
	  option.isQuickStart = false;
	  break;
	case 'd':
	  if (strcmp (optarg, "none") == 0)
	    option.deviceResetType = DeviceResetType::NONE;
	  else if (strcasecmp (optarg, "gm") == 0)
	    option.deviceResetType = DeviceResetType::GM_SYSTEM_ON;
	  else if (strcasecmp (optarg, "gs") == 0)
	    option.deviceResetType = DeviceResetType::GS_RESET;
	  else if (strcasecmp (optarg, "88") == 0)
	    option.deviceResetType = DeviceResetType::SYSTEM_MODE_SET;
	  else if (strcasecmp (optarg, "xg") == 0)
	    option.deviceResetType = DeviceResetType::XG_SYSTEM_ON;
	  else
	    {
	      printUsage (progName);
	      exit (1);
	    }
	  break;
	case 'f':
	  formatList.push_back (optarg);
	  break;
	case 'h':
	  printUsage (progName);
	  exit (0);
	  break;
	case 'i':
	  if (optarg != NULL)
	    {
	      option.ioNames.push_back (optarg);
	    }
	  else
	    {
	      printUsage (progName);
	      exit (1);
	    }
	  break;
	case 'l':
	  option.loopStyle = PlayingStyle::LOOP;
	  break;
	case 'm':
	  if (isdigit (optarg[0]))
	    option.midiDeviceNum = atoi (optarg);
	  else
	    {
	      printUsage (progName);
	      exit (1);
	    }
	  break;
	case 'n':
	  option.isRealTime = false;
	  break;
	case 'p':
	  option.preloadIoNames.push_back (optarg);
	  break;
	case 'q':
	  option.isQuiet = true;
	  break;
	case 'r':
	  option.playingStyle = PlayingStyle::RANDOM;
	  break;
	case 's':
	  sequencerName = optarg;
	  break;
	default:
	  cerr << "ERROR:getopt returned character code=" << c << endl;
	  printUsage (progName);
	  exit (1);
	  break;
	}
    }
  
  for (list<string>::iterator i = formatList.begin ();
       i != formatList.end ();
       i++)
    SongFactory::addSongModule (*i, &option);
  
  DlHandler* dlHandler = NULL;
  Sequencer* sequencer;
  try
    {
      sequencer = Sequencer::newSequencer (&dlHandler, sequencerName,
					   &state, &option);
    }
  catch (Sequencer::SequencerError&)
    {
      cerr << "widi:can not initialize midi device" << endl;
      delete dlHandler;
      exit (1);
    }
    
  WidiServer* server = new WidiServer (&state, &option);

  list<string> files;
  for (int i = optind; i < argc; i++)
    files.push_back (argv[i]);
  server->mainLoop (files, sequencer);
  
  delete server;
  delete sequencer;
  
  delete dlHandler;
  exit (0);
}

